@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-2006 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2007 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: CmdExec.cc,v 1.125 2006/08/22 06:47:25 lav Exp $ */
+/* $Id: CmdExec.cc,v 1.140 2007/10/24 14:38:48 lav Exp $ */
 
 #include <config.h>
 #include <errno.h>
@@ -26,7 +26,6 @@
 #include <ctype.h>
 #include <fcntl.h>
 #include "CmdExec.h"
-#include "xmalloc.h"
 #include "xstring.h"
 #include "SignalHook.h"
 #include "alias.h"
@@ -43,6 +42,7 @@
 #define RL_PROMPT_END_IGNORE	'\002'
 
 #define super SessionJob
+#define waiting_num waiting.count()
 
 static ResDecl
    res_default_proto	   ("cmd:default-protocol","ftp",0,0),
@@ -68,7 +68,7 @@ static ResDecl
 
 CmdExec	 *CmdExec::cwd_owner;
 CmdExec	 *CmdExec::chain;
-CmdExec  *CmdExec::top;
+JobRef<CmdExec> CmdExec::top;
 
 void  CmdExec::SaveCWD()
 {
@@ -100,15 +100,7 @@ void CmdExec::FeedCmd(const char *c)
 {
    partial_cmd=false;
    start_time=now;
-   if(cmd_buf==0)
-   {
-      cmd_buf=next_cmd=xstrdup(c);
-      return;
-   }
-   int len=strlen(next_cmd);
-   memmove(cmd_buf,next_cmd,len);
-   cmd_buf=next_cmd=(char*)xrealloc(cmd_buf,len+strlen(c)+1);
-   strcpy(next_cmd+len,c);
+   cmd_buf.Put(c);
 };
 
 void CmdExec::PrependCmd(const char *c)
@@ -118,17 +110,9 @@ void CmdExec::PrependCmd(const char *c)
    int len=strlen(c);
    int nl=(len>0 && c[len-1]!='\n');
 
-   int next_cmd_len=xstrlen(next_cmd);
-   int next_cmd_offs=next_cmd-cmd_buf;
-   if(next_cmd_offs<len+1)
-      cmd_buf=(char*)xrealloc(cmd_buf,len+nl+next_cmd_len+1);
-   if(next_cmd_len>0)
-      memmove(cmd_buf+len+nl,cmd_buf+next_cmd_offs,next_cmd_len);
-   cmd_buf[len+nl+next_cmd_len]=0;
-   memcpy(cmd_buf,c,len);
    if(nl)
-      cmd_buf[len]='\n';
-   next_cmd=cmd_buf;
+      cmd_buf.Prepend("\n",1);
+   cmd_buf.Prepend(c,len);
 
    if(alias_field>0)
       alias_field+=len+nl;
@@ -137,8 +121,9 @@ void CmdExec::PrependCmd(const char *c)
 int CmdExec::find_cmd(const char *cmd_name,const struct cmd_rec **ret)
 {
    int part=0;
-   const cmd_rec *c;
-   for(c=dyn_cmd_table?dyn_cmd_table:static_cmd_table; c->name; c++)
+   const cmd_rec *c=dyn_cmd_table?dyn_cmd_table.get():static_cmd_table;
+   const int count=dyn_cmd_table?dyn_cmd_table.count():1024;
+   for(int i=0; i<count && c->name; i++,c++)
    {
       if(!strcasecmp(c->name,cmd_name))
       {
@@ -195,9 +180,8 @@ void  CmdExec::exec_parsed_command()
 
    if(ResMgr::QueryBool("cmd:trace",0))
    {
-      char *c=args->CombineQuoted();
-      printf("+ %s\n",c);
-      xfree(c);
+      xstring_ca c(args->CombineQuoted());
+      printf("+ %s\n",c.get());
    }
 
 restart:
@@ -220,9 +204,7 @@ restart:
       args->setarg(0,c->name); // in case it was abbreviated
       args->rewind();
 
-      if(cmd==0)
-	 cmd=args->Combine();
-
+      xstring_ca cmdline(args->Combine());   // save the cmdline
       Job *new_job=0;
 
       if(c->creator==0)
@@ -253,11 +235,8 @@ restart:
       {
 	 if(new_job->jobno<0)
 	    new_job->AllocJobno();
-	 if(cmd && new_job->cmdline==0)
-	 {
-	    new_job->cmdline=cmd;
-	    cmd=0;
-      	 }
+	 if(!new_job->cmdline)
+	    new_job->cmdline.set_allocated(cmdline.borrow());
 	 new_job->SetParentFg(this,!background);
       }
       AddWaiting(new_job);
@@ -287,12 +266,7 @@ void CmdExec::ExecParsed(ArgV *a,FDStream *o,bool b)
 {
    Enter();
 
-   if(args)
-      delete args;
    args=a;
-   xfree(cmd);
-   cmd=args->Combine();
-   delete output;
    output=o;
    background=b;
    condition=COND_ANY;
@@ -303,7 +277,7 @@ void CmdExec::ExecParsed(ArgV *a,FDStream *o,bool b)
 
 bool CmdExec::Idle()
 {
-   return(waiting_num==0 && (next_cmd==0 || *next_cmd==0 || partial_cmd));
+   return(waiting_num==0 && (cmd_buf.Size()==0 || partial_cmd));
 }
 
 int CmdExec::Done()
@@ -313,8 +287,6 @@ int CmdExec::Done()
 
 void CmdExec::RemoveFeeder()
 {
-   alias_field=0;
-   xfree(cmd_buf); next_cmd=cmd_buf=0;
    free_used_aliases();
 
    if(!feeder)
@@ -324,40 +296,38 @@ void CmdExec::RemoveFeeder()
    if(interactive && feeder->prev==0)
       cwd_history.Set(session);
 
-   CmdFeeder *tmp=feeder->prev;
-   next_cmd=cmd_buf=feeder->saved_buf;
+   cmd_buf.Empty();
+   cmd_buf.Put(feeder->saved_buf);
    partial_cmd=false;
    if(feeder==queue_feeder)
       queue_feeder=0;
-   delete feeder;
-   feeder=tmp;
+   delete replace_value(feeder,feeder->prev);
    Reconfig(0);
 }
 
 void CmdExec::ReuseSavedSession()
 {
-   Reuse(saved_session);
    saved_session=0;
 }
 void CmdExec::RevertToSavedSession()
 {
    if(saved_session==0)
       return;
-   ChangeSession(saved_session);
-   saved_session=0;
+   ChangeSession(saved_session.borrow());
 }
 void CmdExec::ChangeSlot(const char *n)
 {
-   xfree(slot);
-   slot=0;
    if(!n || !*n)
+   {
+      slot.set(0);
       return;
-   FileAccess *s=ConnectionSlot::FindSession(n);
+   }
+   slot.set(n);
+   const FileAccess *s=ConnectionSlot::FindSession(n);
    if(!s)
       ConnectionSlot::Set(n,session);
    else
       ChangeSession(s->Clone());
-   slot=xstrdup(n);
 }
 
 int CmdExec::Do()
@@ -415,7 +385,7 @@ int CmdExec::Do()
 		     FileAccess::Path new_cwd(session->GetNewCwd());
 		     new_cwd.Change(0,is_file,loc);
 		     session->PathVerify(new_cwd);
-		     Roll(session);
+		     session->Roll();
 		     return MOVED;
 		  }
 		  session->Close();
@@ -490,12 +460,9 @@ int CmdExec::Do()
 	    const char *pat=args->getnext();
 	    if(!pat)
 	    {
-	       delete glob;
 	       glob=0;
 	       // it was last argument
-	       delete args;
-	       args=args_glob;
-	       args_glob=0;
+	       args=args_glob.borrow();
 	       builtin=BUILTIN_NONE;
 	       redirections=0;
 	       if(status_line)
@@ -604,7 +571,7 @@ int CmdExec::Do()
    }
 
 try_get_cmd:
-   if(next_cmd==0 || *next_cmd==0 || partial_cmd)
+   if(cmd_buf.Size()==0 || partial_cmd)
    {
       if(feeder)
       {
@@ -618,21 +585,22 @@ try_get_cmd:
 
 	 if(status_line)
 	 {
-	    char *def_title = FormatPrompt(res_default_title.Query(getenv("TERM")));
+	    const char *def_title = FormatPrompt(res_default_title.Query(getenv("TERM")));
 	    status_line->DefaultTitle(def_title);
 	    status_line->Clear();
 	 }
 
-	 char *prompt=MakePrompt();
+	 const char *prompt=MakePrompt();
 	 feeder_called=true;
 	 if(fg)
 	    feeder->Fg();
 	 const char *cmd=feeder->NextCmd(this,prompt);
 	 feeder_called=false;
-	 if(cmd==0)
+	 if(!cmd)
 	 {
-	    if(next_cmd && *next_cmd && partial_cmd)
+	    if(cmd_buf.Size()>0 && partial_cmd)
 	    {
+	       const char *next_cmd=cmd_buf.Get();
 	       if(last_char(next_cmd)!='\n')
 	       {
 		  // missing EOL on last line, add it
@@ -643,7 +611,7 @@ try_get_cmd:
 	    }
 	    if(!feeder->RealEOF() && top_level)
 	    {
-	       *next_cmd=0;
+	       cmd_buf.Empty();
 	       FeedCmd("exit;");
 	       return MOVED;
 	    }
@@ -663,8 +631,7 @@ try_get_cmd:
 	    if(SignalHook::GetCount(SIGINT)>0)
 	    {
 	       SignalHook::ResetCount(SIGINT);
-	       if(next_cmd)
-		  *next_cmd=0;	 // flush unparsed command
+	       cmd_buf.Empty();	 // flush unparsed command
 	       return MOVED;
 	    }
 	 }
@@ -701,7 +668,7 @@ try_get_cmd:
    return MOVED;
 }
 
-void CmdExec::ShowRunStatus(StatusLine *s)
+void CmdExec::ShowRunStatus(const SMTaskRef<StatusLine>& s)
 {
    switch(builtin)
    {
@@ -732,9 +699,8 @@ void CmdExec::PrintStatus(int v,const char *prefix)
    SessionJob::PrintStatus(v,prefix);
    if(builtin)
    {
-      char *s=args->Combine();
-      printf(_("\tExecuting builtin `%s' [%s]\n"),s,session->CurrentStatus());
-      xfree(s);
+      xstring_ca s(args->Combine());
+      printf(_("\tExecuting builtin `%s' [%s]\n"),s.get(),session->CurrentStatus());
       return;
    }
    if(queue_feeder)
@@ -771,7 +737,7 @@ void CmdExec::PrintStatus(int v,const char *prefix)
       }
       return;
    }
-   if(next_cmd && next_cmd[0])
+   if(cmd_buf.Size()>0)
    {
       // xgettext:c-format
       printf(_("\tRunning\n"));
@@ -788,19 +754,14 @@ CmdExec::CmdExec(FileAccess *f,LocalDirectory *c) : SessionJob(f?f:new DummyProt
    next=chain;
    chain=this;
 
-   cmd=0;
-   args=0;
-   output=0;
    background=false;
 
    interactive=false;
    top_level=false;
-   status_line=0;
    feeder=0;
    feeder_called=false;
    used_aliases=0;
 
-   next_cmd=cmd_buf=0;
    partial_cmd=false;
    alias_field=0;
    default_output=0;
@@ -820,11 +781,6 @@ CmdExec::CmdExec(FileAccess *f,LocalDirectory *c) : SessionJob(f?f:new DummyProt
    verify_path_cached=false;
 
    start_time=0;
-   old_lcwd=0;
-   slot=0;
-
-   glob=0;
-   args_glob=0;
 
    redirections=0;
 
@@ -851,23 +807,11 @@ CmdExec::~CmdExec()
    }
 
    free_used_aliases();
-   xfree(cmd);
-   delete args;
-   delete output;
-   xfree(cmd_buf);
-   delete cwd;
    if(cwd_owner==this)
       cwd_owner=0;
-   xfree(old_lcwd);
-   delete glob;
-   delete args_glob;
-
-   Reuse(saved_session);
-
-   Delete(status_line);
 }
 
-char *CmdExec::FormatPrompt(const char *scan)
+const char *CmdExec::FormatPrompt(const char *scan)
 {
    const char *cwd=session->GetCwd();
    if(cwd==0 || cwd[0]==0)
@@ -877,8 +821,7 @@ char *CmdExec::FormatPrompt(const char *scan)
       if(home && strcmp(home,"/") && !strncmp(cwd,home,strlen(home))
 	    && (cwd[strlen(home)]=='/' || cwd[strlen(home)]==0))
       {
-	 static char *cwdbuf=0;
-	 cwdbuf=(char*)xrealloc(cwdbuf,strlen(cwd)-strlen(home)+2);
+	 char *cwdbuf=string_alloca(strlen(cwd)-strlen(home)+2);
 	 sprintf(cwdbuf,"~%s",cwd+strlen(home));
 	 cwd=cwdbuf;
       }
@@ -891,14 +834,8 @@ char *CmdExec::FormatPrompt(const char *scan)
    if(p && p>cwdb)
       cwdb=p+1;
 
-   char StartIgn[3], EndIgn[3];
-   /* bash adds the extra \001 too, don't know why */
-   StartIgn[0] = '\001';
-   StartIgn[1] = RL_PROMPT_START_IGNORE;
-   StartIgn[2] = 0;
-   EndIgn[0] = '\001';
-   EndIgn[1] = RL_PROMPT_END_IGNORE;
-   EndIgn[2] = 0;
+   static const char StartIgn[]={RL_PROMPT_START_IGNORE,0};
+   static const char EndIgn[]={RL_PROMPT_END_IGNORE,0};
 
    subst_t subst[] = {
       { 'a', "\007" },
@@ -912,21 +849,20 @@ char *CmdExec::FormatPrompt(const char *scan)
  // @ if non-default user
       { '@', session->GetUser()?"@":"" },
       { 'U', session->GetConnectURL() },
-      { 'S', slot?slot:"" },
+      { 'S', slot?slot.get():"" },
       { 'w', cwd },
       { 'W', cwdb },
       { '[', StartIgn },
       { ']', EndIgn },
       { 0, "" }
    };
-   static char *prompt=0;
-   xfree(prompt);
-   prompt = Subst(scan, subst);
+   static xstring_c prompt;
+   prompt.set_allocated(Subst(scan, subst));
 
    return(prompt);
 }
 
-char *CmdExec::MakePrompt()
+const char *CmdExec::MakePrompt()
 {
    if(partial_cmd)
    {
@@ -983,10 +919,9 @@ void CmdExec::top_vfprintf(FILE *file,const char *f,va_list v)
 void CmdExec::SetCmdFeeder(CmdFeeder *new_feeder)
 {
    new_feeder->prev=feeder;
-   new_feeder->saved_buf=xstrdup(next_cmd);
-   xfree(cmd_buf);
-   cmd_buf=next_cmd=0;
+   new_feeder->saved_buf.set(cmd_buf.Get());
    feeder=new_feeder;
+   cmd_buf.Empty();
 }
 
 int CmdExec::AcceptSig(int sig)
@@ -1005,9 +940,7 @@ int CmdExec::AcceptSig(int sig)
 	 RevertToSavedSession();
 	 break;
       case(BUILTIN_GLOB):
-	 delete glob;
 	 glob=0;
-	 delete args_glob;
 	 args_glob=0;
 	 break;
       case(BUILTIN_NONE):
@@ -1071,8 +1004,9 @@ void CmdExec::SetInteractive(bool i)
    interactive=i;
 }
 
-void CmdExec::unquote(char *buf,const char *str)
+int CmdExec::unquote(char *buf,const char *str)
 {
+   char *buf0=buf;
    while(*str)
    {
       if(*str=='"' || *str=='\\')
@@ -1080,17 +1014,14 @@ void CmdExec::unquote(char *buf,const char *str)
       *buf++=*str++;
    }
    *buf=0;
+   return buf-buf0;
 }
 
-char *CmdExec::unquote(const char *str)
+const char *CmdExec::unquote(const char *str)
 {
-   static char *ret = 0;
-   if(ret)
-      xfree(ret);
-   ret=(char *)xmalloc(strlen(str)*2+1);
-
-   unquote(ret, str);
-
+   static xstring ret;
+   ret.get_space(strlen(str)*2);
+   ret.set_length(unquote(ret.get_non_const(),str));
    return ret;
 }
 
@@ -1119,41 +1050,20 @@ void CmdExec::FeedQuoted(const char *c)
 // implementation is here because it depends on CmdExec.
 char *ArgV::CombineQuoted(int start) const
 {
-   int	 i;
-   char  *res;
-   char	 *store;
-   const char *arg;
-   int	 len=0;
-
-   for(i=start; i<Count(); i++)
-      len+=strlen(String(i))*2+3;
-
-   if(len==0)
-      return(xstrdup(""));
-
-   res=(char*)xmalloc(len);
-
-   store=res;
-   for(i=start; i<Count(); i++)
+   xstring res("");
+   if(start>=Count())
+      return res.borrow();
+   for(;;)
    {
-      arg=String(i);
+      const char *arg=String(start++);
       if(CmdExec::needs_quotation(arg))
-      {
-	 *store++='"';
-	 CmdExec::unquote(store,arg);
-	 store+=strlen(store);
-	 *store++='"';
-      }
+	 res.vappend("\"",CmdExec::unquote(arg),"\"",NULL);
       else
-      {
-	 strcpy(store,arg);
-	 store+=strlen(store);
-      }
-      *store++=' ';
+	 res.append(arg);
+      if(start>=Count())
+	 return(res.borrow());
+      res.append(' ');
    }
-   store[-1]=0;
-
-   return(res);
 }
 
 const char *CmdExec::GetFullCommandName(const char *cmd)
@@ -1169,59 +1079,45 @@ void CmdExec::AtExit()
 {
    FeedCmd(res_at_exit.Query(0));
    FeedCmd("\n");
-
-   if(status_line)
-   {
-      /* Clear the title, and ensure we don't write anything else
-       * to it in case we're being backgrounded. */
-      Delete(status_line);
-      status_line = 0;
-   }
+   /* Clear the title, and ensure we don't write anything else
+    * to it in case we're being backgrounded. */
+   status_line=0;
 }
 
 void CmdExec::EmptyCmds()
 {
-   xfree(cmd_buf);
-   next_cmd=cmd_buf=0;
+   cmd_buf.Empty();
 }
 
 bool CmdExec::WriteCmds(int fd) const
 {
-   // FIXME: handle short writes.
-
-   if(next_cmd)
-      return write(fd, next_cmd, strlen(next_cmd)) != -1;
-
-   return true;
+   const char *buf;
+   int len;
+   cmd_buf.Get(&buf,&len);
+   for(;;)
+   {
+      if(len==0)
+	 return true;
+      int res=write(fd,buf,len);
+      if(res<=0)
+	 return false;
+      buf+=res;
+      len-=res;
+   }
 }
 
 bool CmdExec::ReadCmds(int fd)
 {
-   // FIXME: use loop over read until eof.
-
-   int fpos = lseek(fd, 0, SEEK_CUR);
-   if(fpos == -1) return false;
-
-   int size = lseek(fd, 0, SEEK_END) - fpos;
-
-   /* restore the file pos */
-   lseek(fd, fpos, SEEK_SET);
-
-   /* realloc */
-   int len=0;
-   if(next_cmd) {
-      len=strlen(next_cmd);
-      memmove(cmd_buf,next_cmd,len);
+   for(;;)
+   {
+      int size=0x1000;
+      size=read(fd,cmd_buf.GetSpace(size),size);
+      if(size==-1)
+	 return false;
+      if(size==0)
+	 return true;
+      cmd_buf.SpaceAdd(size);
    }
-   cmd_buf=next_cmd=(char*)xrealloc(cmd_buf,size+len+1);
-
-   if(read(fd, cmd_buf+len, size) == -1) {
-      cmd_buf[len] = 0; /* leave it in a sane state */
-      return false;
-   }
-
-   cmd_buf[len+size] = 0;
-   return true;
 }
 
 void CmdExec::free_used_aliases()
@@ -1234,25 +1130,27 @@ void CmdExec::free_used_aliases()
    alias_field=0;
 }
 
+void CmdExec::skip_cmd(int len)
+{
+   cmd_buf.Skip(len);
+   alias_field-=len;
+   if(alias_field<=0)
+      free_used_aliases();
+}
 
-CmdExec::cmd_rec *CmdExec::dyn_cmd_table=0;
-int CmdExec::dyn_cmd_table_count=0;
+xarray<CmdExec::cmd_rec> CmdExec::dyn_cmd_table;
 void CmdExec::RegisterCommand(const char *name,cmd_creator_t creator,const char *short_desc,const char *long_desc)
 {
    if(dyn_cmd_table==0)
    {
-      dyn_cmd_table_count=2;
+      int count=0;
       for(const cmd_rec *c=static_cmd_table; c->name; c++)
-	 dyn_cmd_table_count++;
-      dyn_cmd_table=(cmd_rec*)xmemdup(static_cmd_table,dyn_cmd_table_count*sizeof(cmd_rec));
+        count++;
+      dyn_cmd_table.nset(static_cmd_table,count);
    }
-   else
-   {
-      dyn_cmd_table_count++;
-      dyn_cmd_table=(cmd_rec*)xrealloc(dyn_cmd_table,dyn_cmd_table_count*sizeof(cmd_rec));
-   }
-   for(cmd_rec *c=dyn_cmd_table; c->name; c++)
+   for(int i=0; i<dyn_cmd_table.count(); i++)
    {
+      cmd_rec *const c=&dyn_cmd_table[i];
       if(!strcmp(c->name,name))
       {
 	 c->creator=creator;
@@ -1260,20 +1158,15 @@ void CmdExec::RegisterCommand(const char *name,cmd_creator_t creator,const char
 	    c->short_desc=short_desc;
 	 if(long_desc)
 	    c->long_desc=long_desc;
-	 dyn_cmd_table_count--;
 	 return;
       }
    }
-   dyn_cmd_table[dyn_cmd_table_count-2].name=name;
-   dyn_cmd_table[dyn_cmd_table_count-2].creator=creator;
-   dyn_cmd_table[dyn_cmd_table_count-2].short_desc=short_desc;
-   dyn_cmd_table[dyn_cmd_table_count-2].long_desc=long_desc;
-   memset(&dyn_cmd_table[dyn_cmd_table_count-1],0,sizeof(cmd_rec));
+   cmd_rec new_entry={name,creator,short_desc,long_desc};
+   dyn_cmd_table.append(new_entry);
 }
 
 void CmdExec::ChangeSession(FileAccess *new_session)
 {
-   Reuse(session);
    session=new_session;
    session->SetPriority(fg?1:0);
    Reconfig(0);
@@ -1281,9 +1174,15 @@ void CmdExec::ChangeSession(FileAccess *new_session)
       ConnectionSlot::Set(slot,session);
 }
 
-const CmdExec::cmd_rec *CmdExec::CmdByIndex(int i)
+const char *CmdExec::CmdByIndex(int i)
 {
-   return (dyn_cmd_table?dyn_cmd_table:static_cmd_table)+i;
+   if(dyn_cmd_table)
+   {
+      if(i<dyn_cmd_table.count())
+	 return dyn_cmd_table[i].name;
+      return 0;
+   }
+   return static_cmd_table[i].name;
 }
 
 Job *CmdExec::default_cmd()
@@ -1307,22 +1206,21 @@ Job *CmdExec::default_cmd()
 
 void CmdExec::FeedArgV(const ArgV *args,int start)
 {
-   char *cmd;
+   xstring_c cmd;
 
    if(start+1==args->count())
-      cmd=args->Combine(start);
+      cmd.set_allocated(args->Combine(start));
    else
-      cmd=args->CombineQuoted(start);
+      cmd.set_allocated(args->CombineQuoted(start));
 
    FeedCmd(cmd);
    FeedCmd("\n");
-   xfree(cmd);
 }
 
 bool CmdExec::SameQueueParameters(CmdExec *scan,const char *this_url)
 {
    return !strcmp(this_url,scan->session->GetConnectURL(FA::NO_PATH))
-      && !xstrcmp(this->slot,scan->slot);
+      && this->slot.eq(scan->slot);
 }
 
 /* return the CmdExec containing a queue feeder; create if necessary */
@@ -1340,12 +1238,12 @@ CmdExec  *CmdExec::GetQueue(bool create)
       return NULL;
 
    CmdExec *queue=new CmdExec(session->Clone(),cwd->Clone());
-   queue->slot=xstrdup(slot);
+   queue->slot.set(slot);
 
    queue->SetParentFg(this,false);
    queue->AllocJobno();
    const char *url=session->GetConnectURL(FA::NO_PATH);
-   queue->cmdline=xasprintf("queue (%s%s%s)",url,slot?"; ":"",slot?slot:"");
+   queue->cmdline.vset("queue (",url,slot?"; ":"",slot?slot.get():"",")",NULL);
    queue->queue_feeder=new QueueFeeder(session->GetCwd(), cwd->GetName());
    queue->SetCmdFeeder(queue->queue_feeder);
    queue->Reconfig(0);
