@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: DummyProto.cc,v 1.5 2001/10/03 08:13:36 lav Exp $ */
+/* $Id: DummyProto.cc,v 1.9 2007/04/23 07:11:40 lav Exp $ */
 
 #include <config.h>
 
@@ -26,66 +26,40 @@
 
 int DummyProto::Do() { return STALL; }
 int DummyProto::Done() { return NO_HOST; }
-const char *DummyProto::GetProto() { return ""; }
-FileAccess *DummyProto::Clone() { return new DummyProto; }
+const char *DummyProto::GetProto() const { return ""; }
+FileAccess *DummyProto::Clone() const { return new DummyProto; }
 int DummyProto::Read(void *buf,int size) { return NO_HOST; };
 int DummyProto::Write(const void *buf,int size) { return NO_HOST; };
 int DummyProto::StoreStatus() { return NO_HOST; }
 
 class DummyDirList : public DirList
 {
-   FA *p;
 public:
-   DummyDirList(FA *p1,ArgV *a) : DirList(a) { p=p1; }
-   int Do() { SetError(p->StrError(FA::NO_HOST)); return STALL; }
+   DummyDirList(FA *p1,ArgV *a) : DirList(p1,a) {}
+   int Do() { SetError(session->StrError(FA::NO_HOST)); return STALL; }
    const char *Status() { return ""; }
 };
 class DummyListInfo : public ListInfo
 {
-   FA *p;
 public:
-   DummyListInfo(FA *p1) : ListInfo(), p(p1) { }
-   int Do() { SetError(p->StrError(FA::NO_HOST)); return STALL; }
+   DummyListInfo(FA *p1) : ListInfo(p1,0) {}
+   int Do() { SetError(session->StrError(FA::NO_HOST)); return STALL; }
    const char *Status() { return ""; }
 };
 DirList *DummyProto::MakeDirList(ArgV *a)
 {
    return new DummyDirList(this,a);
 }
-ListInfo *DummyProto::MakeListInfo()
+ListInfo *DummyProto::MakeListInfo(const char *path)
 {
    return new DummyListInfo(this);
 }
 
-
-DummyNoProto::DummyNoProto(const char *p)
-{
-   proto=xstrdup(p);
-}
-DummyNoProto::~DummyNoProto()
-{
-   xfree(proto);
-}
-int DummyNoProto::Do() { return STALL; }
-int DummyNoProto::Done() { return NO_HOST; }
-const char *DummyNoProto::GetProto() { return proto; }
-FileAccess *DummyNoProto::Clone() { return new DummyNoProto(proto); }
-int DummyNoProto::Read(void *buf,int size) { return NO_HOST; };
-int DummyNoProto::Write(const void *buf,int size) { return NO_HOST; };
-int DummyNoProto::StoreStatus() { return NO_HOST; }
+const char *DummyNoProto::GetProto() const { return proto; }
+FileAccess *DummyNoProto::Clone() const { return new DummyNoProto(proto); }
 const char *DummyNoProto::StrError(int)
 {
-   static char str[128];
-   sprintf(str,"%.32s%s",proto,_(" - not supported protocol"));
+   static xstring str;
+   str.vset(proto.get(),_(" - not supported protocol"),NULL);
    return str;
 }
-
-DirList *DummyNoProto::MakeDirList(ArgV *a)
-{
-   return new DummyDirList(this,a);
-}
-ListInfo *DummyNoProto::MakeListInfo()
-{
-   return new DummyListInfo(this);
-}
-	
