@@ -18,12 +18,11 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FileAccess.cc,v 1.133 2006/07/04 05:15:35 lav Exp $ */
+/* $Id: FileAccess.cc,v 1.151 2007/11/03 15:08:13 lav Exp $ */
 
 #include <config.h>
 
 #include "FileAccess.h"
-#include "xmalloc.h"
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
@@ -41,26 +40,20 @@
 #include "ArgV.h"
 #include "ConnectionSlot.h"
 #include "SignalHook.h"
+#include "FileGlob.h"
 #ifdef WITH_MODULES
 # include "module.h"
 #endif
 
 FileAccess *FileAccess::chain=0;
+const FileAccessRef FileAccessRef::null;
 
 void FileAccess::Init()
 {
    ClassInit();
 
-   vproto=0;
-   portname=0;
-   hostname=0;
-   user=pass=0;
    pass_open=false;
 
-   file=0;
-   file_url=0;
-   file1=0;
-   real_cwd=0;
    default_cwd="~";
    cwd.Set(default_cwd,false,0);
    new_cwd=0;
@@ -72,22 +65,16 @@ void FileAccess::Init()
    retries=0;
    opt_date=0;
    opt_size=0;
-   error=0;
    error_code=OK;
    saved_errno=0;
    mkdir_p=false;
    ascii=false;
    norest_manual=false;
-   location=0;
-   suggested_filename=0;
-   entity_content_type=0;
-   entity_charset=0;
    array_ptr=array_cnt=0;
 
    entity_size=NO_SIZE;
    entity_date=NO_DATE;
 
-   closure=0;
    res_prefix=0;
 
    chmod_mode=0644;
@@ -104,55 +91,58 @@ FileAccess::FileAccess(const FileAccess *fa)
 
    cwd=fa->cwd;
    home=fa->home;
-   xfree(user);
-   user=xstrdup(fa->user);
-   xfree(pass);
-   pass=xstrdup(fa->pass);
+   user.set(fa->user);
+   pass.set(fa->pass);
    pass_open=fa->pass_open;
-   xfree(hostname);
-   hostname=xstrdup(fa->hostname);
-   xfree(portname);
-   portname=xstrdup(fa->portname);
-   vproto=xstrdup(fa->vproto);
+   hostname.set(fa->hostname);
+   portname.set(fa->portname);
+   vproto.set(fa->vproto);
 }
 
 FileAccess::~FileAccess()
 {
-   xfree(vproto);
-   xfree(file);
-   xfree(file_url);
-   xfree(real_cwd);
-   xfree(error);
-   xfree(user);
-   xfree(pass);
-   xfree(hostname);
-   xfree(portname);
-   xfree(closure);
-   xfree(location);
-   xfree(entity_content_type);
-   xfree(entity_charset);
-   for(FileAccess **scan=&chain; *scan; scan=&((*scan)->next))
-   {
-      if(*scan==this)
-      {
-	 *scan=next;
-	 break;
-      }
-   }
+   ListDel(FileAccess,chain,this,next);
 }
 
-void  FileAccess::DebugPrint(const char *prefix,const char *str,int level)
+void  FileAccess::Log2(int level,const char *str0)
 {
-   if(!Log::global)
-      return;
-   int len=strlen(str);
-   if(len>0 && str[len-1]=='\n')
-      len--;
-   if(len>0 && str[len-1]=='\r')
-      len--;
-   char *buf=string_alloca(strlen(prefix)+len+2);
-   sprintf(buf,"%s%.*s\n",prefix,len,str);
-   Log::global->Write(level,buf);
+   xstring &str=xstring::get_tmp();
+   str.set(str0);
+   str.chomp('\n');
+   str.chomp('\r');
+   str.append('\n');
+   Log::global->Write(level,str);
+}
+void  FileAccess::Log3(int level,const char *prefix,const char *str)
+{
+   Log::global->Write(level,prefix);
+   Log2(level,str);
+}
+void FileAccess::LogError(int level,const char *fmt,...)
+{
+   va_list v;
+   va_start(v,fmt);
+   Log::global->Write(level,"**** ");
+   Log::global->vFormat(level,fmt,v);
+   Log::global->Write(level,"\n");
+   va_end(v);
+}
+void FileAccess::LogNote(int level,const char *fmt,...)
+{
+   va_list v;
+   va_start(v,fmt);
+   Log::global->Write(level,"---- ");
+   Log::global->vFormat(level,fmt,v);
+   Log::global->Write(level,"\n");
+   va_end(v);
+}
+void FileAccess::LogRecv(int level,const char *line)
+{
+   Log3(level,"<--- ",line);
+}
+void FileAccess::LogSend(int level,const char *line)
+{
+   Log3(level,"---> ",line);
 }
 
 void FileAccess::NonBlock(int fd)
@@ -174,7 +164,7 @@ void  FileAccess::Open(const char *fn,int mode,off_t offs)
    if(IsOpen())
       Close();
    Resume();
-   file=xstrdup(fn);
+   file.set(fn);
    real_pos=UNKNOWN_POS;
    pos=offs;
    this->mode=mode;
@@ -200,14 +190,7 @@ void  FileAccess::Open(const char *fn,int mode,off_t offs)
 
 const char *FileAccess::StrError(int err)
 {
-   static char *str=0;
-   static unsigned str_allocated=0;
-
-   if(error)
-   {
-      if(str_allocated<128+strlen(error))
-	 str=(char*)xrealloc(str,str_allocated=128+strlen(error));
-   }
+   static xstring str;
 
    // note to translators: several errors should not be displayed to user;
    // so no need to translate them.
@@ -219,10 +202,7 @@ const char *FileAccess::StrError(int err)
       return("Error 0");
    case(SEE_ERRNO):
       if(error)
-      {
-   	 sprintf(str,"%s: %s",error,strerror(saved_errno));
-	 return(str);
-      }
+   	 return str.vset(error.get(),": ",strerror(saved_errno),NULL);
       return(strerror(saved_errno));
    case(LOOKUP_ERROR):
       return(error);
@@ -230,57 +210,38 @@ const char *FileAccess::StrError(int err)
       return("Class is not Open()ed");
    case(NO_FILE):
       if(error)
-      {
-   	 sprintf(str,_("Access failed: %s"),error);
-	 return(str);
-      }
+   	 return str.vset(_("Access failed: "),error.get(),NULL);
       return(_("File cannot be accessed"));
    case(NO_HOST):
       return(_("Not connected"));
    case(FATAL):
       if(error)
-      {
-   	 sprintf(str,_("Fatal error: %s"),error);
-	 return(str);
-      }
+   	 return str.vset(_("Fatal error"),": ",error.get(),NULL);
       return(_("Fatal error"));
    case(STORE_FAILED):
       return(_("Store failed - you have to reput"));
    case(LOGIN_FAILED):
       if(error)
-      {
-   	 sprintf(str,"%s: %s",_("Login failed"),error);
-	 return(str);
-      }
+   	 return str.vset(_("Login failed"),": ",error.get(),NULL);
       return(_("Login failed"));
    case(NOT_SUPP):
       if(error)
-      {
-   	 sprintf(str,"%s: %s",_("Operation not supported"),error);
-	 return(str);
-      }
+   	 return str.vset(_("Operation not supported"),": ",error.get(),NULL);
       return(_("Operation not supported"));
    case(FILE_MOVED):
       if(error)
-      {
-   	 sprintf(str,_("File moved: %s"),error);
-      }
+   	 return str.vset(_("File moved"),": ",error.get(),NULL);
       else
-      {
-	 if(str_allocated<64+xstrlen(location))
-	    str=(char*)xrealloc(str,64+xstrlen(location));
-	 sprintf(str,_("File moved to `%s'"),location?location:"?");
-      }
-      return str;
+   	 return str.vset(_("File moved to `"),location?location.get():"?","'",NULL);
    }
    return("");
 }
 
 void FileAccess::Close()
 {
-   xfree(file); file=0;
-   xfree(file_url); file_url=0;
-   xfree(file1); file1=0;
+   file.set(0);
+   file_url.set(0);
+   file1.set(0);
    delete new_cwd; new_cwd=0;
    mode=CLOSED;
    opt_date=0;
@@ -291,18 +252,16 @@ void FileAccess::Close()
    entity_date=NO_DATE;
    ascii=false;
    norest_manual=false;
-   xfree(location); location=0;
-   xfree(entity_content_type);
-   entity_content_type=0;
-   xfree(entity_charset);
-   entity_charset=0;
+   location.set(0);
+   entity_content_type.set(0);
+   entity_charset.set(0);
    ClearError();
 }
 
 void FileAccess::Rename(const char *f,const char *f1)
 {
    Close();
-   file1=xstrdup(f1);
+   file1.set(f1);
    Open(f,RENAME);
 
    cache->TreeChanged(this,file);
@@ -327,7 +286,7 @@ bool FileAccess::SameSiteAs(const FileAccess *fa) const
 
 const char *FileAccess::GetFileURL(const char *f,int flags) const
 {
-   static char *url;
+   static xstring url;
 
    const char *proto=GetVisualProto();
    if(proto[0]==0)
@@ -335,13 +294,13 @@ const char *FileAccess::GetFileURL(const char *f,int flags) const
 
    ParsedURL u("");
 
-   u.proto=(char*)proto;
+   u.proto.set(proto);
    if(!(flags&NO_USER))
-      u.user=user;
+      u.user.set(user);
    if((pass_open || (flags&WITH_PASSWORD)) && !(flags&NO_PASSWORD))
-      u.pass=pass;
-   u.host=hostname;
-   u.port=portname;
+      u.pass.set(pass);
+   u.host.set(hostname);
+   u.port.set(portname);
    if(!(flags&NO_PATH))
    {
       if(cwd.url)
@@ -349,23 +308,24 @@ const char *FileAccess::GetFileURL(const char *f,int flags) const
 	 Path f_path(cwd);
 	 if(f)
 	    f_path.Change(f,true);
-	 int f_path_index=url::path_index(f_path.url);
-	 xfree(url);
-	 url=u.Combine(home);
-	 url=(char*)xrealloc(url,strlen(url)+strlen(f_path.url+f_path_index)+2);
-	 strcpy(url+strlen(url),f_path.url+f_path_index);
-	 return url;
+	 if(f_path.url)
+	 {
+	    int f_path_index=url::path_index(f_path.url);
+	    url.set_allocated(u.Combine(home));
+	    url.append(f_path.url+f_path_index);
+	    return url;
+	 }
       }
 
+      bool is_dir=((!f || !*f) && !cwd.is_file);
+
       if(!f || (f[0]!='/' && f[0]!='~'))
-	 f=dir_file(cwd.path?cwd.path:"~",f);
-      u.path=(char*)f;
+	 f=dir_file(cwd.path?cwd.path.get():"~",f);
+      u.path.set(f);
+      if(is_dir && url::dir_needs_trailing_slash(proto))
+	 u.path.append('/');
    }
-
-   xfree(url);
-   url=u.Combine(home);
-
-   return url;
+   return url.set_allocated(u.Combine(home));
 }
 
 const char *FileAccess::GetConnectURL(int flags) const
@@ -376,10 +336,8 @@ const char *FileAccess::GetConnectURL(int flags) const
 void FileAccess::Connect(const char *host1,const char *port1)
 {
    Close();
-   xfree(hostname);
-   hostname=xstrdup(host1);
-   xfree(portname);
-   portname=xstrdup(port1);
+   hostname.set(host1);
+   portname.set(port1);
    DontSleep();
    ResetLocationData();
 }
@@ -387,29 +345,26 @@ void FileAccess::Connect(const char *host1,const char *port1)
 void FileAccess::Login(const char *user1,const char *pass1)
 {
    Close();
-   xfree(user);
-   user=xstrdup(user1);
-   xfree(pass);
-   pass=xstrdup(pass1);
+   user.set(user1);
+   pass.set(pass1);
    pass_open=false;
 
    if(user && pass==0)
    {
-      for(FileAccess *o=chain; o!=0; o=o->next)
+      FileAccess *o;
+      for(o=chain; o!=0; o=o->next)
       {
-	 pass=o->pass;
+	 pass.set(o->pass);
 	 if(SameSiteAs(o) && o->pass)
-	 {
-	    pass=xstrdup(o->pass);
 	    break;
-	 }
-	 pass=0;
       }
+      if(!o)
+	 pass.set(0);
       if(pass==0 && hostname) // still no pass? Try .netrc
       {
 	 NetRC::Entry *nrc=NetRC::LookupHost(hostname,user);
 	 if(nrc)
-	    pass=xstrdup(nrc->pass);
+	    pass.set(nrc->pass);
       }
    }
    ResetLocationData();
@@ -418,8 +373,8 @@ void FileAccess::Login(const char *user1,const char *pass1)
 void FileAccess::AnonymousLogin()
 {
    Close();
-   xfree(user); user=0;
-   xfree(pass); pass=0;
+   user.set(0);
+   pass.set(0);
    pass_open=false;
    ResetLocationData();
 }
@@ -432,21 +387,16 @@ void FileAccess::ResetLocationData()
 
 void FileAccess::SetPasswordGlobal(const char *p)
 {
-   xfree(pass);
-   pass=xstrdup(p);
+   pass.set(p);
+   xstring save_pass;
    for(FileAccess *o=chain; o!=0; o=o->next)
    {
       if(o==this)
 	 continue;
-      char *save_pass=o->pass;	 // cheat SameSiteAs.
-      o->pass=pass;
-      if(SameSiteAs(o))
-      {
-	 xfree(save_pass);
-	 o->pass=xstrdup(pass);
-      }
-      else
-	 o->pass=save_pass;
+      save_pass.set(o->pass);	 // cheat SameSiteAs.
+      o->pass.set(pass);
+      if(!SameSiteAs(o))
+	 o->pass.set(save_pass);
    }
 }
 
@@ -458,21 +408,14 @@ void FileAccess::GetInfoArray(struct fileinfo *info,int count)
    array_cnt=count;
 }
 
-static void expand_tilde(char **path, const char *home)
+static void expand_tilde(xstring &path, const char *home, int i=0)
 {
-   if((*path)[0]=='~' && ((*path)[1]==0 || (*path)[1]=='/'))
-   {
-      int home_len=strlen(home);
-      if(home_len==1 && home[0]=='/' && (*path)[1]=='/')
-	 home_len=0;
-      int path_len=strlen(*path+1);
-      // Note: don't shrink the array, because later memmove would break
-      // We could first memmove then shrink, but it won't give much
-      if(home_len>1)
-	 *path=(char*)xrealloc(*path,path_len+home_len+1);
-      memmove(*path+home_len,*path+1,path_len+1);
-      memmove(*path,home,home_len);
-   }
+   if(!(path[i]=='~' && (path[i+1]==0 || path[i+1]=='/')))
+      return;
+   char prefix_len=(last_char(home)=='/' ? 2 : 1);
+   if(home[0]=='/' && i>0 && path[i-1]=='/')
+      home++;
+   path.set_substr(i,prefix_len,home);
 }
 
 void  FileAccess::ExpandTildeInCWD()
@@ -483,11 +426,11 @@ void  FileAccess::ExpandTildeInCWD()
       if(new_cwd)
 	 new_cwd->ExpandTilde(home);
       if(real_cwd)
-	 expand_tilde(&real_cwd,home);
+	 expand_tilde(real_cwd,home);
       if(file)
-	 expand_tilde(&file,home);
+	 expand_tilde(file,home);
       if(file1)
-	 expand_tilde(&file1,home);
+	 expand_tilde(file1,home);
    }
 }
 
@@ -496,19 +439,14 @@ const char *FileAccess::ExpandTildeStatic(const char *s)
    if(!home || !(s[0]=='~' && (s[1]=='/' || s[1]==0)))
       return s;
 
-   static char *buf=0;
-   static int buf_len=0;
-
-   int len=strlen(s)+1;
-   if(len>buf_len)
-      buf=(char*)xrealloc(buf,buf_len=len);
-   strcpy(buf,s);
-   expand_tilde(&buf,home);
+   static xstring buf;
+   buf.set(s);
+   expand_tilde(buf,home);
    return buf;
 }
 
 static inline
-int last_element_is_doubledot(char *path,char *end)
+bool last_element_is_doubledot(const char *path,const char *end)
 {
    return((end==path+2 && !strncmp(path,"..",2))
         || (end>path+2 && !strncmp(end-3,"/..",3)));
@@ -527,11 +465,17 @@ int FileAccess::device_prefix_len(const char *path)
    return 0;
 }
 
-void FileAccess::Path::Optimize(char *path,int device_prefix_len)
+void FileAccess::Path::Optimize(xstring& path,int device_prefix_len)
 {
-   int	 prefix_size=0;
+   int prefix_size=0;
 
-   if(path[0]=='/')
+   if(path[0]=='/' && path[1]=='~' && device_prefix_len==1)
+   {
+      prefix_size=2;
+      while(path[prefix_size]!='/' && path[prefix_size]!='\0')
+	 prefix_size++;
+   }
+   else if(path[0]=='/')
    {
       prefix_size=1;
       if(path[1]=='/' && (!path[2] || path[2]!='/'))
@@ -552,7 +496,7 @@ void FileAccess::Path::Optimize(char *path,int device_prefix_len)
    char	 *in;
    char	 *out;
 
-   in=out=path+prefix_size;
+   in=out=path.get_non_const()+prefix_size;
 
    while((in[0]=='.' && (in[1]=='/' || in[1]==0))
    || (in>path && in[-1]=='/' && (in[0]=='/'
@@ -565,12 +509,12 @@ void FileAccess::Path::Optimize(char *path,int device_prefix_len)
 	 in++;
    }
 
-   for(;;)
+   while(*in)
    {
       if(in[0]=='/')
       {
 	 // double slash
-	 if(in[1]=='/') // || (strip_trailing_slash && in[1]=='\0'))
+	 if(in[1]=='/')
 	 {
 	    in++;
 	    continue;
@@ -611,9 +555,9 @@ void FileAccess::Path::Optimize(char *path,int device_prefix_len)
 	    continue;
 	 }
       }
-      if((*out++=*in++)=='\0')
-	 break;
+      *out++=*in++;
    }
+   path.truncate(path.length()-(in-out));
 }
 
 void FileAccess::Chdir(const char *path,bool verify)
@@ -652,21 +596,16 @@ void FileAccess::SetError(int ec,const char *e)
    if(ec==SEE_ERRNO)
       saved_errno=errno;
    if(ec==NO_FILE && file && file[0] && !strstr(e,file))
-   {
-      char *m=string_alloca(strlen(e)+2+strlen(file)+2);
-      sprintf(m,"%s (%s)",e,file);
-      e=m;
-   }
-   xfree(error);
-   error=xstrdup(e);
+      error.vset(e," (",file.get(),")",NULL);
+   else
+      error.set(e);
    error_code=ec;
 }
 
 void FileAccess::ClearError()
 {
    error_code=OK;
-   xfree(error);
-   error=0;
+   error.set(0);
 }
 
 void FileAccess::Fatal(const char *e)
@@ -676,8 +615,7 @@ void FileAccess::Fatal(const char *e)
 
 void FileAccess::SetSuggestedFileName(const char *fn)
 {
-   xfree(suggested_filename);
-   suggested_filename=0;
+   suggested_filename.set(0);
    if(fn==0)
       return;
 
@@ -692,13 +630,12 @@ void FileAccess::SetSuggestedFileName(const char *fn)
    }
    if(!*fn)
       return;
-   suggested_filename=xstrdup(fn);
+   suggested_filename.set(fn);
 }
 
 void FileAccess::SetFileURL(const char *u)
 {
-   xfree(file_url);
-   file_url=xstrdup(u);
+   file_url.set(u);
    if(new_cwd && mode==CHANGE_DIR)
       new_cwd->SetURL(u);
 }
@@ -831,16 +768,6 @@ void FileAccess::SetTryTime(time_t t)
    try_time=t;
 }
 
-ResValue FileAccess::Query(const char *name,const char *closure) const
-{
-   const char *prefix=res_prefix;
-   if(!prefix)
-      prefix=GetProto();
-   char *fullname=(char*)alloca(3+strlen(prefix)+1+strlen(name)+1);
-   sprintf(fullname,"%s:%s",prefix,name);
-   return ResMgr::Query(fullname,closure);
-}
-
 bool FileAccess::IsBetterThan(const FileAccess *fa) const
 {
    return(SameProtoAs(fa) && this->IsConnected() > fa->IsConnected());
@@ -858,15 +785,8 @@ bool FileAccess::NeedSizeDateBeforehand() { return false; }
 void FileAccess::Cleanup() {}
 void FileAccess::CleanupThis() {}
 ListInfo *FileAccess::MakeListInfo(const char *path) { return 0; }
-Glob *FileAccess::MakeGlob(const char *pattern) { return 0; }
-DirList *FileAccess::MakeDirList(ArgV *a) { if(a) delete a; return 0; }
-
-DirList::~DirList()
-{
-   delete buf;
-   if(args)
-      delete args;
-}
+Glob *FileAccess::MakeGlob(const char *pattern) { return new NoGlob(pattern); }
+DirList *FileAccess::MakeDirList(ArgV *a) { delete a; return 0; }
 
 void FileAccess::CleanupAll()
 {
@@ -896,7 +816,7 @@ FileAccess *FileAccess::New(const char *proto,const char *host,const char *port)
 
    if(!strcmp(proto,"slot"))
    {
-      FA *session=ConnectionSlot::FindSession(host);
+      const FA *session=ConnectionSlot::FindSession(host);
       return session?session->Clone():0;
    }
 
@@ -983,38 +903,29 @@ FileAccess *FileAccess::Protocol::NewSession(const char *proto)
 }
 
 // FileAccessOperation implementation
-FileAccessOperation::FileAccessOperation()
-{
-   done=false;
-   error_text=0;
-   use_cache=true;
-}
-
-FileAccessOperation::~FileAccessOperation()
-{
-   xfree(error_text);
-}
-
 void FileAccessOperation::SetError(const char *e)
 {
-   xfree(error_text);
-   error_text=xstrdup(e);
+   error_text.set(e);
    done=true;
 }
 void FileAccessOperation::SetErrorCached(const char *e)
 {
-   xfree(error_text);
-   error_text=xasprintf(_("%s [cached]"),e);
-   done=true;
+   SetError(e);
+   error_text.append(_(" [cached]"));
 }
 
+DirList::DirList(FileAccess *s,ArgV *a)
+   : FileAccessOperation(s), buf(new Buffer()), args(a), color(false)
+{
+}
+DirList::~DirList()
+{
+}
 
 // ListInfo implementation
 ListInfo::ListInfo(FileAccess *s,const char *p)
+   : FileAccessOperation(s)
 {
-   session=s;
-   result=0;
-
    exclude=0;
    exclude_prefix=0;
 
@@ -1033,37 +944,22 @@ ListInfo::~ListInfo()
 {
    if(session)
       session->Close();
-   if(session && saved_cwd.path)
+   if(session && saved_cwd)
       session->SetCwd(saved_cwd);
-   delete result;
 }
 
-void ListInfo::SetExclude(const char *p,PatternSet *x)
-{
-   exclude=x;
-   exclude_prefix=p;
-}
 
 // Path implementation
 void FileAccess::Path::init()
 {
    device_prefix_len=0;
-   path=0;
    is_file=false;
-   url=0;
-}
-FileAccess::Path::~Path()
-{
-   xfree(path);
-   xfree(url);
 }
 void FileAccess::Path::Set(const char *new_path,bool new_is_file,const char *new_url,int new_device_prefix_len)
 {
-   xfree(path);
-   path=xstrdup(new_path);
+   path.set(new_path);
    is_file=new_is_file;
-   xfree(url);
-   url=xstrdup(new_url);
+   url.set(new_url);
    device_prefix_len=new_device_prefix_len;
 }
 void FileAccess::Path::Set(const Path *o)
@@ -1083,50 +979,36 @@ void FileAccess::Path::Change(const char *new_path,bool new_is_file,const char *
    const char *bn=basename_ptr(new_path);
    if(!strcmp(bn,".") || !strcmp(bn,".."))
       new_is_file=false;
+
+   int path_index=0;
    if(url)
    {
-      int path_index=url::path_index(url);
-      const char *old_url_path=url+path_index;
-      char *new_url_path=string_alloca(strlen(old_url_path)+xstrlen(new_path)*3+2);
-      strcpy(new_url_path,old_url_path);
+      path_index=url::path_index(url);
+      xstring new_url_path(url+path_index);
       if(is_file)
       {
 	 dirname_modify(new_url_path);
 	 if(!new_url_path[0])
-	    strcpy(new_url_path,"/~");
+	    new_url_path.set("/~");
       }
-      if(last_char(new_url_path)!='/')
-	 strcat(new_url_path,"/");
-      if(new_path[0]!='/' && new_path[0]!='~' && new_device_prefix_len==0)
+      if(new_url_path.last_char()!='/')
+	 new_url_path.append("/");
+      if(new_path[0]=='/' || new_path[0]=='~' || new_device_prefix_len!=0)
       {
-	 if(new_path_enc)
-	    strcat(new_url_path,new_path_enc);
-	 else
-	 {
-	    char *np=alloca_strdup(new_path);
-	    Optimize(np);
-	    if(np[0]=='.' && np[1]=='.' && (np[2]=='/' || np[2]==0))
-	    {
-	       xfree(url);
-	       url=0;
-	       goto url_done;
-	    }
-	    else
-	       url::encode_string(np,new_url_path+strlen(new_url_path),URL_PATH_UNSAFE);
-	 }
+	 bool have_slash=((new_path_enc?new_path_enc:new_path)[0]=='/');
+	 new_url_path.set(have_slash?"":"/");
       }
+      if(new_path_enc)
+	 new_url_path.append(new_path_enc);
       else
-      {
-	 if(new_path_enc)
-	    strcpy(new_url_path,new_path_enc);
-	 else
-	    url::encode_string(new_path,new_url_path,URL_PATH_UNSAFE);
-      }
-      Optimize(new_url_path+(!strncmp(new_url_path,"/~",2)));
-      url=(char*)xrealloc(url,path_index+strlen(new_url_path)+1);
-      strcpy(url+path_index,new_url_path);
+	 new_url_path.append(url::encode(new_path,URL_PATH_UNSAFE));
+      if(!new_is_file && url::dir_needs_trailing_slash(url) && new_url_path.last_char()!='/')
+	 new_url_path.append('/');
+      Optimize(new_url_path,!strncmp(new_url_path,"/~",2));
+      url.truncate(path_index);
+      url.append(new_url_path);
    }
-url_done:
+
    if(new_path[0]!='/' && new_path[0]!='~' && new_device_prefix_len==0
    && path && path[0])
    {
@@ -1134,23 +1016,39 @@ url_done:
       {
 	 dirname_modify(path);
 	 if(!path[0])
-	    strcpy(path,"~");
+	    path.set("~");
       }
       char *newcwd=string_alloca(xstrlen(path)+xstrlen(new_path)+2);
       if(last_char(path)=='/')
-	 sprintf(newcwd,"%s%s",path,new_path);
+	 sprintf(newcwd,"%s%s",path.get(),new_path);
       else
-	 sprintf(newcwd,"%s/%s",path,new_path);
+	 sprintf(newcwd,"%s/%s",path.get(),new_path);
       new_path=newcwd;
    }
-   xfree(path);
-   path=xstrdup(new_path);
+   path.set(new_path);
    device_prefix_len=new_device_prefix_len;
    Optimize();
    strip_trailing_slashes(path);
    is_file=new_is_file;
    if(!strcmp(path,"/") || !strcmp(path,"//"))
       is_file=false;
+
+   // sanity check
+   if(url)
+   {
+      const char *url_path=url+path_index;
+      if(url_path[0]=='/' && url_path[1]=='~')
+	 url_path++;
+      url_path=url::decode(url_path);
+      int up_len=strlen(url_path);
+      if(up_len>1 && url_path[up_len-1]=='/')
+	 up_len--;
+      if(strncmp(url_path,path,up_len))
+      {
+	 Log::global->Format(0,"(BUG?) URL mismatch %s vs %s, dropping URL\n",url.get(),path.get());
+	 url.set(0);
+      }
+   }
 }
 bool FileAccess::Path::operator==(const Path &p2) const
 {
@@ -1173,9 +1071,19 @@ void FileAccess::Path::ExpandTilde(const Path &home)
       if(path[1]=='\0')
 	 is_file=home.is_file;
    }
-   expand_tilde(&path,home.path);
+   if(url)
+   {
+      int pi=url::path_index(url);
+      if(url[pi]=='/' && url[pi+1]=='~')
+	 pi++;
+      expand_tilde(url,home.url?home.url.get():url::encode(home.path,URL_PATH_UNSAFE),pi);
+   }
+   expand_tilde(path,home.path);
 }
 
+#include "DirColors.h"
+#include "LocalDir.h"
+#include "FileCopy.h"
 #include "modconfig.h"
 #ifndef MODULE_PROTO_FTP
 # include "ftpclass.h"
@@ -1224,9 +1132,14 @@ void FileAccess::ClassInit()
    _http;
    _fish;
    _sftp;
+
+   // make it link in classes required by modules.
+   LocalDirectory ld;
 }
 void FileAccess::ClassCleanup()
 {
+   DirColors::DeleteInstance();
    delete cache;
    cache=0;
+   FileCopy::fxp_create=0;
 }
