@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FileCopy.cc,v 1.126.2.3 2007/08/23 06:58:18 lav Exp $ */
+/* $Id: FileCopy.cc,v 1.142 2007/09/10 09:57:53 lav Exp $ */
 
 /* FileCopyPeer behaviour:
     1) when suspended, does nothing
@@ -79,7 +79,9 @@ int FileCopy::Do()
       if(remove_target_first && !put->FileRemoved())
 	 return m;
       remove_target_first=false;
-      if(put->NeedSizeDateBeforehand() || (cont && put->CanSeek()))
+      if(cont && put->CanSeek())
+	 put->WantSize();
+      if(put->NeedSizeDateBeforehand() || (cont && put->CanSeek() && put->GetSize()==NO_SIZE_YET))
       {
 	 if(get->GetSize()==NO_SIZE_YET || get->GetDate()==NO_DATE_YET)
 	 {
@@ -155,7 +157,7 @@ int FileCopy::Do()
 	 if(put->GetPos()>0)
 	 {
 	    put->PutEOF();
-	    Roll(put);
+	    put->Roll();
 	 }
       get_error:
 	 SetError(get->ErrorText());
@@ -274,7 +276,11 @@ int FileCopy::Do()
       m=MOVED;
 
       if(get->range_limit!=FILE_END && get->range_limit<get->GetRealPos()+s)
+      {
 	 s=get->range_limit-get->GetRealPos();
+	 if(s<0)
+	    s=0;
+      }
 
       if(line_buffer)
       {
@@ -382,6 +388,7 @@ int FileCopy::Do()
       debug((10,"copy: get is finished - all done\n"));
       set_state(ALL_DONE);
       get->Suspend();
+      LogTransfer();
       return MOVED;
 
    pre_GET_INFO_WAIT:
@@ -400,16 +407,13 @@ int FileCopy::Do()
    return m;
 }
 
-void FileCopy::Init()
+FileCopy::FileCopy(FileCopyPeer *s,FileCopyPeer *d,bool c)
+   : get(s), put(d), cont(c),
+   rate(new Speedometer("xfer:rate-period")),
+   rate_for_eta(new Speedometer("xfer:eta-period"))
 {
-   get=0;
-   put=0;
    set_state(INITIAL);
    max_buf=0x10000;
-   cont=false;
-   error_text=0;
-   rate        =MakeRef(new Speedometer("xfer:rate-period"));
-   rate_for_eta=MakeRef(new Speedometer("xfer:eta-period"));
    put_buf=0;
    put_eof_pos=0;
    bytes_count=0;
@@ -417,25 +421,10 @@ void FileCopy::Init()
    fail_if_broken=true;
    remove_source_later=false;
    remove_target_first=false;
-   line_buffer=0;
    line_buffer_max=0;
 }
-
-FileCopy::FileCopy(FileCopyPeer *s,FileCopyPeer *d,bool c)
-{
-   Init();
-   get=MakeRef(s);
-   put=MakeRef(d);
-   cont=c;
-}
 FileCopy::~FileCopy()
 {
-   DeleteRef(get);
-   DeleteRef(put);
-   delete line_buffer;
-   xfree(error_text);
-   DeleteRef(rate);
-   DeleteRef(rate_for_eta);
 }
 FileCopy *FileCopy::New(FileCopyPeer *s,FileCopyPeer *d,bool c)
 {
@@ -467,16 +456,11 @@ void FileCopy::Bg()
    if(put) put->Bg();
 }
 
-void FileCopy::Reconfig(const char *s)
-{
-}
-
 void FileCopy::SetError(const char *str)
 {
-   xfree(error_text);
-   error_text=xstrdup(str);
-   DeleteRef(get);
-   DeleteRef(put);
+   error_text.set(str);
+   get=0;
+   put=0;
 }
 
 void FileCopy::LineBuffered(int s)
@@ -572,20 +556,15 @@ long FileCopy::GetETA(off_t b)
 }
 const char *FileCopy::GetStatus()
 {
-   static char *buf;
-   xfree(buf); buf=0;
-   const char *get_st=0;
-   if(get)
-      get_st=get->GetStatus();
-   const char *put_st=0;
-   if(put)
-      put_st=put->GetStatus();
+   static xstring buf;
+   const char *get_st=get?get->GetStatus():0;
+   const char *put_st=put?put->GetStatus():0;
    if(get_st && get_st[0] && put_st && put_st[0])
-      buf=xasprintf("[%s->%s]",get_st,put_st);
+      buf.vset("[",get_st,"->",put_st,"]",NULL);
    else if(get_st && get_st[0])
-      buf=xasprintf("[%s]",get_st);
+      buf.vset("[",get_st,"]",NULL);
    else if(put_st && put_st[0])
-      buf=xasprintf("[%s]",put_st);
+      buf.vset("[",put_st,"]",NULL);
    else
       return "";
    return buf;
@@ -623,6 +602,36 @@ void FileCopy::Kill(int sig)
    if(put) put->Kill(sig);
 }
 
+SMTaskRef<Log> FileCopy::transfer_log;
+
+void FileCopy::LogTransfer()
+{
+   if(!ResMgr::QueryBool("xfer:log",0))
+      return;
+   const char *src=get->GetURL();
+   const char *dst=put->GetURL();
+   if(!src || !dst)
+      return;
+   if(!transfer_log)
+   {
+      int fd=open(dir_file(get_lftp_home(),"transfer_log"),O_WRONLY|O_APPEND|O_CREAT,0600);
+      if(fd==-1)
+	 return;
+      transfer_log=new Log;
+      transfer_log->SetOutput(fd,true);
+      transfer_log->ShowNothing();
+      transfer_log->ShowTime();	 // nothing but time
+      transfer_log->Enable();
+   }
+   long long range_limit=GetRangeLimit();
+   if(range_limit==FILE_END)
+      range_limit=get->GetPos();
+   transfer_log->Format(0,"%s -> %s %lld-%lld %s\n",src,dst,
+      (long long)GetRangeStart(),range_limit,
+      Speedometer::GetStr(GetBytesCount()/GetTimeSpent()));
+}
+
+
 // FileCopyPeer implementation
 #undef super
 #define super Buffer
@@ -650,7 +659,7 @@ bool FileCopyPeer::Done()
 {
    if(Error())
       return true;
-   if(eof && in_buffer==0)
+   if(eof && Size()==0)
    {
       if(removing)
 	 return false;
@@ -667,7 +676,7 @@ void FileCopyPeer::Seek(off_t offs)
 {
    seek_pos=offs;
    if(mode==PUT)
-      pos-=in_buffer;
+      pos-=Size();
    Empty();
    eof=false;
    broken=false;
@@ -693,14 +702,9 @@ FileCopyPeer::FileCopyPeer(dir_t m) : IOBuffer(m)
    use_cache=true;
    write_allowed=true;
    done=false;
-   suggested_filename=0;
    auto_rename=false;
    Suspend();  // don't do anything too early
 }
-FileCopyPeer::~FileCopyPeer()
-{
-   xfree(suggested_filename);
-}
 
 // FileCopyPeerFA implementation
 #undef super
@@ -719,10 +723,9 @@ int FileCopyPeerFA::Do()
 	 file_removed=true;
 	 session->Close();
 	 Suspend();
-	 return MOVED;
+	 m=MOVED;
       }
-      else
-	 return m;
+      return m;
    }
 
    if(Done() || Error())
@@ -773,16 +776,15 @@ int FileCopyPeerFA::Do()
 	    goto fxp_eof;
 	 return m;
       }
-      res=Put_LL(buffer+buffer_ptr,in_buffer);
+      res=Put_LL(buffer+buffer_ptr,Size());
       if(res>0)
       {
-	 in_buffer-=res;
 	 buffer_ptr+=res;
 	 m=MOVED;
       }
       else if(res<0)
 	 return MOVED;
-      if(in_buffer==0)
+      if(Size()==0)
       {
 	 if(eof)
 	 {
@@ -798,7 +800,7 @@ int FileCopyPeerFA::Do()
 	       // FIXME: set date for real.
 	       date_set=true;
 	       if(!verify)
-		  verify=MakeRef(new FileVerificator(session,file));
+		  verify=new FileVerificator(session,file);
 	       return MOVED;
 	    }
 	    else if(res==FA::IN_PROGRESS)
@@ -835,7 +837,7 @@ int FileCopyPeerFA::Do()
       res=Get_LL(GET_BUFSIZE);
       if(res>0)
       {
-	 in_buffer+=res;
+	 EmbraceNewData(res);
 	 SaveMaxCheck(0);
 	 return MOVED;
       }
@@ -864,7 +866,8 @@ void FileCopyPeerFA::SuspendInternal()
 {
    if(fxp && mode==PUT)
       return;
-   session->SuspendSlave();
+   if(session->IsOpen())
+      session->SuspendSlave();
 }
 void FileCopyPeerFA::ResumeInternal()
 {
@@ -922,9 +925,7 @@ void FileCopyPeerFA::OpenSession()
 	 b+=seek_pos;
 	 s-=seek_pos;
 	 Save(0);
-	 Allocate(s);
-	 memmove(buffer+buffer_ptr,b,s);
-	 in_buffer=s;
+	 Put(b,s);
 	 pos=seek_pos;
 	 eof=true;
 	 return;
@@ -963,14 +964,9 @@ void FileCopyPeerFA::OpenSession()
    if(want_date && (date==NO_DATE_YET || date.ts_prec>1))
       session->WantDate(&date);
    if(mode==GET)
-   {
       SaveRollback(seek_pos);
-      pos=seek_pos;
-   }
    else
-   {
-      pos=seek_pos+in_buffer;
-   }
+      pos=seek_pos+Size();
 }
 
 void FileCopyPeerFA::WantSize()
@@ -1003,9 +999,7 @@ int FileCopyPeerFA::Get_LL(int len)
    if(GetRealPos()!=io_at) // GetRealPos can alter pos.
       return 0;
 
-   Allocate(len);
-
-   res=session->Read(buffer+buffer_ptr+in_buffer,len);
+   res=session->Read(GetSpace(len),len);
    if(res<0)
    {
       if(res==FA::DO_AGAIN)
@@ -1034,16 +1028,11 @@ int FileCopyPeerFA::Get_LL(int len)
 
 	    if(u.proto)
 	    {
-	       if(reuse_later)
-		  SessionPool::Reuse(session);
+	       my_session=FileAccess::New(&u);
+	       session=my_session;
 
-	       session=FileAccess::New(&u);
-	       reuse_later=true;
-
-	       xfree(file);
-	       file=xstrdup(u.path?u.path:"");
-	       xfree(orig_url);
-	       orig_url=xstrdup(loc);
+	       file.set(u.path?u.path.get():"");
+	       orig_url.set(loc);
 	    }
 	    else // !proto
 	    {
@@ -1056,31 +1045,26 @@ int FileCopyPeerFA::Get_LL(int len)
 		     s_ind=p_ind=strlen(orig_url);
 		  if(loc[0]=='/')
 		  {
-		     orig_url=(char*)xrealloc(orig_url,p_ind+strlen(loc)+1);
-		     strcpy(orig_url+p_ind,loc);
+		     orig_url.truncate(p_ind);
+		     orig_url.append(loc);
 		  }
 		  else
 		  {
-		     orig_url=(char*)xrealloc(orig_url,s_ind+1+strlen(loc)+1);
-		     strcpy(orig_url+s_ind,"/");
-		     strcpy(orig_url+s_ind+1,loc);
+		     orig_url.truncate(s_ind);
+		     orig_url.append('/');
+		     orig_url.append(loc);
 		  }
 	       }
 
 	       url::decode_string(loc);
-	       char *slash=strrchr(file,'/');
-	       char *new_file;
+	       const char *slash=strrchr(file,'/');
 	       if(loc[0]!='/' && slash)
 	       {
-		  *slash=0;
-		  new_file=xstrdup(dir_file(file,loc));
+		  file.truncate(slash-file);
+		  file.set(dir_file(file,loc));
 	       }
 	       else
-	       {
-		  new_file=xstrdup(loc);
-	       }
-	       xfree(file);
-	       file=new_file;
+		  file.set(loc);
 	    }
 
 	    size=NO_SIZE_YET;
@@ -1100,6 +1084,7 @@ int FileCopyPeerFA::Get_LL(int len)
       eof=true;
       FileAccess::cache->Add(session,file,FAmode,FA::OK,this);
       SetSuggestedFileName(session->GetSuggestedFileName());
+      session->Close();
    }
    return res;
 }
@@ -1140,13 +1125,20 @@ int FileCopyPeerFA::Put_LL(const char *buf,int len)
    return res;
 }
 
+int FileCopyPeerFA::PutEOF_LL()
+{
+   if(mode==GET && session)
+      session->Close();
+   return 0;
+}
+
 off_t FileCopyPeerFA::GetRealPos()
 {
    if(session->OpenMode()!=FAmode || fxp)
       return pos;
    if(mode==PUT)
    {
-      if(pos-in_buffer!=session->GetPos())
+      if(pos-Size()!=session->GetPos())
       {
 	 Empty();
 	 can_seek=false;
@@ -1162,97 +1154,91 @@ off_t FileCopyPeerFA::GetRealPos()
 	 can_seek=false;
 	 session->SeekReal();
       }
-      if(pos+in_buffer!=session->GetPos())
-      {
+      if(pos+Size()!=session->GetPos())
 	 SaveRollback(session->GetPos());
-	 pos=session->GetPos();
-      }
    }
    return pos;
 }
 
 void FileCopyPeerFA::Init()
 {
-   FAmode=FA::RETRIEVE;
-   file=0;
-   session=0;
-   reuse_later=false;
-   orig_url=0;
    fxp=false;
    try_time=0;
    retries=0;
    redirections=0;
    can_seek=true;
    can_seek0=true;
-   verify=0;
-}
-
-FileCopyPeerFA::FileCopyPeerFA(FileAccess *s,const char *f,int m)
-   : FileCopyPeer(m==FA::STORE ? PUT : GET)
-{
-   Init();
-   FAmode=m;
-   file=xstrdup(f);
-   session=s;
-   reuse_later=true;
    if(FAmode==FA::LIST || FAmode==FA::LONG_LIST)
       Save(FileAccess::cache->SizeLimit());
 }
+
+FileCopyPeerFA::FileCopyPeerFA(FileAccess *s,const char *f,int m)
+   : FileCopyPeer(m==FA::STORE ? PUT : GET), file(f),
+     my_session(s), session(my_session), FAmode(m)
+{
+   Init();
+}
+FileCopyPeerFA::FileCopyPeerFA(const FileAccessRef& s,const char *f,int m)
+   : FileCopyPeer(m==FA::STORE ? PUT : GET), file(f),
+     session(s), FAmode(m)
+{
+   Init();
+}
+FileCopyPeerFA::FileCopyPeerFA(const ParsedURL *u,int m)
+   : FileCopyPeer(m==FA::STORE ? PUT : GET), file(u->path), orig_url(u->orig_url),
+     my_session(FileAccess::New(u)), session(my_session), FAmode(m)
+{
+   Init();
+   if(!file)
+      SetError(_("file name missed in URL"));
+}
+
 FileCopyPeerFA::~FileCopyPeerFA()
 {
    if(session)
-   {
       session->Close();
-      if(reuse_later)
-	 SessionPool::Reuse(session);
-   }
-   xfree(file);
-   xfree(orig_url);
-   DeleteRef(verify);
 }
 
-FileCopyPeerFA::FileCopyPeerFA(ParsedURL *u,int m)
-   : FileCopyPeer(m==FA::STORE ? PUT : GET)
-{
-   Init();
-   FAmode=m;
-   file=xstrdup(u->path);
-   session=FileAccess::New(u);
-   reuse_later=true;
-   orig_url=u->orig_url;
-   u->orig_url=0;
-   if(!file)
-   {
-      SetError(_("file name missed in URL"));
-   }
-}
-
-FileCopyPeerFA *FileCopyPeerFA::New(FileAccess *s,const char *url,int m,bool reuse)
+FileCopyPeerFA *FileCopyPeerFA::New(FileAccess *s,const char *url,int m)
 {
    ParsedURL u(url,true);
    if(u.proto)
    {
-      if(reuse)
-	 SessionPool::Reuse(s);
+      SessionPool::Reuse(s);
       return new FileCopyPeerFA(&u,m);
    }
-   FileCopyPeerFA *peer=new FileCopyPeerFA(s,url,m);
-   if(!reuse)
-      peer->DontReuseSession();
-   return peer;
+   return new FileCopyPeerFA(s,url,m);
+}
+FileCopyPeerFA *FileCopyPeerFA::New(const FileAccessRef& s,const char *url,int m)
+{
+   ParsedURL u(url,true);
+   if(u.proto)
+      return new FileCopyPeerFA(&u,m);
+   return new FileCopyPeerFA(s,url,m);
+}
+FileCopyPeer *FileCopyPeerFA::Clone()
+{
+   return new FileCopyPeerFA(session->Clone(),file,FAmode);
 }
 
 // FileCopyPeerFDStream
 #undef super
 #define super FileCopyPeer
+
 FileCopyPeerFDStream::FileCopyPeerFDStream(FDStream *o,dir_t m)
-   : FileCopyPeer(m)
+   : FileCopyPeer(m), my_stream(o?o:new FDStream(1,"<stdout>")), stream(my_stream)
+{
+   Init();
+}
+FileCopyPeerFDStream::FileCopyPeerFDStream(const Ref<FDStream>& o,dir_t m)
+   : FileCopyPeer(m), stream(o)
+{
+   Init();
+}
+
+void FileCopyPeerFDStream::Init()
 {
-   if(o==0 && m==PUT)
-      o=new FDStream(1,"<stdout>");
-   stream=o;
    seek_base=0;
-   delete_stream=true;
    create_fg_data=true;
    need_seek=false;
    can_seek = can_seek0 = stream->can_seek();
@@ -1268,17 +1254,8 @@ FileCopyPeerFDStream::FileCopyPeerFDStream(FDStream *o,dir_t m)
    }
    if(stream->usesfd(1))
       write_allowed=false;
-   put_ll_timer=0;
-   if(m==PUT)
+   if(mode==PUT)
       put_ll_timer=new Timer(0,200);
-   verify=0;
-}
-FileCopyPeerFDStream::~FileCopyPeerFDStream()
-{
-   if(delete_stream)
-      delete stream;
-   delete put_ll_timer;
-   DeleteRef(verify);
 }
 
 void FileCopyPeerFDStream::Seek_LL()
@@ -1317,7 +1294,7 @@ void FileCopyPeerFDStream::Seek_LL()
 	 pos=seek_pos;
       }
       if(mode==PUT)
-	 pos+=in_buffer;
+	 pos+=Size();
    }
    else
    {
@@ -1346,7 +1323,7 @@ int FileCopyPeerFDStream::getfd()
    stream->clear_status();
    pos=0;
    if(mode==PUT)
-      pos+=in_buffer;
+      pos+=Size();
    Seek_LL();
    return fd;
 }
@@ -1364,11 +1341,9 @@ int FileCopyPeerFDStream::Do()
       {
 	 if(suggested_filename && stream && stream->full_name && auto_rename)
 	 {
-	    char *dir=alloca_strdup(stream->full_name);
-	    dirname_modify(dir);
-	    debug((5,"copy: renaming `%s' to `%s'\n",stream->full_name,suggested_filename));
-	    if(rename(stream->full_name,dir_file(dir,suggested_filename))==-1)
-	       debug((3,"rename(%s, %s): %s\n",stream->full_name,suggested_filename,strerror(errno)));
+	    debug((5,"copy: renaming `%s' to `%s'\n",stream->full_name.get(),suggested_filename.get()));
+	    if(rename(stream->full_name,dir_file(dirname(stream->full_name),suggested_filename))==-1)
+	       debug((3,"rename(%s, %s): %s\n",stream->full_name.get(),suggested_filename.get(),strerror(errno)));
 	 }
 	 done=true;
 	 m=MOVED;
@@ -1384,7 +1359,7 @@ int FileCopyPeerFDStream::Do()
    switch(mode)
    {
    case PUT:
-      if(in_buffer==0)
+      if(Size()==0)
       {
 	 if(eof)
 	 {
@@ -1397,10 +1372,10 @@ int FileCopyPeerFDStream::Do()
 	       date_set=true;
 	       m=MOVED;
 	    }
-	    if(stream && delete_stream && !stream->Done())
+	    if(stream && my_stream && !stream->Done())
 	       return m;
 	    if(!verify)
-	       verify=MakeRef(new FileVerificator(stream));
+	       verify=new FileVerificator(stream);
 	    return MOVED;
 	 }
 	 if(seek_pos==0)
@@ -1408,15 +1383,16 @@ int FileCopyPeerFDStream::Do()
       }
       if(!write_allowed)
 	 return m;
-      while(in_buffer>0)
+      if(getfd()==-1)
+	 return m;
+      while(Size()>0)
       {
-	 if(check_min_size && !eof && in_buffer<PUT_LL_MIN
+	 if(check_min_size && !eof && Size()<PUT_LL_MIN
 	 && put_ll_timer && !put_ll_timer->Stopped())
 	    break;
-	 int res=Put_LL(buffer+buffer_ptr,in_buffer);
+	 int res=Put_LL(buffer+buffer_ptr,Size());
 	 if(res>0)
 	 {
-	    in_buffer-=res;
 	    buffer_ptr+=res;
 	    m=MOVED;
 	 }
@@ -1430,12 +1406,12 @@ int FileCopyPeerFDStream::Do()
    case GET:
       if(eof)
 	 return m;
-      while(in_buffer<GET_BUFSIZE)
+      while(Size()<GET_BUFSIZE)
       {
 	 int res=Get_LL(GET_BUFSIZE);
 	 if(res>0)
 	 {
-	    in_buffer+=res;
+	    EmbraceNewData(res);
 	    SaveMaxCheck(0);
 	    m=MOVED;
 	 }
@@ -1476,7 +1452,7 @@ void FileCopyPeerFDStream::Seek(off_t new_pos)
    {
       if(seek_pos!=FILE_END)
       {
-	 pos=seek_pos+(mode==PUT)*in_buffer;
+	 pos=seek_pos+(mode==PUT?Size():0);
 	 return;
       }
       else
@@ -1485,7 +1461,7 @@ void FileCopyPeerFDStream::Seek(off_t new_pos)
 	 if(s!=-1)
 	 {
 	    SetSize(s);
-	    pos=seek_pos+(mode==PUT)*in_buffer;
+	    pos=seek_pos+(mode==PUT)?Size():0;
 	    return;
 	 }
 	 else
@@ -1524,17 +1500,11 @@ int FileCopyPeerFDStream::Get_LL(int len)
       }
    }
 
-#ifndef NATIVE_CRLF
-   if(ascii)
-      Allocate(len*2);
-   else
-#endif
-      Allocate(len);
-
    if(need_seek)  // this does not combine with ascii.
       lseek(fd,seek_base+pos,SEEK_SET);
 
-   res=read(fd,buffer+buffer_ptr+in_buffer,len);
+   char *p=GetSpace(ascii?len*2:len);
+   res=read(fd,p,len);
    if(res==-1)
    {
       if(E_RETRY(errno))
@@ -1553,7 +1523,6 @@ int FileCopyPeerFDStream::Get_LL(int len)
 #ifndef NATIVE_CRLF
    if(ascii)
    {
-      char *p=buffer+buffer_ptr+in_buffer;
       for(int i=res; i>0; i--)
       {
 	 if(*p=='\n')
@@ -1615,7 +1584,7 @@ int FileCopyPeerFDStream::Put_LL(const char *buf,int len)
       return skip_cr;
 
    if(need_seek)  // this does not combine with ascii.
-      lseek(fd,seek_base+pos-in_buffer,SEEK_SET);
+      lseek(fd,seek_base+pos-Size(),SEEK_SET);
 
    int res=write(fd,buf,len);
    if(res<0)
@@ -1628,7 +1597,7 @@ int FileCopyPeerFDStream::Put_LL(const char *buf,int len)
       if(errno==EPIPE)
       {
 	 broken=true;
-	 in_buffer=0;
+	 buffer.truncate(buffer_ptr);
 	 eof=true;
 	 return -1;
       }
@@ -1640,11 +1609,11 @@ int FileCopyPeerFDStream::Put_LL(const char *buf,int len)
 	    struct stat st;
 	    if(fstat(fd,&st)!=-1)
 	    {
-	       if(st.st_size<seek_base+pos-in_buffer)
+	       if(st.st_size<seek_base+pos-Size())
 	       {
 		  // workaround solaris nfs bug. It can lose data if disk is full.
-		  if(buffer_ptr>=seek_base+pos-in_buffer-buffer_ptr-st.st_size)
-		     UnSkip(seek_base+pos-in_buffer-st.st_size);
+		  if(buffer_ptr>=seek_base+pos-Size()-buffer_ptr-st.st_size)
+		     UnSkip(seek_base+pos-Size()-st.st_size);
 		  else
 		  {
 		     Empty();
@@ -1674,7 +1643,7 @@ int FileCopyPeerFDStream::Put_LL(const char *buf,int len)
 }
 FgData *FileCopyPeerFDStream::GetFgData(bool fg)
 {
-   if(!delete_stream || !create_fg_data)
+   if(!my_stream || !create_fg_data)
       return 0;	  // if we don't own the stream, don't create FgData.
    if(stream->GetProcGroup())
       return new FgData(stream->GetProcGroup(),fg);
@@ -1728,26 +1697,26 @@ FileCopyPeerFDStream *FileCopyPeerFDStream::NewGet(const char *file)
    return new FileCopyPeerFDStream(new FileStream(file,O_RDONLY),
 				    FileCopyPeer::GET);
 }
-
+FileCopyPeer *FileCopyPeerFDStream::Clone()
+{
+   NeedSeek();
+   FileCopyPeerFDStream *peer=new FileCopyPeerFDStream(stream,mode);
+   peer->NeedSeek();
+   peer->SetBase(0);
+   return peer;
+}
 
 // FileCopyPeerDirList
 FileCopyPeerDirList::FileCopyPeerDirList(FA *s,ArgV *v)
-   : FileCopyPeer(GET)
+   : FileCopyPeer(GET), session(s)
 {
-   session=s;
-   dl=MakeRef(session->MakeDirList(v));
+   dl=session->MakeDirList(v);
    if(dl==0)
       eof=true;
    can_seek=false;
    can_seek0=false;
 }
 
-FileCopyPeerDirList::~FileCopyPeerDirList()
-{
-   DeleteRef(dl);
-   SessionPool::Reuse(session);
-}
-
 int FileCopyPeerDirList::Do()
 {
    if(Done())
@@ -1768,9 +1737,8 @@ int FileCopyPeerDirList::Do()
    }
    if(s==0)
       return STALL;
-   Allocate(s);
-   memmove(buffer+buffer_ptr+in_buffer,b,s);
-   in_buffer+=s;
+   memcpy(GetSpace(s),b,s);
+   SpaceAdd(s);
    dl->Skip(s);
    return MOVED;
 }
@@ -1779,7 +1747,6 @@ int FileCopyPeerDirList::Do()
 void FileVerificator::Init0()
 {
    done=false;
-   error_text=0;
    verify_buffer=0;
    verify_process=0;
    if(!ResMgr::QueryBool("xfer:verify",0))
@@ -1828,7 +1795,7 @@ FileVerificator::FileVerificator(const FDStream *stream)
       verify_process->SetCwd(cwd);
    }
 }
-FileVerificator::FileVerificator(FileAccess *session,const char *f)
+FileVerificator::FileVerificator(const FileAccess *session,const char *f)
 {
    Init0();
    if(done)
@@ -1843,7 +1810,6 @@ FileVerificator::FileVerificator(FileAccess *session,const char *f)
 }
 FileVerificator::~FileVerificator()
 {
-   xfree(error_text);
    // verify_process is deleted by verify_buffer dtor
    Delete(verify_buffer);
 }
@@ -1861,23 +1827,13 @@ int FileVerificator::Do()
    m=MOVED;
    if(verify_process->GetProcExitCode()!=0)
    {
-      verify_buffer->Put("",1);
-      int len;
-      const char *s;
-      verify_buffer->Get(&s,&len);
-      len=strlen(s);
-      while(len>0 && s[len-1]=='\n')
-	 len--;
-      if(len==0)
-      {
-	 s=_("Verify command failed without a message");
-	 len=strlen(s);
-      }
-      error_text=xstrdup(s);
-      error_text[len]=0;
-      char *nl=strrchr(error_text,'\n');
+      error_text.set(verify_buffer->Get());
+      error_text.rtrim('\n');
+      if(error_text.length()==0)
+	 error_text.set(_("Verify command failed without a message"));
+      const char *nl=strrchr(error_text,'\n');
       if(nl)
-	 memmove(error_text,nl+1,len-(nl+1-error_text)+1);
+	 error_text.set(nl+1);
    }
    return m;
 }
