@@ -18,13 +18,13 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FileSet.h,v 1.38.2.1 2007/04/05 13:23:37 lav Exp $ */
+/* $Id: FileSet.h,v 1.48 2007/09/20 11:29:40 lav Exp $ */
 
 #ifndef FILESET_H
 #define FILESET_H
 
 #include <sys/types.h>
-#include "xmalloc.h"
+#include "xarray.h"
 
 #undef TYPE
 
@@ -50,11 +50,13 @@ struct FileTimestamp
 class FileInfo
 {
 public:
-   char	    *name;
+   xstring  name;
+   xstring  longname;
+   xstring_c symlink;
    mode_t   mode;
    FileTimestamp date;
    off_t    size;
-   void	    *data;
+   xstring  data;
    const char *user, *group;
    int      nlinks;
 
@@ -65,9 +67,7 @@ public:
       SYMLINK,
       NORMAL
    };
-
    type	 filetype;
-   char	 *symlink;
 
    int	 defined;
    enum defined_bits
@@ -82,15 +82,13 @@ public:
    };
 
    int rank;
-   char *longname;
 
-   ~FileInfo();
    void Init();
    FileInfo() { Init(); }
    FileInfo(const FileInfo &fi);
    FileInfo(const char *n) { Init(); SetName(n); }
 
-   void SetName(const char *n);
+   void SetName(const char *n) { name.set(n); defined|=NAME; }
    void SetUser(const char *n);
    void SetGroup(const char *n);
    void LocalFile(const char *name, bool follow_symlinks);
@@ -99,29 +97,26 @@ public:
    void SetMode(mode_t m) { mode=m; defined|=MODE; }
    void SetDate(time_t t,int prec) { date.set(t,prec); defined|=DATE; }
    void SetType(type t) { filetype=t; defined|=TYPE; }
-   void SetSymlink(const char *s) { xfree(symlink); symlink=xstrdup(s);
-      filetype=SYMLINK; defined|=TYPE|SYMLINK_DEF; }
+   void SetSymlink(const char *s) { symlink.set(s); filetype=SYMLINK; defined|=TYPE|SYMLINK_DEF; }
    void	SetSize(off_t s) { size=s; defined|=SIZE; }
    void	SetNlink(int n) { nlinks=n; defined|=NLINKS; }
 
    void	 Merge(const FileInfo&);
 
-   bool	 SameAs(const FileInfo *,int ignore);
-   bool	 OlderThan(time_t t);
-   bool	 NewerThan(time_t t);
-   bool  SizeOutside(const Range *r);
+   bool	 SameAs(const FileInfo *,int ignore) const;
+   bool	 OlderThan(time_t t) const;
+   bool	 NewerThan(time_t t) const;
+   bool	 NotOlderThan(time_t t) const;
+   bool	 NotNewerThan(time_t t) const;
+   bool  SizeOutside(const Range *r) const;
 
-   void	 SetAssociatedData(void *d,int len)
-      {
-	 xfree(data);
-	 data=xmemdup(d,len);
-      }
-   void  *GetAssociatedData() { return data; }
+   void	 SetAssociatedData(const void *d,int len) { data.nset((const char*)d,len); }
+   const void *GetAssociatedData() { return data; }
 
    void SetRank(int r) { rank=r; }
    int GetRank() const { return rank; }
    void MakeLongName();
-   void SetLongName(const char *s) { xfree(longname); longname=xstrdup(s); }
+   void SetLongName(const char *s) { longname.set(s); }
    const char *GetLongName() { if(!longname) MakeLongName(); return longname; }
 
    operator const char *() { return name; }
@@ -135,42 +130,36 @@ public:
    enum sort_e { BYNAME, BYSIZE, DIRSFIRST, BYRANK, BYDATE };
 
 private:
-   FileInfo **files;
+   RefArray<FileInfo> files;
 
-   /* Alternate pointers when sort != NAME: */
-   FileInfo **files_sort;
-
-   int	 fnum;
-   int	 allocated;
+   /* indexes when sort != NAME: */
+   xarray<int> sorted;
 
    int	 ind;
 
    void	 Sub(int);
-   bool sorted;
 
    void add_before(int pos,FileInfo *fi);
 
 public:
-   FileSet()
-   {
-      files=files_sort=0;
-      sorted=false;
-      fnum=0;
-      allocated=0;
-      ind=0;
-   }
+   FileSet();
    FileSet(const FileSet *s);
-   ~FileSet();
    void Empty();
 
-   int	 get_fnum() const { return fnum; }
+   int	 get_fnum() const { return files.count(); }
+   int	 count() const { return files.count(); }
+   int	 curr_index() const { return ind; }
+   int	 curr_pct() const { return count()==0 ? 100 : ind*100/count(); }
 
    void	 Add(FileInfo *);
    void	 Merge(const FileSet *);
    void	 SubtractSame(const FileSet *,int ignore);
    void	 SubtractAny(const FileSet *);
-   void  SubtractOlderThan(time_t t);
-   void  SubtractNewerThan(time_t t);
+   void  SubtractTimeCmp(bool (FileInfo::*cmp)(time_t) const,time_t);
+   void  SubtractOlderThan(time_t t) { SubtractTimeCmp(&FileInfo::OlderThan,t); }
+   void  SubtractNewerThan(time_t t) { SubtractTimeCmp(&FileInfo::NewerThan,t); }
+   void  SubtractNotOlderThan(time_t t) { SubtractTimeCmp(&FileInfo::NotOlderThan,t); }
+   void  SubtractNotNewerThan(time_t t) { SubtractTimeCmp(&FileInfo::NotNewerThan,t); }
    void  SubtractSizeOutside(const Range *r);
    void  SubtractDirs();
    void  SubtractNotDirs();
@@ -181,7 +170,7 @@ public:
    void  Unsort();
    void	 SortByPatternList(const char *list_c);
 
-   void	 Exclude(const char *prefix,PatternSet *x);
+   void	 Exclude(const char *prefix,const PatternSet *x);
    void	 ExcludeDots();
    void	 ExcludeUnaccessible();
 
@@ -194,7 +183,7 @@ public:
    void	 LocalChmod(const char *dir,mode_t mask=0);
    void	 LocalChown(const char *dir);
 
-   void Count(int *d,int *f,int *s,int *o);
+   void Count(int *d,int *f,int *s,int *o) const;
 
    int FindGEIndByName(const char *name) const;
    FileInfo *FindByName(const char *name) const;
@@ -220,7 +209,8 @@ public:
 
    FileInfo * operator[](int i) const;
 
-   int EstimateMemory() const;
+   size_t EstimateMemory() const;
+   void Dump(const char *tag) const;
 };
 
 #endif // FILESET_H
