@@ -72,7 +72,7 @@ int Fish::Do()
    // check if idle time exceeded
    if(mode==CLOSED && send_buf && idle_timer.Stopped())
    {
-      DebugPrint("---- ",_("Closing idle connection"),1);
+      LogNote(1,_("Closing idle connection"));
       Disconnect();
       return m;
    }
@@ -130,9 +130,7 @@ int Fish::Do()
 	 return MOVED;
 
       const char *shell=Query("shell",hostname);
-      char *init=alloca_strdup2("echo FISH:;",strlen(shell));
-      strcat(init,shell);
-
+      const char *init=xstring::cat("echo FISH:;",shell,NULL);
       const char *prog=Query("connect-program",hostname);
       if(!prog || !prog[0])
 	 prog="ssh -a -x";
@@ -152,11 +150,8 @@ int Fish::Do()
       }
       cmd->Add(hostname);
       cmd->Add(init);
-      {
-      char *cmd_str=cmd->Combine(0);
-      Log::global->Format(9,"---- %s (%s)\n",_("Running connect program"),cmd_str);
-      xfree(cmd_str);
-      }
+      xstring_ca cmd_str(cmd->Combine(0));
+      Log::global->Format(9,"---- %s (%s)\n",_("Running connect program"),cmd_str.get());
       ssh=new PtyShell(cmd);
       state=CONNECTING;
       timeout_timer.Reset();
@@ -175,8 +170,7 @@ int Fish::Do()
 	 return m;
       }
       ssh->Kill(SIGCONT);
-      send_buf=new IOBufferFDStream(ssh,IOBuffer::PUT);
-      ssh=0;
+      send_buf=new IOBufferFDStream(ssh.borrow(),IOBuffer::PUT);
       recv_buf=new IOBufferFDStream(new FDStream(fd,"pseudo-tty"),IOBuffer::GET);
       set_real_cwd("~");
       state=CONNECTING_1;
@@ -218,10 +212,10 @@ int Fish::Do()
       ExpandTildeInCWD();
       if(mode!=CHANGE_DIR && xstrcmp(cwd,real_cwd))
       {
-	 if(path_queue_len==0 || strcmp(path_queue[path_queue_len-1],cwd))
+	 if(xstrcmp(path_queue.LastString(),cwd))
 	 {
 	    Send("#CWD %s\n"
-		 "cd %s; echo '### 000'\n",cwd.path,shell_encode(cwd));
+		 "cd %s; echo '### 000'\n",cwd.path.get(),shell_encode(cwd));
 	    PushExpect(EXPECT_CWD);
 	    PushDirectory(cwd);
 	 }
@@ -298,18 +292,13 @@ usual_return:
 
 void Fish::MoveConnectionHere(Fish *o)
 {
-   send_buf=o->send_buf; o->send_buf=0;
-   recv_buf=o->recv_buf; o->recv_buf=0;
-   rate_limit=o->rate_limit; o->rate_limit=0;
-   path_queue=o->path_queue; o->path_queue=0;
-   path_queue_len=o->path_queue_len; o->path_queue_len=0;
-   RespQueue=o->RespQueue; o->RespQueue=0;
-   RQ_alloc=o->RQ_alloc; o->RQ_alloc=0;
-   RQ_head=o->RQ_head; o->RQ_head=0;
-   RQ_tail=o->RQ_tail; o->RQ_tail=0;
+   send_buf=o->send_buf.borrow();
+   recv_buf=o->recv_buf.borrow();
+   rate_limit=o->rate_limit.borrow();
+   path_queue.MoveHere(o->path_queue);
+   RespQueue.move_here(o->RespQueue);
    timeout_timer.Reset(o->timeout_timer);
    set_real_cwd(o->real_cwd);
-   o->set_real_cwd(0);
    state=CONNECTED;
    o->Disconnect();
    if(!home)
@@ -320,10 +309,10 @@ void Fish::MoveConnectionHere(Fish *o)
 void Fish::Disconnect()
 {
    if(send_buf)
-      DebugPrint("---- ",_("Disconnecting"),9);
-   Delete(send_buf); send_buf=0;
-   Delete(recv_buf); recv_buf=0;
-   delete ssh; ssh=0;
+      LogNote(9,_("Disconnecting"));
+   send_buf=0;
+   recv_buf=0;
+   ssh=0;
    EmptyRespQueue();
    EmptyPathQueue();
    state=DISCONNECTED;
@@ -331,15 +320,7 @@ void Fish::Disconnect()
       SetError(STORE_FAILED,0);
    received_greeting=false;
    password_sent=0;
-   xfree(home_auto); home_auto=0;
-   home_auto=xstrdup(FindHomeAuto());
-}
-
-void Fish::EmptyPathQueue()
-{
-   for(int i=0; i<path_queue_len; i++)
-      xfree(path_queue[i]);
-   path_queue_len=0;
+   home_auto.set(FindHomeAuto());
 }
 
 void Fish::Init()
@@ -347,16 +328,8 @@ void Fish::Init()
    state=DISCONNECTED;
    send_buf=0;
    recv_buf=0;
-   ssh=0;
    max_send=0;
-   line=0;
-   message=0;
-   RespQueue=0;
-   RQ_alloc=0;
-   RQ_head=RQ_tail=0;
    eof=false;
-   path_queue=0;
-   path_queue_len=0;
    received_greeting=false;
    password_sent=0;
 }
@@ -370,12 +343,6 @@ Fish::Fish()
 Fish::~Fish()
 {
    Disconnect();
-   xfree(line);
-   xfree(message);
-   EmptyRespQueue();
-   xfree(RespQueue);
-   EmptyPathQueue();
-   xfree(path_queue);
 }
 
 Fish::Fish(const Fish *o) : super(o)
@@ -414,75 +381,13 @@ void Fish::Close()
 void Fish::Send(const char *format,...)
 {
    va_list va;
-   char *str;
-
-   static int max_send=256;
-   for(;;)
-   {
-      va_start(va,format);
-      str=string_alloca(max_send);
-      int res=vsnprintf(str,max_send,format,va);
-      va_end(va);
-      if(res>=0 && res<max_send)
-      {
-	 if(res<max_send/16)
-	    max_send/=2;
-	 break;
-      }
-      max_send*=2;
-   }
-
-   DebugPrint("---> ",str,5);
+   va_start(va,format);
+   xstring& str=xstring::vformat(format,va);
+   va_end(va);
+   LogSend(5,str);
    send_buf->Put(str);
 }
 
-const char *Fish::shell_encode(const char *string)
-{
-   if(!string)
-      return 0;
-
-   int c;
-   static char *result=0;
-   char *r;
-   const char *s;
-
-   result = (char*)xrealloc (result, 2 + 2 * strlen (string) + 1);
-
-   r = result;
-   if(string[0]=='-' || string[0]=='~')
-   {
-      *r++='.';
-      *r++='/';
-   }
-
-   for (s = string; s && (c = *s); s++)
-   {
-      switch (c)
-      {
-      case '\'':
-      case '(': case ')':
-      case '!': case '{': case '}':		/* reserved words */
-      case '^':
-      case '$': case '`':			/* expansion chars */
-      case '*': case '[': case '?': case ']':	/* globbing chars */
-      case ' ': case '\t': case '\n':		/* IFS white space */
-      case '"': case '\\':		/* quoting chars */
-      case '|': case '&': case ';':		/* shell metacharacters */
-      case '<': case '>':
-      case '#':				/* comment char */
-	 *r++ = '\\';
-	 *r++ = c;
-	 break;
-      default:
-	 *r++ = c;
-	 break;
-      }
-   }
-
-   *r = '\0';
-   return (result);
-}
-
 void Fish::SendArrayInfoRequests()
 {
    for(int i=array_ptr; i<array_cnt; i++)
@@ -628,7 +533,7 @@ void Fish::SendMethod()
       break;
    case QUOTE_CMD:
       Send("#EXEC %s\n"
-	   "%s; echo '### 200'\n",file,file);
+	   "%s; echo '### 200'\n",e,file.get());
       PushExpect(EXPECT_QUOTE);
       real_pos=0;
       break;
@@ -663,8 +568,9 @@ int Fish::HandleReplies()
       }
       if(recv_buf->Eof())
       {
-	 DebugPrint("**** ",_("Peer closed connection"),0);
-	 if(!RespQueueIsEmpty() && RespQueue[RQ_head]==EXPECT_CWD && message)
+	 LogError(0,_("Peer closed connection"));
+	 // Solaris' shell exists when is given with wrong directory
+	 if(!RespQueueIsEmpty() && RespQueue[0]==EXPECT_CWD && message)
 	    SetError(NO_FILE,message);
 	 Disconnect();
 	 m=MOVED;
@@ -717,9 +623,7 @@ int Fish::HandleReplies()
    }
    m=MOVED;
    s=eol-b+1;
-   xfree(line);
-   line=(char*)xmemdup(b,s);
-   line[s-1]=0;
+   line.nset(b,s-1);
    recv_buf->Skip(s);
 
    int code=-1;
@@ -728,7 +632,7 @@ int Fish::HandleReplies()
 	 code=-1;
    }
 
-   DebugPrint("<--- ",line,ReplyLogPriority(code));
+   LogRecv(ReplyLogPriority(code),line);
    if(code==-1)
    {
       if(!received_greeting && !strcmp(line,"FISH:"))
@@ -737,28 +641,22 @@ int Fish::HandleReplies()
 	 return m;
       }
       if(message==0)
-	 message=xstrdup(line);
+	 message.set(line);
       else
-      {
-	 message=(char*)xrealloc(message,xstrlen(message)+s+1);
-	 strcat(message,"\n");
-	 strcat(message,line);
-      }
+	 message.vappend("\n",line.get(),NULL);
       return m;
    }
 
    if(RespQueueIsEmpty())
    {
-      DebugPrint("**** ",_("extra server response"),3);
-      xfree(message);
-      message=0;
+      LogError(3,_("extra server response"));
+      message.set(0);
       return m;
    }
-   expect_t &e=RespQueue[RQ_head];
-   RQ_head++;
+   expect_t e=RespQueue.next();
 
    bool keep_message=false;
-   char *p;
+   xstring p;
 
    switch(e)
    {
@@ -769,15 +667,15 @@ int Fish::HandleReplies()
    case EXPECT_PWD:
       if(!message)
 	 break;
-      home_auto=xstrdup(message);
-      Log::global->Format(9,"---- home set to %s\n",home_auto);
+      home_auto.set(message);
+      Log::global->Format(9,"---- home set to %s\n",home_auto.get());
       PropagateHomeAuto();
       if(!home)
 	 set_home(home_auto);
       cache->SetDirectory(this, home, true);
       break;
    case EXPECT_CWD:
-      p=PopDirectory();
+      PopDirectory(&p);
       if(message==0)
       {
 	 set_real_cwd(p);
@@ -790,7 +688,6 @@ int Fish::HandleReplies()
       }
       else
 	 SetError(NO_FILE,message);
-      xfree(p);
       break;
    case EXPECT_RETR_INFO:
       if(message && is_ascii_digit(message[0]) && !strchr(message,':'))
@@ -871,32 +768,18 @@ int Fish::HandleReplies()
    }
 
    if(!keep_message)
-   {
-      xfree(message);
-      message=0;
-   }
+      message.set(0);
 
    return m;
 }
 void Fish::PushExpect(expect_t e)
 {
-   int newtail=RQ_tail+1;
-   if(newtail>RQ_alloc)
-   {
-      if(RQ_head-0<newtail-RQ_alloc)
-	 RespQueue=(expect_t*)
-	    xrealloc(RespQueue,(RQ_alloc=newtail+16)*sizeof(*RespQueue));
-      memmove(RespQueue,RespQueue+RQ_head,(RQ_tail-RQ_head)*sizeof(*RespQueue));
-      RQ_tail=0+(RQ_tail-RQ_head);
-      RQ_head=0;
-      newtail=RQ_tail+1;
-   }
-   RespQueue[RQ_tail]=e;
-   RQ_tail=newtail;
+   RespQueue.push(e);
 }
 void Fish::CloseExpectQueue()
 {
-   for(int i=RQ_head; i<RQ_tail; i++)
+   int count=RespQueue.count();
+   for(int i=0; i<count; i++)
    {
       switch(RespQueue[i])
       {
@@ -922,19 +805,6 @@ void Fish::CloseExpectQueue()
    }
 }
 
-void Fish::PushDirectory(const char *p)
-{
-   path_queue=(char**)xrealloc(path_queue,++path_queue_len*sizeof(*path_queue));
-   path_queue[path_queue_len-1]=xstrdup(p);
-}
-char *Fish::PopDirectory()
-{
-   assert(path_queue_len>0);
-   char *p=path_queue[0];
-   memmove(path_queue,path_queue+1,--path_queue_len*sizeof(*path_queue));
-   return p; // caller should free it.
-}
-
 const char *memstr(const char *mem,size_t len,const char *str)
 {
    size_t str_len=strlen(str);
@@ -1211,7 +1081,7 @@ FileAccess *Fish::New() { return new Fish(); }
 
 DirList *Fish::MakeDirList(ArgV *args)
 {
-   return new FishDirList(args,this);
+   return new FishDirList(this,args);
 }
 #include "FileGlob.h"
 Glob *Fish::MakeGlob(const char *pattern)
@@ -1258,7 +1128,7 @@ int FishDirList::Do()
       else
       {
 	 session->Open(pattern,FA::LONG_LIST);
-	 ((Fish*)session)->DontEncodeFile();
+	 session.Cast<Fish>()->DontEncodeFile();
 	 ubuf=new IOBufferFileAccess(session);
 	 if(FileAccess::cache->IsEnabled(session->GetHostName()))
 	    ubuf->Save(FileAccess::cache->SizeLimit());
@@ -1292,20 +1162,6 @@ int FishDirList::Do()
    return m;
 }
 
-FishDirList::FishDirList(ArgV *a,FileAccess *fa)
-   : DirList(a)
-{
-   session=fa;
-   ubuf=0;
-   pattern=args->Combine(1);
-}
-
-FishDirList::~FishDirList()
-{
-   Delete(ubuf);
-   xfree(pattern);
-}
-
 const char *FishDirList::Status()
 {
    static char s[256];
