@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FtpDirList.cc,v 1.27 2006/06/13 14:35:36 lav Exp $ */
+/* $Id: FtpDirList.cc,v 1.30 2007/08/13 18:58:26 lav Exp $ */
 
 #include <config.h>
 
@@ -119,20 +119,6 @@ int FtpDirList::Do()
    return m;
 }
 
-FtpDirList::FtpDirList(ArgV *a,FileAccess *fa)
-   : DirList(a)
-{
-   session=fa;
-   ubuf=0;
-   pattern=args->Combine(1);
-}
-
-FtpDirList::~FtpDirList()
-{
-   Delete(ubuf);
-   xfree(pattern);
-}
-
 const char *FtpDirList::Status()
 {
    static char s[256];
@@ -231,7 +217,11 @@ bool FtpDirList::TryColor(const char *line_c,int len)
       group[0]=0;
       n=sscanf(line,"%11s %d %31s %lld %3s %2d %5s%n",perms,&nlink,
 	    user,&size,month_name,&day,year_or_time,&consumed);
+      if(n!=7)
+	 return false;
    }
+   else if(n!=8)
+      return false;
    if(consumed>0 && -1!=(parse_perms(perms+1))
    && -1!=(parse_month(month_name))
    && -1!=parse_year_or_time(year_or_time,&year,&hour,&minute)
