@@ -82,36 +82,32 @@
 #include "misc.h"
 #include "LsCache.h"
 
-GetFileInfo::GetFileInfo(FileAccess *a, const char *_dir, bool _showdir)
-   : ListInfo(a,0)
+GetFileInfo::GetFileInfo(const FileAccessRef& a, const char *_dir, bool _showdir)
+   : ListInfo(0,0), session(a), dir(_dir?_dir:""), origdir(a->GetCwd())
 {
-  dir=xstrdup(_dir? _dir:"");
-
    showdir=_showdir;
    state=INITIAL;
    tried_dir=tried_file=tried_info=false;
    result=0;
-   path_to_prefix=0;
-   verify_fn=0;
    li=0;
    from_cache=0;
-   saved_error_text=0;
-   was_directory=0;
+   was_directory=false;
    prepend_path=true;
 
-   origdir=xstrdup(session->GetCwd());
+   const char *bn=basename_ptr(dir);
+   if(bn[0]=='.' && (bn[1]==0 || bn[1]=='/' ||
+                     (bn[1]=='.' && (bn[2]==0 || bn[2]=='/')))
+   || bn[0]=='/')
+   {
+      // . .. / are directories, don't try them as a file.
+      tried_file=true;
+   }
 }
 
 GetFileInfo::~GetFileInfo()
 {
    session->Close();
    session->SetCwd(origdir);
-   Delete(li);
-   xfree(saved_error_text);
-   xfree(dir);
-   xfree(path_to_prefix);
-   xfree(origdir);
-   xfree(verify_fn);
 }
 
 int GetFileInfo::Do()
@@ -130,7 +126,12 @@ int GetFileInfo::Do()
 
       if(use_cache)
       {
-	 switch(LsCache::IsDirectory(session,dir))
+	 int is_dir=-1;
+	 if(last_char(dir)=='/')
+	    is_dir=1;
+	 else
+	    is_dir=FileAccess::cache->IsDirectory(session,dir);
+	 switch(is_dir)
 	 {
 	 case 0:
 	    tried_dir = true; /* it's a file */
@@ -171,8 +172,7 @@ int GetFileInfo::Do()
 	  * if we are going to show its contents */
 	 tried_dir=true;
 	 cd_path = dir;
-	 xfree(path_to_prefix);
-	 path_to_prefix=xstrdup(dir);
+	 path_to_prefix.set(dir);
 	 was_directory=true;
       }
       else if(!tried_file)
@@ -185,8 +185,7 @@ int GetFileInfo::Do()
 	 session->Chdir(dir, false);
 	 cd_path = "..";
 
-	 xfree(path_to_prefix);
-	 path_to_prefix=dirname_alloc(dir);
+	 path_to_prefix.set(dirname(dir));
 	 was_directory=false;
       }
       else if(!tried_info)
@@ -197,8 +196,7 @@ int GetFileInfo::Do()
 	 session->Chdir(dir, false);
 	 session->Chdir("..", false);
 
-	 xfree(path_to_prefix);
-	 path_to_prefix=dirname_alloc(dir);
+	 path_to_prefix.set(dirname(dir));
 	 was_directory=false;
 
 	 /* We tried both; no luck. Fall back on ARRAY_INFO. */
@@ -237,7 +235,7 @@ int GetFileInfo::Do()
 	  * Only save the first error, so error texts contain the full
 	  * path. */
 	 if(!saved_error_text)
-	    saved_error_text = xstrdup(session->StrError(res));
+	    saved_error_text.set(session->StrError(res));
 	 session->Close();
 	 if(res==FA::NO_FILE)
 	 {
@@ -255,19 +253,18 @@ int GetFileInfo::Do()
       }
       session->Close();
       if(!from_cache)
-	 LsCache::SetDirectory(session,"",true);
+	 FileAccess::cache->SetDirectory(session,"",true);
 
       /* Now that we've connected, we should have the home directory path. Find out
        * the real name of the path.  (We may have something like "~/..".) */
       if(!verify_fn)
       {
-	 char *pwd = alloca_strdup(session->GetCwd());
+	 FileAccess::Path pwd = session->GetCwd();
 
 	 session->SetCwd(origdir);
 	 session->Chdir(dir, false);
 
-	 xfree(verify_fn);
-	 verify_fn = xstrdup(basename_ptr(session->GetCwd()));
+	 verify_fn.set(basename_ptr(session->GetCwd()));
 
 	 /* go back */
 	 session->SetCwd(pwd);
@@ -281,6 +278,9 @@ int GetFileInfo::Do()
 
 	 result = new FileSet;
 	 result->Add(fi);
+
+	 path_to_prefix.set(dirname(dir));
+
 	 state=DONE;
 	 return MOVED;
       }
@@ -293,8 +293,7 @@ int GetFileInfo::Do()
 	  * directory index. */
 	 FileInfo *fi = new FileInfo(dir);
 	 fi->SetType(fi->DIRECTORY);
-	 xfree(path_to_prefix);
-	 path_to_prefix=dirname_alloc(dir);
+	 path_to_prefix.set(dirname(dir));
 
 	 result = new FileSet;
 	 result->Add(fi);
@@ -303,13 +302,12 @@ int GetFileInfo::Do()
       }
 
       /* Get a listing: */
-      Delete(li);
       li=session->MakeListInfo();
       if(follow_symlinks) li->FollowSymlinks();
       li->UseCache(use_cache);
       li->NoNeed(FileInfo::ALL_INFO); /* clear need */
       li->Need(need);
-      SetExclude(exclude_prefix, exclude);
+      li->SetExclude(exclude_prefix, exclude);
       state=GETTING_LIST;
       m=MOVED;
 
@@ -324,7 +322,7 @@ int GetFileInfo::Do()
 	 }
 
 	 if(!saved_error_text)
-	    saved_error_text = xstrdup(li->ErrorText());
+	    saved_error_text.set(li->ErrorText());
 
 	 /* Otherwise, go on to try the next mode. */
 	 state=CHANGE_DIR;
@@ -339,20 +337,18 @@ int GetFileInfo::Do()
 
       /* Got the list.  Steal it from the listinfo: */
       result=li->GetResult();
-      Delete(li); li=0;
+      li=0;
 
       /* If this was a listing of the basename: */
       if(!was_directory) {
-	 int len=strlen(verify_fn);
-	 while(len>0 && verify_fn[len-1] == '/')
-	    verify_fn[--len] = 0;
+	 verify_fn.rtrim('/');
 
 	 /* Find the file with our filename: */
 	 const FileInfo *file = result->FindByName(verify_fn);
 
 	 if(!file) {
 	    /* It doesn't exist, or we have no result (failed). */
-	    delete result; result=0;
+	    result=0;
 	    tried_file=true;
 	    from_cache=false;
 	    state=CHANGE_DIR;
@@ -363,7 +359,7 @@ int GetFileInfo::Do()
 	  * directory, we should have been able to Chdir into it to begin
 	  * with.  We probably got Access Denied.  Fail. */
 	 if(!showdir && (file->defined&file->TYPE) && file->filetype==FileInfo::DIRECTORY) {
-	    delete result; result=0;
+	    result=0;
 	    if(saved_error_text)
 	    {
 	       SetError(saved_error_text);
@@ -375,12 +371,9 @@ int GetFileInfo::Do()
 	    return MOVED;
 	 }
 
-	 FileSet *newresult=new FileSet();
 	 FileInfo *copy = new FileInfo(*file);
-
-	 newresult->Add(copy);
-	 delete result;
-	 result=newresult;
+	 result=new FileSet();
+	 result->Add(copy);
       }
 
       goto done;
@@ -412,7 +405,7 @@ int GetFileInfo::Do()
       if(res < 0)
       {
 	 if(!saved_error_text)
-	    saved_error_text = xstrdup(session->StrError(res));
+	    saved_error_text.set(session->StrError(res));
 	 state=CHANGE_DIR;
 	 return MOVED;
       }
@@ -456,11 +449,11 @@ done:
    case DONE:
       if(!done)
       {
-	 if(result && showdir && result->get_fnum())
+	 if(result && showdir && result->get_fnum()>0)
 	 {
 	    FileInfo *f = (*result)[0];
 	    /* Make sure the filename is what was requested (ie ".."). */
-	    char *fn = basename_ptr(dir);
+	    const char *fn = basename_ptr(dir);
 	    f->SetName(*fn? fn:".");
 
 	    /* If we're in show_dir mode, was_directory will always be false;
