@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: GetJob.cc,v 1.19 2001/10/25 10:46:34 lav Exp $ */
+/* $Id: GetJob.cc,v 1.28 2007/04/23 07:11:45 lav Exp $ */
 
 #include <config.h>
 #include <sys/types.h>
@@ -56,30 +56,17 @@ int   GetJob::Do()
    return m;
 }
 
-FileCopyPeer *GetJob::NoProtoSrc(const char *src)
+FileCopyPeer *GetJob::SrcLocal(const char *src)
 {
-   if(reverse)
-   {
-      const char *f=(cwd && src[0]!='/') ? dir_file(cwd,src) : src;
-      return FileCopyPeerFDStream::NewGet(f);
-   }
-
-   FileCopyPeerFA *s=new FileCopyPeerFA(session,src,FA::RETRIEVE);
-   s->DontReuseSession();
-   return s;
+   const char *f=(cwd && src[0]!='/') ? dir_file(cwd,src) : src;
+   return FileCopyPeerFDStream::NewGet(f);
 }
-FileCopyPeer *GetJob::NoProtoDst(const char *dst)
+FileCopyPeer *GetJob::DstLocal(const char *dst)
 {
-   if(reverse)
-   {
-      FileCopyPeerFA *s=new FileCopyPeerFA(session,dst,FA::STORE);
-      s->DontReuseSession();
-      return s;
-   }
-
-   int flags=O_WRONLY|O_CREAT|(cont?0:O_TRUNC);
+   int flags=O_WRONLY|O_CREAT|(truncate_target_first?O_TRUNC:0);
+   dst=expand_home_relative(dst);
    const char *f=(cwd && dst[0]!='/') ? dir_file(cwd,dst) : dst;
-   if(!cont && res_make_backup.QueryBool(0))
+   if(truncate_target_first && res_make_backup.QueryBool(0))
    {
       /* rename old file if exists and size>0 */
       struct stat st;
@@ -94,40 +81,58 @@ FileCopyPeer *GetJob::NoProtoDst(const char *dst)
 	       count++;
 	       return 0;
 	    }
-	    backup_file=xstrdup(f,+1);
-	    strcat(backup_file,"~");
+	    backup_file.vset(f,"~",NULL);
 	    if(rename(f,backup_file)!=0)
-	    {
-	       xfree(backup_file);
-	       backup_file=0;
-	    }
+	       backup_file.set(0);
 	    else
-	    {
 	       file_mode=st.st_mode;
-	    }
 	 }
       }
    }
    local=new FileStream(f,flags); // local is for pget.
    FileCopyPeerFDStream *p=new FileCopyPeerFDStream(local,FileCopyPeer::PUT);
-   p->DontDeleteStream();
    return p;
 }
 
+bool GetJob::IsRemoteNonURL(const ParsedURL &url,FA::open_mode mode)
+{
+   // store & put || !store & get
+   return (!url.proto && (mode==FA::STORE ^ !reverse));
+}
+bool GetJob::IsLocal(const ParsedURL &url)
+{
+   return !url.proto || !strcasecmp(url.proto,"file");
+}
+FileCopyPeer *GetJob::CreateCopyPeer(FileAccess *session,const char *path,FA::open_mode mode)
+{
+   ParsedURL url(path,true);
+   if(IsRemoteNonURL(url,mode))
+      return new FileCopyPeerFA(session,path,mode);
+   Delete(session);	// delete cloned session.
+   return CreateCopyPeer(url,mode);
+}
+FileCopyPeer *GetJob::CreateCopyPeer(const FileAccessRef& session,const char *path,FA::open_mode mode)
+{
+   ParsedURL url(path,true);
+   if(IsRemoteNonURL(url,mode))
+      return new FileCopyPeerFA(session,path,mode);
+   return CreateCopyPeer(url,mode);
+}
+FileCopyPeer *GetJob::CreateCopyPeer(const ParsedURL &url,FA::open_mode mode)
+{
+   if(IsLocal(url))
+      return (mode==FA::STORE)
+	 ? DstLocal(url.path)
+	 : SrcLocal(url.path);
+   return new FileCopyPeerFA(&url,mode);
+}
+
 void GetJob::NextFile()
 {
 try_next:
    file_mode=NO_MODE;
-   if(backup_file)
-   {
-      xfree(backup_file);
-      backup_file=0;
-   }
-   if(local)
-   {
-      delete local;
-      local=0;
-   }
+   backup_file.set(0);
+   local=0;
 
    if(!args)
       return;
@@ -140,29 +145,17 @@ try_next:
       return;
    }
 
-   ParsedURL src_url(src,true);
-   ParsedURL dst_url(dst,true);
-
-   FileCopyPeer *src_peer=0;
-   FileCopyPeer *dst_peer=0;
-
-   if(dst_url.proto==0)
-      dst_peer=NoProtoDst(dst);
-   else
-      dst_peer=new FileCopyPeerFA(&dst_url,FA::STORE);
-
+   FileCopyPeer *dst_peer=CreateCopyPeer(session,dst,FA::STORE);
    if(!dst_peer)
       goto try_next;
-
-   if(src_url.proto==0)
-      src_peer=NoProtoSrc(src);
-   else
-      src_peer=new FileCopyPeerFA(&src_url,FA::RETRIEVE);
+   FileCopyPeer *src_peer=CreateCopyPeer(session,src,FA::RETRIEVE);
 
    FileCopy *c=FileCopy::New(src_peer,dst_peer,cont);
 
    if(delete_files)
       c->RemoveSourceLater();
+   if(remove_target_first)
+      c->RemoveTargetFirst();
 
    SetCopier(c,src);
 }
@@ -172,8 +165,7 @@ void GetJob::RemoveBackupFile()
    if(backup_file)
    {
       remove(backup_file);
-      xfree(backup_file);
-      backup_file=0;
+      backup_file.set(0);
    }
 }
 
@@ -181,14 +173,8 @@ GetJob::GetJob(FileAccess *s,ArgV *a,bool c)
    : CopyJobEnv(s,a,c)
 {
    delete_files=false;
-   backup_file=0;
+   remove_target_first=false;
+   truncate_target_first=!cont;
    file_mode=NO_MODE;
-   local=0;
    reverse=false;
 }
-GetJob::~GetJob()
-{
-   xfree(backup_file);
-   if(local)
-      delete local;
-}
