@@ -21,9 +21,9 @@
 #ifndef GETPASS_H
 #define GETPASS_H
 
-#include <stdio.h>
+#include "trio.h"
 
-char *GetPass(const char *p);
-char *readline_from_file(FILE *f);
+const char *GetPass(const char *p);
+char *readline_from_file(int fd);
 
 #endif//GETPASS_H
