@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1999 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1999-2006 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,13 +18,14 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: Http.h,v 1.45 2005/03/05 09:10:45 lav Exp $ */
+/* $Id: Http.h,v 1.54 2007/08/24 15:26:36 lav Exp $ */
 
 #ifndef HTTP_H
 #define HTTP_H
 
 #include "NetAccess.h"
 #include "buffer.h"
+#include "lftp_ssl.h"
 
 class Http : public NetAccess
 {
@@ -51,14 +52,13 @@ class Http : public NetAccess
 
    void	Send(const char *format,...) PRINTF_LIKE(2,3);
 
-   IOBuffer *send_buf;
-   IOBuffer *recv_buf;
-   bool recv_buf_suspended;
+   SMTaskRef<IOBuffer> send_buf;
+   SMTaskRef<IOBuffer> recv_buf;
    void SendMethod(const char *,const char *);
    const char *last_method;
    enum { HTTP_NONE=0, HTTP_POST, HTTP_MOVE, HTTP_COPY } special;
-   char *special_data;
-   void DirFile(char *path_base,const char *ecwd,const char *efile);
+   xstring special_data;
+   void DirFile(xstring& path,const char *ecwd,const char *efile);
    void SendAuth();
    void SendCacheControl();
    void SendBasicAuth(const char *tag,const char *auth);
@@ -73,15 +73,15 @@ class Http : public NetAccess
    void HandleHeaderLine(const char *name,const char *value);
    void GetBetterConnection(int level);
    void SetCookie(const char *val);
-   char *MakeCookie(const char *host,const char *path);
-   void CookieMerge(char **c,const char *add);
+   void MakeCookie(xstring &cookie,const char *host,const char *path);
+   void CookieMerge(xstring &c,const char *add);
    bool CookieClosureMatch(const char *closure,const char *host,const char *path);
 
    int sock;
    void Disconnect();
    void ResetRequestData();
    void MoveConnectionHere(Http *o);
-   int IsConnected()
+   int IsConnected() const
       {
 	 if(sock==-1)
 	    return 0;
@@ -89,12 +89,16 @@ class Http : public NetAccess
 	    return 1;
 	 return 2;
       }
+#if USE_SSL
+   Ref<lftp_ssl> ssl;
    void MakeSSLBuffers();
+#endif
+   void LogErrorText();
 
-   char *status;
-   int   status_consumed;
+   xstring status;
+   int status_consumed;
    int proto_version;
-   char *line;
+   xstring line;
    off_t body_size;
    off_t bytes_received;
    bool sent_eot;
@@ -117,8 +121,12 @@ class Http : public NetAccess
    bool no_cache;
    bool no_cache_this;
 
+   bool use_propfind_now;
+
    const char *user_agent;
 
+   int _Read(void *,int);  // does not update pos, rate_limit, retries, does not check state.
+
 protected:
    bool hftp;  // ftp over http proxy.
    bool https; // secure http
@@ -131,11 +139,11 @@ public:
    Http(const Http *);
    ~Http();
 
-   const char *GetProto() { return "http"; }
+   const char *GetProto() const { return "http"; }
 
-   FileAccess *Clone() { return new Http(this); }
+   FileAccess *Clone() const { return new Http(this); }
    static FileAccess *New();
-   FileSet *ParseLongList(const char *buf,int len,int *err=0);
+   FileSet *ParseLongList(const char *buf,int len,int *err=0) const;
 
    int Do();
    int Done();
@@ -152,8 +160,8 @@ public:
 
    void Reconfig(const char *name=0);
 
-   bool SameSiteAs(FileAccess *fa);
-   bool SameLocationAs(FileAccess *fa);
+   bool SameSiteAs(const FileAccess *fa) const;
+   bool SameLocationAs(const FileAccess *fa) const;
 
    DirList *MakeDirList(ArgV *a);
    Glob *MakeGlob(const char *pattern);
@@ -163,8 +171,8 @@ public:
 
    bool NeedSizeDateBeforehand() { return true; }
 
-   void Suspend();
-   void Resume();
+   void SuspendInternal();
+   void ResumeInternal();
 
    void Cleanup();
    void CleanupThis();
@@ -179,9 +187,9 @@ public:
    HFtp(const HFtp *);
    ~HFtp();
 
-   const char *GetProto() { return "hftp"; }
+   const char *GetProto() const { return "hftp"; }
 
-   FileAccess *Clone() { return new HFtp(this); }
+   FileAccess *Clone() const { return new HFtp(this); }
    static FileAccess *New();
 
    virtual void Login(const char *,const char *);
@@ -195,9 +203,9 @@ public:
    Https(const Https *);
    ~Https();
 
-   const char *GetProto() { return "https"; }
+   const char *GetProto() const { return "https"; }
 
-   FileAccess *Clone() { return new Https(this); }
+   FileAccess *Clone() const { return new Https(this); }
    static FileAccess *New();
 };
 
