@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-2005 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2007 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: MirrorJob.h,v 1.45 2006/02/06 11:00:40 lav Exp $ */
+/* $Id: MirrorJob.h,v 1.52 2007/09/17 10:39:30 lav Exp $ */
 
 #ifndef MIRRORJOB_H
 #define MIRRORJOB_H
@@ -42,37 +42,38 @@ class MirrorJob : public Job
       TARGET_REMOVE_OLD_FIRST,
       TARGET_CHMOD,
       FINISHING,
+      LAST_EXEC,
       DONE
    };
    state_t state;
 
-   FileAccess *source_session;
-   FileAccess *target_session;
+   FileAccessRef source_session;
+   FileAccessRef target_session;
    bool target_is_local;
    bool source_is_local;
 
-   FileSet *target_set;
-   FileSet *source_set;
-   FileSet *to_transfer;
-   FileSet *same;
-   FileSet *to_rm;
-   FileSet *to_rm_mismatched;
-   FileSet *old_files_set;
-   FileSet *new_files_set;
-   void	 InitSets(FileSet *src,FileSet *dst);
+   Ref<FileSet> target_set;
+   Ref<FileSet> source_set;
+   Ref<FileSet> to_transfer;
+   Ref<FileSet> same;
+   Ref<FileSet> to_rm;
+   Ref<FileSet> to_rm_mismatched;
+   Ref<FileSet> old_files_set;
+   Ref<FileSet> new_files_set;
+   void	 InitSets(const FileSet *src,const FileSet *dst);
 
    void	 HandleFile(FileInfo *);
 
    bool create_target_dir;
    bool	no_target_dir;	   // target directory does not exist (for script_only)
 
-   ListInfo *source_list_info;
-   ListInfo *target_list_info;
+   Ref<ListInfo> source_list_info;
+   Ref<ListInfo> target_list_info;
 
-   char	 *source_dir;
-   char	 *source_relative_dir;
-   char	 *target_dir;
-   char	 *target_relative_dir;
+   xstring_c source_dir;
+   xstring_c source_relative_dir;
+   xstring_c target_dir;
+   xstring_c target_relative_dir;
 
    struct Statistics
    {
@@ -95,7 +96,8 @@ class MirrorJob : public Job
    int	 flags;
    int	 max_error_count;
 
-   PatternSet *exclude;
+   Ref<PatternSet> my_exclude;
+   const PatternSet *exclude;
 
    bool	 create_remote_dir;
 
@@ -106,30 +108,32 @@ class MirrorJob : public Job
 
    time_t newer_than;
    time_t older_than;
-   Range *size_range;
+   Ref<Range> my_size_range;
+   const Range *size_range;
 
-   char *script_name;
+   xstring_c script_name;
    FILE *script;
    bool script_only;
    bool script_needs_closing;
-
    bool use_cache;
-
    bool remove_source_files;
+   bool skip_noaccess;
 
    int parallel;
    int pget_n;
    int pget_minchunk;
 
+   xstring_c on_change;
+
    mode_t get_mode_mask();
 
    int source_redirections;
    int target_redirections;
 
-   void HandleChdir(FileAccess * &session, int &redirections);
-   void HandleListInfoCreation(FileAccess * &session,ListInfo * &list_info,
+   void HandleChdir(FileAccessRef& session, int &redirections);
+   void HandleListInfoCreation(const FileAccessRef& session,Ref<ListInfo>& list_info,
 	    const char *relative_dir);
-   void HandleListInfo(ListInfo * &list_info, FileSet * &set);
+   void HandleListInfo(Ref<ListInfo>& list_info,Ref<FileSet>& set);
 
 public:
    enum
@@ -148,16 +152,23 @@ public:
       REMOVE_FIRST=1<<11,
       IGNORE_SIZE=1<<12,
       NO_SYMLINKS=1<<13,
-      LOOP=1<<14
+      LOOP=1<<14,
+      ONLY_EXISTING=1<<15,
+      NO_EMPTY_DIRS=1<<16,
+      DEPTH_FIRST=1<<17,
+      ASCII=1<<18,
    };
 
-   void SetFlags(int f,int v)
+   void SetFlags(int f,bool v)
    {
       if(v)
 	 flags|=f;
       else
 	 flags&=~f;
    }
+   bool FlagsSet(int f)	   { return (flags&f)==f; }
+   bool FlagSet(int f)	   { return (flags&f); }
+   bool AnyFlagSet(int f)  { return (flags&f); }
 
    MirrorJob(MirrorJob *parent,FileAccess *f,FileAccess *target,
       const char *new_source_dir,const char *new_target_dir);
@@ -165,13 +176,15 @@ public:
 
    int	 Do();
    int	 Done() { return state==DONE; }
-   void	 ShowRunStatus(StatusLine *);
+   void	 ShowRunStatus(const SMTaskRef<StatusLine>&);
    void	 PrintStatus(int v,const char *);
    void	 SayFinal() { PrintStatus(0,""); }
    int	 ExitCode() { return stats.error_count; }
 
-   void	 SetExclude(PatternSet *x) { exclude=x; }
-   void	 SetSizeRange(Range *r) { size_range=r; }
+   void	 SetExclude(PatternSet *x) { my_exclude=x; exclude=my_exclude; }
+   void	 SetExclude(const PatternSet *x) { exclude=x; }
+   void	 SetSizeRange(Range *r) { my_size_range=r; size_range=my_size_range; }
+   void	 SetSizeRange(const Range *r) { size_range=r; }
 
    void	 SetVerbose(int v) { verbose_report=v; }
 
@@ -182,6 +195,7 @@ public:
 
    void  UseCache(bool u) { use_cache=u; }
    void	 RemoveSourceFiles() { remove_source_files=true; }
+   void	 SkipNoAccess() { skip_noaccess=true; }
 
    void  SetParallel(int p) { parallel=p; }
    void  SetPGet(int n) { pget_n=n; }
@@ -195,6 +209,7 @@ public:
 	 script_only=yes;
       }
    void SetMaxErrorCount(int ec) { max_error_count=ec; }
+   void SetOnChange(const char *oc);
 };
 
 #endif//MIRRORJOB_H
