@@ -26,23 +26,24 @@
 class QueueFeeder : public CmdFeeder
 {
    struct QueueJob {
-      char *cmd;
-      char *pwd, *lpwd;
+      xstring_c cmd;
+      xstring_c pwd;
+      xstring_c lpwd;
 
       QueueJob *next, *prev;
-      QueueJob(): cmd(0), pwd(0), lpwd(0), next(0), prev(0) { }
-      ~QueueJob();
+      QueueJob(): next(0), prev(0) {}
    } *jobs, *lastjob;
-   char *cur_pwd, *cur_lpwd;
+   xstring_c cur_pwd;
+   xstring_c cur_lpwd;
 
-   char *buffer;
+   xstring buffer;
 
    /* remove the given job from the list */
    void unlink_job(QueueJob *job);
-      
+
    /* get the n'th job */
    QueueJob *get_job(int n);
-      
+
    /* get the n'th job, removed from the list: */
    QueueJob *grab_job(int n);
 
@@ -50,34 +51,35 @@ class QueueFeeder : public CmdFeeder
     * that match the cmd: */
    QueueJob *grab_job(const char *cmd);
 
+   /* get the next job in j that matches cmd (including j) */
+   static QueueJob *get_next_match(const char *cmd, QueueJob *j);
+   void PrintJobs(const QueueJob *job, int v, const char *plur) const;
+
    void insert_jobs(QueueJob *job,
 		   QueueJob *&lst_head, QueueJob *&lst_tail,
 		   QueueJob *before);
-      
+
    void FreeList(QueueJob *j);
 
 public:
    const char *NextCmd(CmdExec *exec,const char *prompt);
 
    /* Add a command to the queue at a given position; a 0 position inserts at the end. */
-   void QueueCmd(const char *cmd, const char *pwd, const char *lpwd, int pos = 0);
+   void QueueCmd(const char *cmd, const char *pwd, const char *lpwd, int pos = 0, int verbose = 0);
 
    /* delete jobs (by index or wildcard expr) */
-   bool DelJob(int from);
-   bool DelJob(const char *cmd);
+   bool DelJob(int from, int v = 0);
+   bool DelJob(const char *cmd, int v = 0);
 
    /* move one or more jobs (by index or wildcard expr). */
-   bool MoveJob(int from, int to);
-   bool MoveJob(const char *cmd, int to);
+   bool MoveJob(int from, int to, int v = 0);
+   bool MoveJob(const char *cmd, int to, int v = 0);
 
-   void PrintStatus(int v) const;
+   enum { PrintRequeue = 9999 };
+   void PrintStatus(int v,const char *prefix="\t") const;
 
    QueueFeeder(const char *pwd, const char *lpwd):
-      jobs(0), lastjob(0), cur_pwd(0), cur_lpwd(0), buffer(0)
-   {
-      cur_pwd = xstrdup(pwd);
-      cur_lpwd = xstrdup(lpwd);
-   }
+      jobs(0), lastjob(0), cur_pwd(pwd), cur_lpwd(lpwd) {}
    virtual ~QueueFeeder();
 };
 
