@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: Resolver.cc,v 1.70 2006/06/13 14:35:40 lav Exp $ */
+/* $Id: Resolver.cc,v 1.78 2007/10/08 13:29:14 lav Exp $ */
 
 #include <config.h>
 
@@ -44,7 +44,6 @@
 #endif
 
 #include "xstring.h"
-#include "xmalloc.h"
 #include "ResMgr.h"
 #include "log.h"
 #include "plural.h"
@@ -95,24 +94,14 @@ ResolverCache *Resolver::cache;
 
 Resolver::Resolver(const char *h,const char *p,const char *defp,
 		   const char *ser,const char *pr)
+   : hostname(h), portname(p), service(ser), proto(pr), defport(defp)
 {
-   hostname=xstrdup(h);
-   portname=xstrdup(p);
    port_number=0;
 
-   service=xstrdup(ser);
-   proto=xstrdup(pr);
-   defport=xstrdup(defp);
-
    pipe_to_child[0]=pipe_to_child[1]=-1;
-   w=0;
-   err_msg=0;
    done=false;
    timeout_timer.SetResource("dns:fatal-timeout",hostname);
    Reconfig();
-   addr=0;
-   addr_num=0;
-   buf=0;
    use_fork=ResMgr::QueryBool("dns:use-fork",0);
 
    error=0;
@@ -127,20 +116,11 @@ Resolver::~Resolver()
    if(pipe_to_child[1]!=-1)
       close(pipe_to_child[1]);
 
-   xfree(hostname);
-   xfree(portname);
-   xfree(service);
-   xfree(proto);
-   xfree(defport);
-
-   xfree(err_msg);
-   xfree(addr);
    if(w)
    {
       w->Kill(SIGKILL);
-      w->Auto();
+      w.borrow()->Auto();
    }
-   Delete(buf);
 }
 
 int   Resolver::Do()
@@ -157,10 +137,8 @@ int   Resolver::Do()
       cache->Find(hostname,portname,defport,service,proto,&a,&n);
       if(a && n>0)
       {
-	 Log::global->Write(10,"dns cache hit\n");
-	 addr_num=n;
-	 addr=(sockaddr_u*)xmalloc(n*sizeof(*addr));
-	 memcpy(addr,a,n*sizeof(*addr));
+	 FA::LogNote(10,"dns cache hit");
+	 addr.nset(a,n);
 	 done=true;
 	 return MOVED;
       }
@@ -183,7 +161,7 @@ int   Resolver::Do()
 	 fcntl(pipe_to_child[0],F_SETFD,FD_CLOEXEC);
 	 fcntl(pipe_to_child[1],F_SETFD,FD_CLOEXEC);
 	 m=MOVED;
-	 Log::global->Format(4,"---- %s\n",_("Resolving host address..."));
+	 FA::LogNote(4,_("Resolving host address..."));
       }
 
       if(!w && !buf)
@@ -225,7 +203,7 @@ int   Resolver::Do()
    {
       if(!buf)
       {
-	 Log::global->Format(4,"---- %s\n",_("Resolving host address..."));
+	 FA::LogNote(4,_("Resolving host address..."));
 	 buf=new IOBuffer(IOBuffer::GET);
 	 DoGethostbyname();
 	 if(deleting)
@@ -235,7 +213,7 @@ int   Resolver::Do()
 
    if(buf->Error())
    {
-      err_msg=xstrdup(buf->ErrorText());
+      err_msg.set(buf->ErrorText());
       done=true;
       return MOVED;
    }
@@ -244,7 +222,7 @@ int   Resolver::Do()
    {
       if(timeout_timer.Stopped())
       {
-	 err_msg=xstrdup(_("host name resolve timeout"));
+	 err_msg.set(_("host name resolve timeout"));
 	 done=true;
 	 return MOVED;
       }
@@ -263,93 +241,76 @@ int   Resolver::Do()
    buf->Get(&s,&n);
    if(c=='E' || c=='P') // error
    {
-      const char *tport=portname?portname:defport;
-      err_msg=(char*)xmalloc(strlen(hostname)+strlen(tport)+n+3);
-      sprintf(err_msg,"%s: ",(c=='E'?hostname:tport));
-      char *e=err_msg+strlen(err_msg);
-      memcpy(e,s,n);
-      e[n]=0;
+      const char *tport=portname?portname.get():defport.get();
+      err_msg.vset(c=='E'?hostname.get():tport,": ",s,NULL);
       done=true;
       return MOVED;
    }
-   if((unsigned)n<sizeof(sockaddr_u))
+   if((unsigned)n<addr.get_element_size())
    {
    proto_error:
       if(use_fork)
       {
 	 // e.g. under gdb child fails.
-	 Log::global->Format(4,"**** %s\n","child failed, retrying with dns:use-fork=no");
+	 FA::LogError(4,"child failed, retrying with dns:use-fork=no");
 	 use_fork=false;
-	 Delete(buf);
 	 buf=0;
 	 return MOVED;
       }
-      err_msg=xstrdup("BUG: internal class Resolver error");
+      err_msg.set("BUG: internal class Resolver error");
       done=true;
       return MOVED;
    }
-   addr_num=n/sizeof(*addr);
-   addr=(sockaddr_u*)xmalloc(n);
-   memcpy(addr,s,n);
+   addr.nset((const sockaddr_u*)s,n/addr.get_element_size());
    done=true;
    if(!cache)
       cache=new ResolverCache;
-   cache->Add(hostname,portname,defport,service,proto,addr,addr_num);
-   Log::global->Format(4,plural("---- %d address$|es$ found\n",addr_num),addr_num);
+   cache->Add(hostname,portname,defport,service,proto,addr.get(),addr.count());
+   FA::LogNote(4,plural("%d address$|es$ found",addr.count()),addr.count());
    return MOVED;
 }
 
 void Resolver::MakeErrMsg(const char *f)
 {
    const char *e=strerror(errno);
-   err_msg=(char*)xmalloc(strlen(e)+strlen(f)+3);
-   sprintf(err_msg,"%s: %s",f,e);
+   err_msg.vset(f,": ",e,NULL);
    done=true;
 }
 
 void Resolver::AddAddress(int family,const char *address,int len)
 {
-   addr=(sockaddr_u*)xrealloc(addr,(addr_num+1)*sizeof(*addr));
-   sockaddr_u *add=addr+addr_num;
-   addr_num++;
+   sockaddr_u add;
+   memset(&add,0,sizeof(add));
 
-   memset(add,0,sizeof(*add));
-
-   add->sa.sa_family=family;
+   add.sa.sa_family=family;
    switch(family)
    {
    case AF_INET:
-      if(sizeof(add->in.sin_addr) != len)
-      {
-         addr_num--;
+      if(sizeof(add.in.sin_addr) != len)
          return;
-      }
-      memcpy(&add->in.sin_addr,address,len);
-      add->in.sin_port=port_number;
+      memcpy(&add.in.sin_addr,address,len);
+      add.in.sin_port=port_number;
 #ifdef HAVE_STRUCT_SOCKADDR_SA_LEN
-      add->sa.sa_len=sizeof(add->in);
+      add.sa.sa_len=sizeof(add.in);
 #endif
       break;
 
 #if INET6
    case AF_INET6:
-      if(sizeof(add->in6.sin6_addr) != len)
-      {
-         addr_num--;
+      if(sizeof(add.in6.sin6_addr) != len)
          return;
-      }
-      memcpy(&add->in6.sin6_addr,address,len);
-      add->in6.sin6_port=port_number;
+      memcpy(&add.in6.sin6_addr,address,len);
+      add.in6.sin6_port=port_number;
 #ifdef HAVE_STRUCT_SOCKADDR_SA_LEN
-      add->sa.sa_len=sizeof(add->in6);
+      add.sa.sa_len=sizeof(add.in6);
 #endif
       break;
 #endif
 
    default:
-      addr_num--;
       return;
    }
+   addr.append(add);
 }
 
 int Resolver::FindAddressFamily(const char *name)
@@ -471,10 +432,8 @@ struct SRV
 };
 
 static
-int SRV_compare(const void *a,const void *b)
+int SRV_compare(const SRV *sa,const SRV *sb)
 {
-   struct SRV *sa=(struct SRV*)a;
-   struct SRV *sb=(struct SRV*)b;
    if(sa->priority < sb->priority)
       return -1;
    if(sa->priority > sb->priority)
@@ -496,11 +455,11 @@ void Resolver::LookupSRV_RR()
    if(!ResMgr::QueryBool("dns:SRV-query",hostname))
       return;
 #ifdef HAVE_RES_SEARCH
-   const char *tproto=proto?proto:"tcp";
+   const char *tproto=proto?proto.get():"tcp";
    time_t try_time;
    unsigned char answer[0x1000];
    char *srv_name=string_alloca(strlen(service)+1+strlen(tproto)+1+strlen(hostname)+1);
-   sprintf(srv_name,"_%s._%s.%s",service,tproto,hostname);
+   sprintf(srv_name,"_%s._%s.%s",service.get(),tproto,hostname.get());
 
    int retries=0;
    int max_retries=ResMgr::Query("dns:max-retries",hostname);
@@ -557,18 +516,14 @@ void Resolver::LookupSRV_RR()
       len-=4;
    }
 
-   struct SRV *SRVs=0;
-   int SRV_num=0;
+   xarray<SRV> SRVs;
 
    // now process answers section
    for( ; answer_count>0; answer_count--)
    {
       int dom_len=extract_domain(answer,scan,len,0,0);
       if(dom_len<0)
-      {
-	 xfree(SRVs);
 	 return;
-      }
       scan+=dom_len;
       len-=dom_len;
       if(len<8)
@@ -589,49 +544,44 @@ void Resolver::LookupSRV_RR()
       if(data_len<6)
 	 return;
 
-      SRV_num++;
-      SRVs=(struct SRV*)xrealloc(SRVs,sizeof(*SRVs)*SRV_num);
-      struct SRV *t=SRVs+SRV_num-1;
-
-      t->priority=(scan[0]<<8)+scan[1];
-      t->weight  =(scan[2]<<8)+scan[3];
-      t->port    =(scan[4]<<8)+scan[5];
-
-      t->order=0;
+      struct SRV t;
+      t.priority=(scan[0]<<8)+scan[1];
+      t.weight  =(scan[2]<<8)+scan[3];
+      t.port    =(scan[4]<<8)+scan[5];
+      t.order=0;
 
       scan+=6;
       len-=6;
 
-      dom_len=extract_domain(answer,scan,len,t->domain,sizeof(t->domain));
+      dom_len=extract_domain(answer,scan,len,t.domain,sizeof(t.domain));
       if(dom_len<0)
-      {
-	 xfree(SRVs);
 	 return;
-      }
+
       scan+=dom_len;
       len-=dom_len;
 
-      // check if the service is decidedly not available at this domain.
-      if(!strcmp(t->domain,"."))
-	 SRV_num--;
+      // add unless the service is decidedly not available at this domain.
+      if(strcmp(t.domain,"."))
+	 SRVs.append(t);
    }
 
    // now sort and randomize the list.
-   qsort(SRVs,SRV_num,sizeof(*SRVs),SRV_compare);
+   SRVs.qsort(SRV_compare);
 
    srand(time(0));
 
-   struct SRV *SRVscan,*base=0;
+   int SRVscan;
+   int base=0;
    int curr_priority=-1;
    int weight_sum=0;
-   for(SRVscan=SRVs; ; SRVscan++)
+   for(SRVscan=0; ; SRVscan++)
    {
-      if(SRVscan-SRVs==SRV_num || SRVscan->priority!=curr_priority)
+      if(SRVscan==SRVs.count() || SRVs[SRVscan].priority!=curr_priority)
       {
 	 if(base)
 	 {
 	    int o=1;
-	    struct SRV *s;
+	    int s;
 	    while(weight_sum>0)
 	    {
 	       int r=int(rand()/(RAND_MAX+1.0)*weight_sum);
@@ -640,39 +590,38 @@ void Resolver::LookupSRV_RR()
 	       int w=0;
 	       for(s=base; s<SRVscan; s++)
 	       {
-		  if(s->order!=0)
+		  if(SRVs[s].order!=0)
 		     continue;
-		  w+=s->weight;
+		  w+=SRVs[s].weight;
 		  if(r<w)
 		  {
-		     s->order=o;
+		     SRVs[s].order=o;
 		     o++;
-		     weight_sum-=s->weight;
+		     weight_sum-=SRVs[s].weight;
 		     break;
 		  }
 	       }
 	    }
 	 }
-	 if(SRVscan-SRVs==SRV_num)
+	 if(SRVscan==SRVs.count())
 	    break;
 	 base=SRVscan;
-	 curr_priority=SRVscan->priority;
+	 curr_priority=SRVs[SRVscan].priority;
 	 weight_sum=0;
       }
-      weight_sum+=SRVscan->weight;
+      weight_sum+=SRVs[SRVscan].weight;
    }
 
-   qsort(SRVs,SRV_num,sizeof(*SRVs),SRV_compare);
+   SRVs.qsort(SRV_compare);
 
    int oldport=port_number;
-   for(SRVscan=SRVs; SRVscan-SRVs<SRV_num; SRVscan++)
+   for(SRVscan=0; SRVscan<SRVs.count(); SRVscan++)
    {
-      port_number=htons(SRVscan->port);
-      LookupOne(SRVscan->domain);
+      port_number=htons(SRVs[SRVscan].port);
+      LookupOne(SRVs[SRVscan].domain);
    }
    port_number=oldport;
 
-   xfree(SRVs);
 #endif // HAVE_RES_SEARCH
 }
 
@@ -846,8 +795,8 @@ void Resolver::DoGethostbyname()
 {
    if(port_number==0)
    {
-      const char *tproto=proto?proto:"tcp";
-      const char *tport=portname?portname:defport;
+      const char *tproto=proto?proto.get():"tcp";
+      const char *tport=portname?portname.get():defport.get();
 
       if(isdigit((unsigned char)tport[0]))
 	 port_number=htons(atoi(tport));
@@ -878,7 +827,7 @@ void Resolver::DoGethostbyname()
    if(!use_fork && deleting)
       return;
 
-   if(addr_num==0)
+   if(addr.count()==0)
    {
       buf->Put("E");
       if(error==0)
@@ -887,16 +836,15 @@ void Resolver::DoGethostbyname()
       goto flush;
    }
    buf->Put("O");
-   buf->Put((char*)addr,addr_num*sizeof(*addr));
-   xfree(addr);
-   addr=0;
+   buf->Put((const char*)addr.get(),addr.count()*addr.get_element_size());
+   addr.unset();
 
 flush:
    buf->PutEOF();
    if(use_fork)
    {
       while(buf->Size()>0 && !buf->Error() && !buf->Broken())
-	 Roll(buf);  // should flush quickly.
+	 buf->Roll();  // should flush quickly.
    }
 }
 
@@ -911,6 +859,12 @@ ResolverCache::ResolverCache()
    : Cache(ResMgr::FindRes("dns:cache-size"),ResMgr::FindRes("dns:cache-enable"))
 {
 }
+void ResolverCache::Reconfig(const char *r)
+{
+   if(!xstrcmp(r,"dns:SRV-query")
+   || !xstrcmp(r,"dns:order"))
+      Flush();
+}
 ResolverCacheEntry *ResolverCache::Find(const char *h,const char *p,const char *defp,const char *ser,const char *pr)
 {
    for(ResolverCacheEntry *c=IterateFirst(); c; c=IterateNext())
@@ -931,7 +885,7 @@ void ResolverCache::Add(const char *h,const char *p,const char *defp,
    {
       if(!IsEnabled(h))
 	 return;
-      chain=new ResolverCacheEntry(chain,h,p,defp,ser,pr,a,n);
+      AddCacheEntry(new ResolverCacheEntry(h,p,defp,ser,pr,a,n));
    }
 }
 bool ResolverCacheEntryLoc::Matches(const char *h,const char *p,
