@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: Resolver.h,v 1.18 2006/06/13 14:35:40 lav Exp $ */
+/* $Id: Resolver.h,v 1.24 2007/10/08 13:29:15 lav Exp $ */
 
 #ifndef RESOLVER_H
 #define RESOLVER_H
@@ -30,7 +30,7 @@
 #include <sys/types.h>
 #include <sys/socket.h>
 #include "buffer.h"
-#include "xmalloc.h"
+#include "xarray.h"
 #include "Cache.h"
 
 union sockaddr_u
@@ -44,26 +44,25 @@ union sockaddr_u
 
 class Resolver : public SMTask
 {
-   char *hostname;
-   char *portname;
+   xstring hostname;
+   xstring portname;
 
-   char *service;
-   char *proto;
-   char *defport;
+   xstring service;
+   xstring proto;
+   xstring defport;
 
    int port_number;
 
    int pipe_to_child[2];
-   ProcWait *w;
-   IOBuffer *buf;
+   SMTaskRef<ProcWait> w;
+   SMTaskRef<IOBuffer> buf;
    Timer timeout_timer;
 
-   int addr_num;
-   sockaddr_u *addr;
+   xarray<sockaddr_u> addr;
 
    void AddAddress(int family,const char *a,int len);
 
-   char *err_msg;
+   xstring err_msg;
    bool done;
 
    void  MakeErrMsg(const char *f);
@@ -81,23 +80,21 @@ class Resolver : public SMTask
    bool no_cache;
    bool use_fork;
 
-protected:
-   ~Resolver();
-
 public:
    int	 Do();
    bool	 Done() { return done; }
    bool	 Error() { return err_msg!=0; }
    const char *ErrorMsg() { return err_msg; }
-   sockaddr_u *Result() { return addr; }
-   size_t GetResultSize() { return addr_num*sizeof(*addr); }
-   int	 GetResultNum() { return addr_num; }
-   void  GetResult(void *m) { memcpy(m,addr,GetResultSize()); }
+   const xarray<sockaddr_u>& Result() { return addr; }
+   size_t GetResultSize() { return addr.count()*addr.get_element_size(); }
+   int	 GetResultNum() { return addr.count(); }
+   void  GetResult(void *m) { memcpy(m,addr.get(),GetResultSize()); }
    void	 UseCache(bool y) { no_cache=!y; }
    void	 NoCache() { UseCache(false); }
 
    Resolver(const char *h,const char *p,const char *defp=0,const char *ser=0,
 	    const char *pr=0);
+   ~Resolver();
 
    void Reconfig(const char *name=0);
    const char *GetLogContext() { return hostname; }
@@ -105,60 +102,41 @@ public:
 
 class ResolverCacheEntryLoc
 {
-   char *hostname;
-   char *portname;
-   char *defport;
-   char *service;
-   char *proto;
+   xstring_c hostname;
+   xstring_c portname;
+   xstring_c defport;
+   xstring_c service;
+   xstring_c proto;
 public:
-   ResolverCacheEntryLoc(const char *h,const char *p,const char *defp,const char *ser,const char *pr) {
-      hostname=xstrdup(h);
-      portname=xstrdup(p);
-      service=xstrdup(ser);
-      proto=xstrdup(pr);
-      defport=xstrdup(defp);
-   }
-   ~ResolverCacheEntryLoc() {
-      xfree(hostname);
-      xfree(portname);
-      xfree(service);
-      xfree(proto);
-      xfree(defport);
-   }
+   ResolverCacheEntryLoc(const char *h,const char *p,const char *defp,const char *ser,const char *pr)
+      : hostname(h), portname(p), defport(defp), service(ser), proto(pr) {}
    const char *GetClosure() const { return hostname; }
    bool Matches(const char *h,const char *p,const char *defp,const char *ser,const char *pr);
 };
 class ResolverCacheEntryData
 {
-   int addr_num;
-   sockaddr_u *addr;
+   xarray<sockaddr_u> addr;
 public:
    ResolverCacheEntryData(const sockaddr_u *a,int n) {
-      addr_num=n;
-      addr=(sockaddr_u*)xmemdup(a,n*sizeof(*addr));
-   }
-   ~ResolverCacheEntryData() {
-      xfree(addr);
+      addr.nset(a,n);
    }
    void SetData(const sockaddr_u *a,int n) {
-      xfree(addr);
-      addr_num=n;
-      addr=(sockaddr_u*)xmemdup(a,n*sizeof(*addr));
+      addr.nset(a,n);
    }
    void GetData(const sockaddr_u **a,int *n) {
-      *n=addr_num;
-      *a=addr;
+      *n=addr.count();
+      *a=addr.get();
    }
 };
 class ResolverCacheEntry : public CacheEntry, public ResolverCacheEntryLoc, public ResolverCacheEntryData
 {
 public:
-   ResolverCacheEntry(CacheEntry *next,const char *h,const char *p,const char *defp,const char *ser,const char *pr,
-	 const sockaddr_u *a,int n) : CacheEntry(next), ResolverCacheEntryLoc(h,p,defp,ser,pr), ResolverCacheEntryData(a,n) {
+   ResolverCacheEntry(const char *h,const char *p,const char *defp,const char *ser,const char *pr,
+	 const sockaddr_u *a,int n) : ResolverCacheEntryLoc(h,p,defp,ser,pr), ResolverCacheEntryData(a,n) {
       SetResource("dns:cache-expire",GetClosure());
    }
 };
-class ResolverCache : public Cache
+class ResolverCache : public Cache, public ResClient
 {
    ResolverCacheEntry *Find(const char *h,const char *p,const char *defp,const char *ser,const char *pr);
    ResolverCacheEntry *IterateFirst() { return (ResolverCacheEntry*)Cache::IterateFirst(); }
@@ -170,6 +148,7 @@ public:
    void Find(const char *h,const char *p,const char *defp,
          const char *ser,const char *pr,const sockaddr_u **a,int *n);
    ResolverCache();
+   void Reconfig(const char *);
 };
 
 #endif // RESOLVER_H
