@@ -23,6 +23,7 @@
 
 #include "PollVec.h"
 #include "TimeDate.h"
+#include "Ref.h"
 
 class SMTask
 {
@@ -31,10 +32,8 @@ class SMTask
    SMTask *next;
 
    static SMTask *chain;
-   static PollVec sched_total;
-   static SMTask **stack;
-   static int stack_ptr;
-   static int stack_size;
+   static PollVec block;
+   static xarray<SMTask*> stack;
 
    bool	 suspended;
    bool	 suspended_slave;
@@ -51,25 +50,21 @@ protected:
       WANTDIE=2	  // for AcceptSig
    };
 
-   virtual ~SMTask();
-
    // SuspendInternal and ResumeInternal usually suspend and resume slave tasks
    virtual void SuspendInternal() {}
    virtual void ResumeInternal() {}
 
 public:
-   PollVec  block;
-
-   void Block(int fd,int mask) { block.AddFD(fd,mask); }
-   void Timeout(int ms) { block.AddTimeout(ms); }
-   void TimeoutS(int s) { Timeout(1000*s); }
+   static void Block(int fd,int mask) { block.AddFD(fd,mask); }
+   static void Timeout(int ms) { block.AddTimeout(ms); }
+   static void TimeoutS(int s) { Timeout(1000*s); }
 
    static TimeDate now;
    static void UpdateNow() { now.SetToCurrentTime(); }
 
    static void Schedule();
    static int CollectGarbage();
-   static void Block() { sched_total.Block(); }
+   static void Block() { block.Block(); }
 
    void Suspend();
    void Resume();
@@ -81,20 +76,21 @@ public:
 
    bool IsSuspended() { return suspended|suspended_slave; }
 
-   virtual void Reconfig(const char *name=0) {};
-   static void ReconfigAll(const char *name);
-
    virtual const char *GetLogContext() { return 0; }
 
    SMTask();
+   virtual ~SMTask();
 
    void DeleteLater() { deleting=true; }
    static void Delete(SMTask *);
-   static SMTask *_MakeRef(SMTask *task) { if(task) task->ref_count++; return task; }
-   static void _DeleteRef(SMTask *task)  { if(task) task->ref_count--; Delete(task); }
-   template<typename T> static void DeleteRef(T *&task) { _DeleteRef(task); task=0; }
+   void IncRefCount() { ref_count++; }
+   void DecRefCount() { if(ref_count>0) ref_count--; }
+   static SMTask *_MakeRef(SMTask *task) { if(task) task->IncRefCount(); return task; }
+   static void _DeleteRef(SMTask *task)  { if(task) task->DecRefCount(); Delete(task); }
+/*   template<typename T> static void DeleteRef(T *&task) { _DeleteRef(task); task=0; }*/
    template<typename T> static T *MakeRef(T *task) { _MakeRef(task); return task; }
    static int Roll(SMTask *);
+   int Roll() { return Roll(this); }
    static void RollAll(const TimeInterval &max_time);
 
    static SMTask *current;
@@ -120,4 +116,28 @@ public:
    ~SMTaskInit();
 };
 
+template<class T> class SMTaskRef : public Ref<T>
+{
+   SMTaskRef<T>(const SMTaskRef<T>&);  // disable cloning
+   void operator=(const SMTaskRef<T>&);   // and assignment
+
+public:
+#define ptr Ref<T>::ptr
+   SMTaskRef() {}
+   SMTaskRef<T>(T *p) : Ref<T>(SMTask::MakeRef(p)) {}
+   ~SMTaskRef<T>() { SMTask::_DeleteRef(ptr); ptr=0; }
+   void operator=(T *p) { SMTask::_DeleteRef(ptr); ptr=SMTask::MakeRef(p); }
+   T *borrow() { if(ptr) ptr->DecRefCount(); return Ref<T>::borrow(); }
+#undef ptr
+   void unset() { *this=0; }
+};
+
+template<typename T>
+class TaskRefArray : public _RefArray< T,SMTaskRef<T> > {
+   TaskRefArray& operator=(const TaskRefArray&); // make assignment fail
+   TaskRefArray(const TaskRefArray&);	       // disable cloning
+public:
+   TaskRefArray() : _RefArray< T,SMTaskRef<T> >() {}
+};
+
 #endif /* SMTASK_H */
