@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1998-2002 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1998-2005 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: SleepJob.cc,v 1.21 2004/02/17 10:40:56 lav Exp $ */
+/* $Id: SleepJob.cc,v 1.34 2007/11/03 15:29:35 lav Exp $ */
 
 #include <config.h>
 #include <ctype.h>
@@ -28,22 +28,22 @@
 #include "LocalDir.h"
 
 SleepJob::SleepJob(const TimeInterval &when,FileAccess *s,LocalDirectory *cwd,char *what)
-   : SessionJob(s), next_time(when)
+   : SessionJob(s), Timer(when), saved_cwd(cwd)
 {
-   start_time=now;
-   cmd=what;
+   cmd.set_allocated(what);
    exit_code=0;
    done=false;
-   saved_cwd=cwd;
    repeat=false;
+   weak=false;
    repeat_count=0;
+   max_repeat_count=0;
    exec=0;
+   continue_code=-1;
+   break_code=-1;
 }
 SleepJob::~SleepJob()
 {
    Delete(exec);
-   xfree(cmd);
-   delete saved_cwd;
 }
 
 int SleepJob::Do()
@@ -51,45 +51,36 @@ int SleepJob::Do()
    if(Done())
       return STALL;
 
-   if(waiting_num>0)
+   if(waiting.count()>0)
    {
       Job *j=FindDoneAwaitedJob();
       if(!j)
 	 return STALL;
-      if(!repeat)
+      exit_code=j->ExitCode();
+      if(!repeat || (++repeat_count>=max_repeat_count && max_repeat_count)
+      || exit_code==break_code || (continue_code!=-1 && exit_code!=continue_code))
       {
-	 exit_code=j->ExitCode();
 	 RemoveWaiting(j);
 	 Delete(j);
 	 exec=0;
 	 done=true;
 	 return MOVED;
       }
-      repeat_count++;
-      start_time=now;
+      Reset();
       exec=(CmdExec*)j; // we are sure it is CmdExec.
       RemoveWaiting(j);
    }
 
-   if(next_time.IsInfty())
-   {
-      TimeoutS(HOUR);  // to avoid deadlock message
-      return STALL;
-   }
-
-   if(now >= start_time+next_time.Seconds())
+   if(Stopped())
    {
       if(cmd)
       {
 	 if(!exec)
 	 {
-	    exec=new CmdExec(session,saved_cwd);
-	    session=0;
-	    saved_cwd=0;
+	    exec=new CmdExec(session.borrow(),saved_cwd.borrow());
 	    exec->SetParentFg(this);
 	    exec->AllocJobno();
-	    exec->cmdline=(char*)xmalloc(3+strlen(cmd));
-	    sprintf(exec->cmdline,"(%s)",cmd);
+	    exec->cmdline.vset("(",cmd.get(),")",NULL);
 	 }
 	 exec->FeedCmd(cmd);
 	 exec->FeedCmd("\n");
@@ -100,10 +91,6 @@ int SleepJob::Do()
       done=true;
       return MOVED;
    }
-   time_t diff=next_time.Seconds()-(time_t(now)-start_time);
-   if(diff>1024)
-      diff=1024;  // prevent overflow
-   TimeoutS(diff);
    return STALL;
 }
 
@@ -118,8 +105,7 @@ void SleepJob::PrintStatus(int,const char *prefix)
 
 void SleepJob::lftpMovesToBackground()
 {
-   if(next_time.IsInfty()
-   || (repeat && cmd[0]==0))
+   if(weak || IsInfty() || (repeat && cmd[0]==0))
    {
       // terminate
       done=true;
@@ -140,7 +126,7 @@ Job *cmd_sleep(CmdExec *parent)
       return 0;
    }
    const char *t=args->getarg(1);
-   TimeInterval delay(t);
+   TimeIntervalR delay(t);
    if(delay.Error())
    {
       eprintf("%s: %s: %s. ",op,t,delay.ErrorText());
@@ -153,23 +139,79 @@ Job *cmd_repeat(CmdExec *parent)
 {
    const char *op=args->a0();
    int cmd_start=1;
-   const char *t=args->getarg(1);
-   TimeInterval delay(1);
-   if(t && isdigit((unsigned char)t[0]))
+   TimeIntervalR delay(1);
+   int max_count=0;
+   const char *delay_str=0;
+   bool while_ok=false;
+   bool until_ok=false;
+   bool weak=false;
+   int opt;
+
+   static struct option repeat_opts[]=
    {
-      delay=TimeInterval(t);
-      if(delay.Error())
+      {"delay",required_argument,0,'d'},
+      {"count",required_argument,0,'c'},
+      {"while-ok",no_argument,0,'o'},
+      {"until-ok",no_argument,0,'O'},
+      {"weak",no_argument,0,'w'},
+      {0},
+   };
+
+   args->rewind();
+   while((opt=args->getopt_long("+c:d:",repeat_opts,0))!=EOF)
+   {
+      switch(opt)
       {
-	 eprintf("%s: %s: %s.\n",op,t,delay.ErrorText());
+      case('c'):
+	 max_count=atoi(optarg);
+	 break;
+      case('d'):
+	 delay_str=optarg;
+	 break;
+      case('o'):
+	 while_ok=true;
+	 break;
+      case('O'):
+	 until_ok=true;
+	 break;
+      case('w'):
+	 weak=true;
+	 break;
+      case('?'):
+	 eprintf(_("Try `help %s' for more information.\n"),args->a0());
+	 return 0;
+      }
+   }
+   if(!delay_str)
+   {
+      const char *t=args->getcurr();
+      if(t && isdigit((unsigned char)t[0]))
+      {
+	 args->getnext();
+	 delay_str=t;
+      }
+   }
+   cmd_start=args->getindex();
+
+   if(delay_str)
+   {
+      delay.Set(delay_str);
+      if(delay.Error())
+      {
+	 eprintf("%s: %s: %s.\n",op,delay_str,delay.ErrorText());
 	 return 0;
       }
-      cmd_start=2;
    }
 
    char *cmd = (args->count()==cmd_start+1
 	        ? args->Combine(cmd_start) : args->CombineQuoted(cmd_start));
    SleepJob *s=new SleepJob(delay,session->Clone(),parent->cwd->Clone(),cmd);
-   s->Repeat();
+   s->Repeat(max_count);
+   s->SetWeak(weak);
+   if(while_ok)
+      s->ContinueCode(0);
+   if(until_ok)
+      s->BreakCode(0);
    return s;
 }
 
@@ -195,17 +237,10 @@ Job *cmd_at(CmdExec *parent)
       count++;
    }
 
-#if 0
-   char **av=(char**)xmemdup(args->GetV(),(count+1)*sizeof(char**));
-   av[count]=0;
-   time_t when=parsetime(count-1,av+1);
-   xfree(av);
-#endif
-   char *date=args->Combine(1);
-   date[date_len]=0;
-   time_t now=time(0);
+   xstring_ca date(args->Combine(1));
+   date.truncate(date_len);
+   time_t now=SMTask::now;
    time_t when=get_date(date,&now);
-   xfree(date);
 
    if(when==0 || when==(time_t)-1)
    {
@@ -230,9 +265,10 @@ Job *cmd_at(CmdExec *parent)
    }
 
    if(!cmd)
-      return new SleepJob(when-now);
+      return new SleepJob(Time(when,0)-SMTask::now);
 
-   return new SleepJob(when-now, session->Clone(), parent->cwd->Clone(), cmd);
+   return new SleepJob(Time(when,0)-SMTask::now,
+			session->Clone(), parent->cwd->Clone(), cmd);
 }
 #undef args
 
