@@ -51,19 +51,26 @@ void Timer::set_last_setting(const TimeInterval &i)
    infty_count-=IsInfty();
    last_setting=i;
    infty_count+=IsInfty();
+   re_set();
+}
+void Timer::re_set()
+{
+   stop=start;
+   stop+=last_setting;
+   re_sort();
 }
 void Timer::Set(const TimeInterval &i)
 {
    resource=closure=0;
+   start=SMTask::now;
    set_last_setting(i);
-   Reset();
 }
 void Timer::Reset(const Time &t)
 {
+   if(start>=t)
+      return;
    start=t;
-   stop=t;
-   stop+=last_setting;
-   re_sort();
+   re_set();
 }
 void Timer::ResetDelayed(int s)
 {
@@ -92,12 +99,7 @@ bool Timer::Stopped() const
 void Timer::reconfig(const char *r)
 {
    if(resource && (!r || !strcmp(r,resource)))
-   {
       set_last_setting(TimeIntervalR(ResMgr::Query(resource,closure)));
-      stop=start;
-      stop+=last_setting;
-      re_sort();
-   }
 }
 void Timer::init()
 {
@@ -107,15 +109,19 @@ void Timer::init()
    next_all=chain_all;
    chain_all=this;
 }
-Timer::~Timer()
+void Timer::remove_from_running_list()
 {
-   infty_count-=IsInfty();
    if(next_running)
       next_running->prev_running=prev_running;
    if(prev_running)
       prev_running->next_running=next_running;
    if(chain_running==this)
       chain_running=next_running;
+}
+Timer::~Timer()
+{
+   remove_from_running_list();
+   infty_count-=IsInfty();
    Timer **scan=&chain_all;
    while(*scan!=this)
       scan=&scan[0]->next_all;
@@ -129,7 +135,7 @@ Timer::Timer(const TimeInterval &d) : last_setting(d)
 {
    init();
    infty_count+=IsInfty();
-   Reset();
+   re_set();
 }
 Timer::Timer(const char *r,const char *c) : last_setting(0,0)
 {
@@ -159,15 +165,16 @@ void Timer::re_sort()
       // find new location in the list.
       Timer *new_next=next_running;
       Timer *new_prev=prev_running;
+
       if(prev_running==0 && next_running==0 && chain_running!=this)
-	 new_next=chain_running;
+	 new_next=chain_running; // it was not in the running list.
       else if((!prev_running || prev_running->stop<stop)
 	   && (!next_running || stop<next_running->stop))
-	 return;
-      if(next_running)
-	 next_running->prev_running=prev_running;
-      if(prev_running)
-	 prev_running->next_running=next_running;
+	 return;  // it was already properly sorted.
+
+      remove_from_running_list();
+
+      // find new position in the list.
       while(new_next && new_next->stop<stop)
       {
 	 new_prev=new_next;
@@ -178,6 +185,8 @@ void Timer::re_sort()
 	 new_next=new_prev;
 	 new_prev=new_prev->prev_running;
       }
+
+      // re-insert it.
       next_running=new_next;
       prev_running=new_prev;
       if(new_next)
