@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-2002 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2006 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: complete.cc,v 1.66 2003/12/17 13:00:23 lav Exp $ */
+/* $Id: complete.cc,v 1.76 2007/09/17 14:01:28 lav Exp $ */
 
 #include <config.h>
 
@@ -56,8 +56,8 @@ CDECL_END
 # define GLOB_PERIOD 0
 #endif
 
-static char *bash_dequote_filename (char *text, int quote_char);
-static int lftp_char_is_quoted(char *string,int eindex);
+static char *bash_dequote_filename (const char *text, int quote_char);
+static int lftp_char_is_quoted(const char *string,int eindex);
 
 static int len;    // lenght of the word to complete
 static int cindex; // index in completion array
@@ -82,7 +82,7 @@ char *command_generator(const char *text,int state)
    }
 
    /* Return the next name which partially matches from the command list. */
-   while ((name=CmdExec::CmdByIndex(cindex)->name)!=0)
+   while ((name=CmdExec::CmdByIndex(cindex))!=0)
    {
       cindex++;
       if(name[0]=='.' && len==0)
@@ -118,15 +118,13 @@ static char *file_generator(const char *text,int state)
 
       if(!name[0])
 	 continue;
-      /* it's not our job to be case-sensitive; that's done (or not done)
-       * when we create the glob_res */
-      if(strncasecmp(name,text,len)==0)
-	 return(xstrdup(name));
+      return(xstrdup(name));
    }
 
    return NULL;
 }
 
+static bool bookmark_prepend_bm;
 static char *bookmark_generator(const char *text,int s)
 {
    static int state;
@@ -149,6 +147,12 @@ static char *bookmark_generator(const char *text,int s)
 	 }
 	 if(!lftp_bookmarks.Next())
 	    state=1;
+	 if(bookmark_prepend_bm)
+	 {
+	    char *e=alloca_strdup2("bm:",strlen(t)*3);
+	    url::encode_string(t,e+3,URL_HOST_UNSAFE);
+	    t=e;
+	 }
 	 if(strncmp(t,text,len)==0)
 	    return xstrdup(t);
 	 break;
@@ -251,23 +255,59 @@ int ignore_non_dirs(char **matches)
 
 static const char *find_word(const char *p)
 {
-   while(*p && isspace(*p))
+   while(CmdExec::is_space(*p))
       p++;
    return p;
 }
+static bool copy_word_skip(const char **p_in,char *buf,int n)
+{
+   const char *&p=*p_in;
+   char in_quotes=0;
+   for(;;)
+   {
+      if(!*p)
+	 break;
+      if(in_quotes)
+      {
+	 if(*p==in_quotes)
+	 {
+	    in_quotes=0,p++;
+	    continue;
+	 }
+	 else if(*p=='\\' && CmdExec::quotable(p[1],in_quotes))
+	    p++;
+	 if(buf && n>0)
+	    *buf++=*p,n--;
+	 p++;
+	 continue;
+      }
+      if(CmdExec::is_space(*p))
+	 break;
+      if(*p=='\\' && CmdExec::quotable(p[1],in_quotes))
+	 p++;
+      else if(CmdExec::is_quote(*p))
+      {
+	 in_quotes=*p++;
+	 continue;
+      }
+      if(buf && n>0)
+	 *buf++=*p,n--;
+      p++;
+   }
+   if(n>0 && buf)
+      *buf=0;
+   return n>0;
+}
 // returns false when buffer overflows
 static bool copy_word(char *buf,const char *p,int n)
 {
-   while(n>0 && *p && !isspace(*p))
-   {
-      *buf++=*p++;
-      n--;
-   }
-   if(n>0)
-      *buf=0;
-   return n>0;
+   return copy_word_skip(&p,buf,n);
+}
+static const char *skip_word(const char *p)
+{
+   copy_word_skip(&p,0,0);
+   return p;
 }
-
 
 enum completion_type
 {
@@ -275,21 +315,20 @@ enum completion_type
    STRING_ARRAY, VARIABLE, NO_COMPLETION
 };
 
+// cmd: ptr to command line being completed
+// start: location of the word being completed
 static completion_type cmd_completion_type(const char *cmd,int start)
 {
-   const char *w=find_word(cmd);
-
-   if(w-cmd == start) // first word is command
-      return COMMAND;
-
-   // try to guess whether the completion word is remote
-
+   const char *w=0;
    char buf[20];  // no commands longer
    TouchedAlias *used_aliases=0;
 
+   // try to guess whether the completion word is remote
    for(;;)
    {
       w=find_word(cmd);
+      if(w-cmd == start) // first word is command
+	 return COMMAND;
       if(w[0]=='!')
 	 shell_cmd=true;
       if(w[0]=='#')
@@ -303,6 +342,7 @@ static completion_type cmd_completion_type(const char *cmd,int start)
 	 return COMMAND;
       if(w[0]=='(')
       {
+	 start-=(w+1-cmd);
 	 cmd=w+1;
 	 continue;
       }
@@ -316,7 +356,11 @@ static completion_type cmd_completion_type(const char *cmd,int start)
       if(alias && !TouchedAlias::IsTouched(alias,used_aliases))
       {
 	 used_aliases=new TouchedAlias(alias,used_aliases);
-	 cmd=alias;
+	 char *cmd1=string_alloca(strlen(alias)+strlen(w)-strlen(buf)+1);
+	 strcpy(cmd1,alias);
+	 strcat(cmd1,w+strlen(buf));
+	 start=start-strlen(buf)+strlen(alias);
+	 cmd=cmd1;
 	 continue;
       }
       const char *full=CmdExec::GetFullCommandName(buf);
@@ -326,11 +370,11 @@ static completion_type cmd_completion_type(const char *cmd,int start)
       break;
    }
 
-   for(char *p=rl_line_buffer+start; p>rl_line_buffer; )
+   for(const char *p=cmd+start; p>cmd; )
    {
       p--;
       if((*p=='>' || *p=='|')
-      && !lftp_char_is_quoted(rl_line_buffer,p-rl_line_buffer))
+      && !lftp_char_is_quoted(cmd,p-cmd))
 	 return LOCAL;
       if(!isspace((unsigned char)*p))
 	 break;
@@ -377,6 +421,8 @@ static completion_type cmd_completion_type(const char *cmd,int start)
    bool was_o=false;
    bool was_N=false;
    bool was_O=false;
+   bool have_R=false;
+   bool have_N=false;
    bool second=false;
    int second_start=-1;
    for(int i=start; i>4; i--)
@@ -404,9 +450,7 @@ static completion_type cmd_completion_type(const char *cmd,int start)
 	 break;
       }
    }
-   w=find_word(cmd);
-   while(*w && !isspace(*w))
-      w++;  // FIXME: handle quotations.
+   w=skip_word(find_word(cmd));
    if(*w)
    {
       w=find_word(w);
@@ -415,6 +459,25 @@ static completion_type cmd_completion_type(const char *cmd,int start)
 	 second=true;
    }
 
+   if(re_match(cmd," -[a-zA-Z]*R[a-zA-Z]* ",0))
+      have_R=true;
+
+   int arg=0;
+   bool opt_stop=false;
+   for(w=cmd; *w; )
+   {
+      w=find_word(w);
+      if(w-cmd==start || w-cmd+CmdExec::is_quote(*w)==start)
+	 break;
+      if(w[0]!='-' || opt_stop)
+	 arg++;
+      else if(re_match(w,"^-[a-zA-Z]*N ",0))
+	 have_N=true;
+      if(!strncmp(w,"-- ",3))
+	 opt_stop=true;
+      w=skip_word(w);
+   }
+
    if(!strcmp(buf,"get")
    || !strcmp(buf,"pget")
    || !strcmp(buf,"get1"))
@@ -436,9 +499,15 @@ static completion_type cmd_completion_type(const char *cmd,int start)
 	 return REMOTE_DIR;
    if(!strcmp(buf,"mirror"))
    {
-      // FIXME: guess -R and take arg number into account.
-      if(!was_N)
-	 return REMOTE_DIR;
+      if(was_N)
+	 return LOCAL;
+      if(have_N)
+	 arg--;
+      if(have_R ^ arg!=1)
+	 return LOCAL_DIR;
+      if(arg>2)
+	 return NO_COMPLETION;
+      return REMOTE_DIR;
    }
    if(!strcmp(buf,"bookmark"))
    {
@@ -572,6 +641,7 @@ static char **lftp_completion (const char *text,int start,int end)
       generator = command_generator;
       break;
    case BOOKMARK:
+      bookmark_prepend_bm=false;
       generator = bookmark_generator;
       break;
    case STRING_ARRAY:
@@ -590,9 +660,8 @@ static char **lftp_completion (const char *text,int start,int end)
 	    type=REMOTE_DIR;
 	 goto really_remote;
       }
-
-      ArgV arg("", ResMgr::Query("cmd:cls-completion-default", 0));
-      fso.parse_argv(&arg);
+   really_local:
+      fso.parse_res(ResMgr::Query("cmd:cls-completion-default", 0));
 
       bool tilde_expanded=false;
       const char *home=getenv("HOME");
@@ -659,8 +728,20 @@ static char **lftp_completion (const char *text,int start,int end)
    case REMOTE_FILE:
    case REMOTE_DIR: {
       if(!remote_completion && !force_remote)
-	 break; // local
+      {
+	 if(type==REMOTE_DIR)
+	    type=LOCAL_DIR;
+	 goto really_local;
+      }
    really_remote:
+      if(!strncmp(text,"bm:",3) && !strchr(text,'/'))
+      {
+	 bookmark_prepend_bm=true;
+	 generator=bookmark_generator;
+	 rl_completion_append_character='/';
+	 break;
+      }
+
       pat=(char*)alloca(len*2+10);
       glob_quote(pat,text,len);
 
@@ -682,8 +763,7 @@ static char **lftp_completion (const char *text,int start,int end)
 
       rl_save_prompt();
 
-      ArgV arg("", ResMgr::Query("cmd:cls-completion-default", 0));
-      fso.parse_argv(&arg);
+      fso.parse_res(ResMgr::Query("cmd:cls-completion-default", 0));
 
       if(rg)
       {
@@ -756,8 +836,7 @@ static char **lftp_completion (const char *text,int start,int end)
    char quoted=((lftp_char_is_quoted(rl_line_buffer,start) &&
 		 strchr(rl_completer_quote_characters,rl_line_buffer[start-1]))
 		? rl_line_buffer[start-1] : 0);
-   char *textq=alloca_strdup(text);
-   textq=bash_dequote_filename(textq, quoted);
+   xstring_ca textq(bash_dequote_filename(text, quoted));
    len=strlen(textq);
 
    char **matches=lftp_rl_completion_matches(textq,generator);
@@ -773,14 +852,12 @@ static char **lftp_completion (const char *text,int start,int end)
    if(!matches)
    {
       rl_attempted_completion_over = 1;
-      goto leave;
+      return 0;
    }
 
    if(type==REMOTE_DIR)
       rl_completion_append_character='/';
 
-leave:
-   xfree(textq);
    return matches;
 }
 
@@ -950,9 +1027,11 @@ contains_shell_metas (char *string)
    allows single quotes to appear within double quotes, and vice versa.
    It should be smarter. */
 static char *
-bash_dequote_filename (char *text, int quote_char)
+bash_dequote_filename (const char *text, int quote_char)
 {
-  char *ret, *p, *r;
+  char *ret;
+  const char *p;
+  char *r;
   int l, quoted;
 
   l = strlen (text);
@@ -1114,7 +1193,7 @@ static int skip_quoted(const char *s, int i, char q)
    return i;
 }
 
-int lftp_char_is_quoted(char *string,int eindex)
+int lftp_char_is_quoted(const char *string,int eindex)
 {
   int i, pass_next;
 
@@ -1213,7 +1292,7 @@ void lftp_readline_init ()
 
 extern "C" void completion_display_list (char **matches, int len)
 {
-   OutputJob *b=new OutputJob((FDStream *) NULL, "completion");
+   JobRef<OutputJob> b(new OutputJob((FDStream *) NULL, "completion"));
 
    if(glob_res) {
       /* Our last completion action was of files, and we kept that
@@ -1231,8 +1310,7 @@ extern "C" void completion_display_list (char **matches, int len)
       FileSetOutput fso;
       fso.config(b);
 
-      ArgV arg("", ResMgr::Query("cmd:cls-completion-default", 0));
-      fso.parse_argv(&arg);
+      fso.parse_res(ResMgr::Query("cmd:cls-completion-default", 0));
 
       fso.print(tmp, b);
    } else {
@@ -1256,6 +1334,4 @@ extern "C" void completion_display_list (char **matches, int len)
 	 break;
       }
    }
-
-   SMTask::Delete(b);
 }
