@@ -18,11 +18,15 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
+/* $Id: fg.h,v 1.1 1998/08/14 16:05:23 lav Exp $ */
+
 #ifndef FG_H
 #define FG_H
 
 #include <sys/types.h>
 
+
+#if 0
 /* Assume Posix termio if we have the header and function */
 #if HAVE_TERMIOS_H && HAVE_TCGETATTR
 #ifndef TERMIOS
@@ -69,11 +73,13 @@
 #define SET_TTY(fd, buf) stty(fd, buf)
 #endif
 
+#endif // 0
+
 class FgData
 {
    pid_t pg;
    pid_t old_pgrp;
-#if 0 // not yet done
+#if 0 // not done yet
    TTY tc;
    TTY old_tc;
 #endif
@@ -81,6 +87,7 @@ public:
    FgData(pid_t npg,bool fg);
    ~FgData();
    void Fg(),Bg();
+   void cont();
 };
 
 #endif//FG_H
