@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-2005 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2007 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: ftpclass.h,v 1.137.2.1 2007/08/20 08:33:58 lav Exp $ */
+/* $Id: ftpclass.h,v 1.152 2007/10/18 13:27:16 lav Exp $ */
 
 #ifndef FTPCLASS_H
 #define FTPCLASS_H
@@ -32,11 +32,21 @@
 # include "lftp_ssl.h"
 #endif
 
+class TelnetEncode : public DataTranslator {
+   void PutTranslated(Buffer *target,const char *buf,int size);
+};
+class TelnetDecode : public DataTranslator {
+   void PutTranslated(Buffer *target,const char *buf,int size);
+};
 class IOBufferTelnet : public IOBufferStacked
 {
-   void PutTranslated(const char *,int);
 public:
-   IOBufferTelnet(IOBuffer *b) : IOBufferStacked(b) {}
+   IOBufferTelnet(IOBuffer *b) : IOBufferStacked(b) {
+      if(mode==PUT)
+	 SetTranslator(new TelnetEncode());
+      else
+	 SetTranslator(new TelnetDecode());
+   }
 };
 
 class Ftp : public NetAccess
@@ -60,15 +70,15 @@ class Ftp : public NetAccess
 
    class Connection
    {
-      char *closure;
+      xstring closure;
    public:
       int control_sock;
-      IOBuffer *control_recv;
-      IOBuffer *control_send;
+      SMTaskRef<IOBuffer> control_recv;
+      SMTaskRef<IOBuffer> control_send;
       IOBufferTelnet *telnet_layer_send;
-      Buffer *send_cmd_buffer; // holds unsent commands.
+      DirectedBuffer send_cmd_buffer; // holds unsent commands.
       int data_sock;
-      IOBuffer *data_iobuf;
+      SMTaskRef<IOBuffer> data_iobuf;
       int aborted_data_sock;
       sockaddr_u peer_sa;
       sockaddr_u data_sa; // address for data accepting
@@ -110,26 +120,25 @@ class Ftp : public NetAccess
       Timer abor_close_timer; // timer for closing aborted connection.
 
 #if USE_SSL
-      lftp_ssl *control_ssl;
+      Ref<lftp_ssl> control_ssl;
       char prot;  // current data protection scheme 'C'lear or 'P'rivate
       bool auth_sent;
       bool auth_supported;
       bool cpsv_supported;
       bool sscn_supported;
       bool sscn_on;
-      char *auth_args_supported;
+      xstring auth_args_supported;
       bool ssl_is_activated() { return control_ssl!=0; }
 #else
       bool ssl_is_activated() { return false; }
 #endif
 
-      char *mlst_attr_supported;
+      xstring mlst_attr_supported;
 
       Connection(const char *c);
       ~Connection();
 
-      bool data_address_ok(sockaddr_u *d,bool verify_address,bool verify_port);
-      void SavePeerAddress();
+      bool data_address_ok(const sockaddr_u *d,bool verify_address,bool verify_port);
 
       void MakeBuffers();
       void MakeSSLBuffers(const char *h);
@@ -141,15 +150,17 @@ class Ftp : public NetAccess
       void AbortDataConnection();
       void CloseAbortedDataConnection();
 
-      void Send(const char *cmd,int len);
+      void Send(const char *cmd);
+      void SendEncoded(const char *url);
       void SendCmd(const char *cmd);
-      void SendCmd2(const char *cmd,const char *f);
+      void SendCmd2(const char *cmd,const char *f,const char *u=0,const char *home=0);
       void SendCmd2(const char *cmd,int v);
       void SendCmdF(const char *fmt,...) PRINTF_LIKE(2,3);
       int FlushSendQueueOneCmd(); // sends single command from send_cmd_buffer
    };
 
-   Connection *conn;
+   Ref<Connection> conn;
+   bool last_connection_failed;
 
    struct Expect
    {
@@ -192,24 +203,14 @@ class Ftp : public NetAccess
       };
 
       expect_t check_case;
-      char *arg;
+      xstring arg;
       Expect *next;
 
-      Expect(expect_t e,const char *a=0)
+      Expect(expect_t e,const char *a=0) : check_case(e), arg(a) {}
+      Expect(expect_t e,char c) : check_case(e)
 	 {
-	    check_case=e;
-	    arg=xstrdup(a);
-	 }
-      Expect(expect_t e,char c)
-	 {
-	    check_case=e;
-	    arg=(char*)xmalloc(2);
-	    arg[0]=c;
-	    arg[1]=0;
-	 }
-      ~Expect()
-	 {
-	    xfree(arg);
+	    char cc[2]={c,0};
+	    arg.set(cc);
 	 }
    };
    class ExpectQueue;
@@ -235,7 +236,7 @@ class Ftp : public NetAccess
       void Close();
    };
 
-   ExpectQueue *expect;
+   Ref<ExpectQueue> expect;
 
    void  CheckResp(int resp);
    int	 ReplyLogPriority(int code);
@@ -254,6 +255,7 @@ class Ftp : public NetAccess
    void	 CatchDATE_opt(int);
    void	 CatchSIZE(int);
    void	 CatchSIZE_opt(int);
+   void	 TurnOffStatForList();
 
    enum pasv_state_t
    {
@@ -282,9 +284,8 @@ private:
    static const bool ftps; // for convenience
 #endif
 
-   char  *line;
-   int	 line_len;
-   char  *all_lines;
+   xstring line;
+   xstring all_lines;
 
    bool	 eof;
 
@@ -314,6 +315,7 @@ private:
 
    const char *QueryStringWithUserAtHost(const char *);
 
+   int	 ReceiveOneLine();
    // If a response is received, it checks it for accordance with
    // response_queue and switch to a state if necessary
    int	 ReceiveResp();
@@ -330,10 +332,10 @@ private:
    void	 HttpProxySendConnect();
    void	 HttpProxySendConnectData();
    // Send http proxy auth.
-   void	 HttpProxySendAuth(IOBuffer *);
+   void	 HttpProxySendAuth(const SMTaskRef<IOBuffer>&);
    // Check if proxy returned a reply, returns true if reply is ok.
    // May disconnect.
-   bool	 HttpProxyReplyCheck(IOBuffer *buf);
+   bool	 HttpProxyReplyCheck(const SMTaskRef<IOBuffer>&);
 
    bool	 AbsolutePath(const char *p);
 
@@ -341,10 +343,10 @@ private:
 
    automate_state state;
 
-   char	 *anon_user;
-   char	 *anon_pass;
+   xstring anon_user;
+   xstring anon_pass;
 
-   char	 *charset;
+   xstring charset;
 
    static const char *DefaultAnonPass();
 
@@ -355,14 +357,14 @@ private:
    bool  verify_data_address;
    bool  verify_data_port;
    bool	 rest_list;
-   char  *list_options;
+   xstring list_options;
 
    bool	 GetBetterConnection(int level,bool limit_reached);
    bool  SameConnection(const Ftp *o) const;
 
    int	 nop_interval;
 
-   char *skey_pass;
+   xstring skey_pass;
    bool allow_skey;
    bool force_skey;
    const char *make_skey_reply();
@@ -384,6 +386,7 @@ private:
    bool copy_failed;
 
    bool use_stat;
+   bool use_stat_for_list;
    bool use_mdtm;
    bool use_size;
    bool use_pret;
@@ -392,7 +395,7 @@ private:
 
    bool use_telnet_iac;
 
-   const char *encode_eprt(sockaddr_u *);
+   const char *encode_eprt(const sockaddr_u *);
 
    typedef FileInfo *(*FtpLineParser)(char *line,int *err,const char *tz);
    static FtpLineParser line_parsers[];
@@ -485,7 +488,7 @@ public:
 	 retries=rnum;
 	 try_time=tt;
       }
-   bool SetCopyAddress(Ftp *o)
+   bool SetCopyAddress(const Ftp *o)
       {
 	 if(copy_addr_valid || !o->copy_addr_valid)
 	    return false;
@@ -512,11 +515,12 @@ public:
 	    return copy_addr_valid && expect->FirstIs(Expect::TRANSFER);
 	 return state==WAITING_STATE && expect->IsEmpty();
       }
-   void CopyCheckTimeout(Ftp *o)
+   void CopyCheckTimeout(const Ftp *o)
       {
 	 timeout_timer.Reset(o->timeout_timer);
 	 CheckTimeout();
       }
+   bool AnonymousQuietMode();
 };
 
 class FtpS : public Ftp
