@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: history.h,v 1.1 1998/10/03 15:27:37 lav Exp $ */
+/* $Id: history.h,v 1.5 2007/04/23 07:11:52 lav Exp $ */
 
 #ifndef HISTORY_H
 #define HISTORY_H
@@ -31,8 +31,9 @@ class History : public KeyValueDB
    KeyValueDB *full;
    time_t stamp;
 
-   char *file;
+   xstring file;
    int fd;
+   bool modified;
 
    void Load();
    void Close();
@@ -42,8 +43,9 @@ class History : public KeyValueDB
    time_t extract_stamp(const char *res);
 
 public:
-   void Set(FileAccess *s,const char *cwd);
-   const char *Lookup(FileAccess *s);
+   void Set(const FileAccess *s,const FileAccess::Path &p);
+   void Set(const FileAccess *s) { Set(s,s->GetCwd()); }
+   const char *Lookup(const FileAccess *s);
    void Save();
    History();
    ~History();
