@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-1997 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2007 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: mgetJob.cc,v 1.17 2000/10/16 13:47:26 lav Exp $ */
+/* $Id: mgetJob.cc,v 1.24 2007/04/13 12:18:33 lav Exp $ */
 
 #include <config.h>
 #include <sys/types.h>
@@ -30,7 +30,7 @@
 #include "url.h"
 #include "mkdirJob.h"
 
-void mgetJob::ShowRunStatus(StatusLine *s)
+void mgetJob::ShowRunStatus(const SMTaskRef<StatusLine>& s)
 {
    if(rg)
    {
@@ -39,7 +39,7 @@ void mgetJob::ShowRunStatus(StatusLine *s)
    }
    GetJob::ShowRunStatus(s);
 }
-void mgetJob::PrintStatus(int v)
+void mgetJob::PrintStatus(int v,const char *prefix)
 {
    if(mkdir_job)
    {
@@ -48,14 +48,14 @@ void mgetJob::PrintStatus(int v)
    }
    if(rg)
    {
-      SessionJob::PrintStatus(v);
+      SessionJob::PrintStatus(v,prefix);
       const char *s=rg->Status();
       if(!s || !s[0])
 	 return;
       printf("\t%s\n",s);
       return;
    }
-   GetJob::PrintStatus(v);
+   GetJob::PrintStatus(v,prefix);
 }
 
 mgetJob::mgetJob(FileAccess *session,ArgV *a,bool c,bool md)
@@ -68,13 +68,13 @@ mgetJob::mgetJob(FileAccess *session,ArgV *a,bool c,bool md)
    mkdir_job=0;
    mkdir_args=0;
    mkdir_base_arg=0;
-   output_dir=0;
 }
 
-void mgetJob::make_directory(char *d)
+void mgetJob::make_directory(const char *d_c)
 {
    if(!make_dirs)
       return;
+   char *d=alloca_strdup(d_c);
    char *slash=strrchr(d,'/');
    if(!slash || slash==d)
       return;
@@ -84,7 +84,7 @@ void mgetJob::make_directory(char *d)
    const char *dir_name_c=output_file_name(d,0,!reverse,output_dir,make_dirs);
    char *dir_name;
    if(dir_name_c==0 || dir_name_c[0]==0)
-      goto leave;
+      return;
    dir_name=alloca_strdup(dir_name_c);
    if(reverse || url::is_url(dir_name))
    {
@@ -95,7 +95,7 @@ void mgetJob::make_directory(char *d)
 	    if(!strcmp(dir_name,mkdir_args->getarg(j)))
 	       break;
 	 if(j<mkdir_args->count()) // don't try to create dir twice
-	    goto leave;
+	    return;
       }
       if(!mkdir_args)
       {
@@ -110,8 +110,6 @@ void mgetJob::make_directory(char *d)
    {
       create_directories(dir_name);
    }
-leave:
-   *slash='/';
 }
 
 void mgetJob::LocalGlob(const char *p)
@@ -162,7 +160,7 @@ int mgetJob::Do()
    {
    next:
       if(rg) { delete rg; rg=0; }
-      char *p=m_args->getnext();
+      const char *p=m_args->getnext();
       if(!p)
       {
 	 delete m_args;
@@ -173,7 +171,7 @@ int mgetJob::Do()
 	    mkdir_job->BeQuiet();
 	    AddWaiting(mkdir_job);
 	    mkdir_job->SetParentFg(this);
-	    mkdir_job->cmdline=mkdir_args->Combine();
+	    mkdir_job->cmdline.set_allocated(mkdir_args->Combine());
 	    // don't delete mkdir_args; -- mkdirJob does it
 	    mkdir_args=0;
 	 }
@@ -182,10 +180,7 @@ int mgetJob::Do()
       if(reverse && !url::is_url(p))
 	 LocalGlob(expand_home_relative(p));
       else
-      {
-	 rg=new GlobURL(session,p);
-	 rg->glob->FilesOnly();
-      }
+	 rg=new GlobURL(session,p,GlobURL::FILES_ONLY);
       m=MOVED;
    }
 
@@ -206,11 +201,11 @@ int mgetJob::Do()
    FileSet *files=rg->GetResult();
    if(files->get_fnum()==0)
    {
-      fprintf(stderr,_("%s: no files found\n"),rg->glob->GetPattern());
+      fprintf(stderr,_("%s: no files found\n"),rg->GetPattern());
       goto next;
    }
    do {
-      char *src=files->curr()->name;
+      const char *src=files->curr()->name;
       args->Append(src);
       make_directory(src);
       args->Append(output_file_name(src,0,!reverse,output_dir,make_dirs));
@@ -220,11 +215,7 @@ int mgetJob::Do()
 
 mgetJob::~mgetJob()
 {
-   if(rg)
-      delete rg;
-   if(m_args)
-      delete m_args;
-   if(mkdir_args)
-      delete mkdir_args;
-   xfree(output_dir);
+   delete rg;
+   delete m_args;
+   delete mkdir_args;
 }
