@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-1997 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2007 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -26,20 +26,21 @@
 #include "url.h"
 #include "misc.h"
 
-mkdirJob::mkdirJob(FileAccess *s,ArgV *a) : SessionJob(s)
+#define super SessionJob
+#define orig_session super::session
+
+mkdirJob::mkdirJob(FileAccess *s,ArgV *a)
+   : super(s), args(a), session(orig_session)
 {
    quiet=false;
    failed=file_count=0;
 
-   args=a;
    args->rewind();
-   char *op=args->getarg(0);
+   const char *op=args->getarg(0);
 
    curr=first=0;
    opt_p=false;
 
-   url_session=0;
-
    int opt;
    while((opt=args->getopt("p"))!=EOF)
    {
@@ -55,75 +56,67 @@ mkdirJob::mkdirJob(FileAccess *s,ArgV *a) : SessionJob(s)
       fprintf(stderr,"Usage: %s [-p] files...\n",op);
       return;
    }
-
-}
-
-mkdirJob::~mkdirJob()
-{
-   delete args;
-   args=0;
-   Reuse();
 }
 
 int mkdirJob::Do()
 {
    if(Done())
       return STALL;
-   if(Session()->IsClosed())
+   if(session->IsClosed())
    {
       ParsedURL u(curr,true);
       if(u.proto)
-	 url_session=FileAccess::New(&u);
-      if(url_session)
-	 url_session->Mkdir(u.path,opt_p);
+      {
+	 session=my_session=FileAccess::New(&u);
+	 session->SetPriority(fg?1:0);
+	 session->Mkdir(u.path,opt_p);
+      }
       else
+      {
+	 session=orig_session;
 	 session->Mkdir(curr,opt_p);
+      }
    }
 
-   int res=Session()->Done();
+   int res=session->Done();
    if(res==FA::DO_AGAIN || res==FA::IN_PROGRESS)
       return STALL;
    if(res<0)
    {
       failed++;
       if(!quiet)
-	 fprintf(stderr,"%s: %s\n",args->getarg(0),Session()->StrError(res));
+	 fprintf(stderr,"%s: %s\n",args->getarg(0),session->StrError(res));
    }
    file_count++;
-   Session()->Close();
-   if(url_session)
-   {
-      SessionPool::Reuse(url_session);
-      url_session=0;
-   }
+   session->Close();
    curr=args->getnext();
    return MOVED;
 }
 
-void  mkdirJob::PrintStatus(int v)
+void  mkdirJob::PrintStatus(int v,const char *prefix)
 {
-   SessionJob::PrintStatus(v);
+   SessionJob::PrintStatus(v,prefix);
    if(Done())
       return;
 
-   printf("\t`%s' [%s]\n",curr,Session()->CurrentStatus());
+   printf("%s`%s' [%s]\n",prefix,curr,session->CurrentStatus());
 }
 
-void  mkdirJob::ShowRunStatus(StatusLine *s)
+void  mkdirJob::ShowRunStatus(const SMTaskRef<StatusLine>& s)
 {
    if(Done())
       return;
 
    s->Show("%s `%s' [%s]",args->getarg(0),
       squeeze_file_name(curr,s->GetWidthDelayed()-40),
-      Session()->CurrentStatus());
+      session->CurrentStatus());
 }
 
 void  mkdirJob::SayFinal()
 {
    if(failed==file_count)
       return;
-   char *op=args->getarg(0);
+   const char *op=args->getarg(0);
    if(file_count==1)
       // A directory has just been created
       printf(_("%s ok, `%s' created\n"),op,first);
@@ -134,3 +127,16 @@ void  mkdirJob::SayFinal()
       printf(plural("%s ok, %d director$y|ies$ created\n",file_count),
 	    op,file_count);
 }
+
+void mkdirJob::Fg()
+{
+   super::Fg();
+   if(orig_session!=session)
+      orig_session->SetPriority(1);
+}
+void mkdirJob::Bg()
+{
+   if(orig_session!=session)
+      orig_session->SetPriority(0);
+   super::Bg();
+}
