@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: resource.cc,v 1.136 2006/08/29 07:25:27 lav Exp $ */
+/* $Id: resource.cc,v 1.146 2007/11/08 07:59:30 lav Exp $ */
 
 #include <config.h>
 
@@ -29,20 +29,21 @@
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <errno.h>
+#include <stddef.h>
 
 #include "ResMgr.h"
 #include "url.h"
 #include "GetPass.h"
 #include "ascii_ctype.h"
 #include "confpaths.h"
+#include "misc.h"
 
-static const char *FtpProxyValidate(char **p)
+static const char *FtpProxyValidate(xstring_c *p)
 {
    ParsedURL url(*p);
    if(url.host==0)
    {
-      if((*p)[0]!=0)
-	 (*p)[0]=0;
+      p->truncate(0);
       return 0;
    }
    if(url.proto)
@@ -52,20 +53,18 @@ static const char *FtpProxyValidate(char **p)
    }
    if(url.user && !url.pass)
    {
-      url.pass=GetPass(_("ftp:proxy password: "));
-      xfree(*p);
-      *p=url.Combine();
+      url.pass.set(GetPass(_("ftp:proxy password: ")));
+      p->set_allocated(url.Combine());
    }
    return 0;
 }
 
-static const char *HttpProxyValidate(char **p)
+static const char *HttpProxyValidate(xstring_c *p)
 {
    ParsedURL url(*p);
    if(url.host==0)
    {
-      if((*p)[0]!=0)
-	 (*p)[0]=0;
+      p->truncate(0);
       return 0;
    }
    if(url.proto)
@@ -77,28 +76,11 @@ static const char *HttpProxyValidate(char **p)
    return 0;
 }
 
-#if 0
-static const char *FtpDefaultAnonPass()
-{
-   static char *pass=0;
-
-   if(pass)
-      return pass;
-
-   struct passwd *pw=getpwuid(getuid());
-   const char *u=pw?pw->pw_name:"unknown";
-   pass=(char*)xmalloc(strlen(u)+3);
-   sprintf(pass,"%s@",u);
-
-   return pass;
-}
-#endif
-
-static const char *PutOrPost(char **s)
+static const char *PutOrPost(xstring_c *s)
 {
    if(strcasecmp(*s,"PUT") && strcasecmp(*s,"POST"))
       return _("only PUT and POST values allowed");
-   for(char *scan=*s; *scan; scan++)
+   for(char *scan=s->get_non_const(); *scan; scan++)
       *scan=to_ascii_upper((unsigned char)*scan);
    return 0;
 }
@@ -112,15 +94,13 @@ static const char *const af_list[]=
    0
 };
 static
-const char *OrderValidate(char **s)
+const char *OrderValidate(xstring_c *s)
 {
-   static char *error=0;
+   static xstring error;
+   xstring fixed;
 
    const char * const delim="\t ";
    char *s1=alloca_strdup(*s);
-   char *fixed=(char*)xmalloc(strlen(s1)+1);
-   fixed[0]=0;
-
    for(s1=strtok(s1,delim); s1; s1=strtok(0,delim))
    {
       const char *const *f;
@@ -131,25 +111,23 @@ const char *OrderValidate(char **s)
       }
       if(!*f)
       {
-	 const char * const format=_("unknown address family `%s'");
-	 error=(char*)xrealloc(error,strlen(format)+strlen(s1));
-	 sprintf(error,format,s1);
+	 error.setf(_("unknown address family `%s'"),s1);
 	 return error;
       }
-      if(fixed[0])
-	 strcat(fixed," ");
-      strcat(fixed,s1);
+      if(fixed)
+	 fixed.vappend(" ",s1,NULL);
+      else
+	 fixed.set(s1);
    }
-   xfree(*s);
-   *s=fixed;
+   s->set(fixed);
    return 0;
 }
 
 #if USE_SSL
 static
-const char *AuthArgValidate(char **s)
+const char *AuthArgValidate(xstring_c *s)
 {
-   for(char *i=*s; *i; i++)
+   for(char *i=s->get_non_const(); *i; i++)
       *i=to_ascii_upper(*i);
 
    if(strcmp(*s,"SSL")
@@ -161,12 +139,12 @@ const char *AuthArgValidate(char **s)
    return 0;
 }
 static
-const char *ProtValidate(char **s)
+const char *ProtValidate(xstring_c *s)
 {
    if(!**s)
       return 0;
 
-   for(char *i=*s; *i; i++)
+   for(char *i=s->get_non_const(); *i; i++)
       *i=to_ascii_upper(*i);
 
    if(strcmp(*s,"P")
@@ -239,6 +217,7 @@ static ResType lftp_vars[] = {
    {"ftp:use-site-utime",	 "no",	  ResMgr::BoolValidate,0},
    {"ftp:use-size",		 "yes",   ResMgr::BoolValidate,0},
    {"ftp:use-stat",		 "yes",   ResMgr::BoolValidate,0},
+   {"ftp:use-stat-for-list",	 "no",	  ResMgr::BoolValidate,0},
    {"ftp:use-telnet-iac",	 "yes",   ResMgr::BoolValidate,0},
    {"ftp:use-quit",		 "yes",   ResMgr::BoolValidate,0},
    {"ftp:verify-address",	 "no",    ResMgr::BoolValidate,0},
@@ -246,7 +225,7 @@ static ResType lftp_vars[] = {
    {"ftp:web-mode",		 "off",   ResMgr::BoolValidate,0},
 #define RETRY_530 \
    "too many|overloaded|try (again |back )?later|is restricted to|"\
-   "maximum number|number of connect|only.*session.*allowed|more connection"
+   "maximum number|number of connect|only.*session.*allowed|more connection|already connected"
    {"ftp:retry-530",		 RETRY_530,ResMgr::ERegExpValidate,0},
    {"ftp:retry-530-anonymous",	 "Login incorrect",ResMgr::ERegExpValidate,0},
    {"hftp:cache",		 "yes",   ResMgr::BoolValidate,0},
@@ -340,12 +319,14 @@ static ResType lftp_vars[] = {
    {"xfer:destination-directory","",	  0,0},
    {"xfer:verify",		 "no",	  ResMgr::BoolValidate,ResMgr::NoClosure},
    {"xfer:verify-command",	 "",	  ResMgr::FileExecutable,0},
+   {"xfer:log",			 "yes",	  ResMgr::BoolValidate,ResMgr::NoClosure},
 
 #if USE_SSL
    {"ssl:ca-file",		 "",	  ResMgr::FileReadable,ResMgr::NoClosure},
    {"ssl:crl-file",		 "",	  ResMgr::FileReadable,ResMgr::NoClosure},
    {"ssl:key-file",		 "",	  ResMgr::FileReadable,0},
    {"ssl:cert-file",		 "",	  ResMgr::FileReadable,0},
+   {"ssl:check-hostname",	 "yes",	  ResMgr::BoolValidate,0},
    {"ssl:verify-certificate",	 "no",	  ResMgr::BoolValidate,0},
 # if USE_OPENSSL
    {"ssl:ca-path",		 "",	  ResMgr::DirReadable,ResMgr::NoClosure},
@@ -370,12 +351,12 @@ void ResMgr::ClassInit()
    {
       if(scan->defvalue && scan->val_valid)
       {
-	 char *dv=xstrdup(scan->defvalue);
+	 xstring_c dv(scan->defvalue);
 	 const char *error=(*scan->val_valid)(&dv);
 	 if(error)
 	    fprintf(stderr,"Default value for %s is invalid: %s\n",scan->name,error);
 	 else if(strcmp(dv,scan->defvalue))
-	    fprintf(stderr,"Default value for %s (%s) is not in canonic form: %s\n",scan->name,scan->defvalue,dv);
+	    fprintf(stderr,"Default value for %s (%s) is not in canonic form: %s\n",scan->name,scan->defvalue,dv.get());
       }
    }
 
