@@ -19,13 +19,12 @@
  * Boston, MA 02111-1307, USA.
  */
 
-/* $Id: xmalloc.h,v 1.7.2.1 2000/06/08 23:08:37 lav Exp $ */
+/* $Id: xmalloc.h,v 1.14 2007/06/08 14:10:17 lav Exp $ */
 
 #ifndef XMALLOC_H
 #define XMALLOC_H
 
 #include <stdlib.h>
-#include "xstring.h"
 #include "xalloca.h"
 
 #ifdef DBMALLOC
@@ -35,7 +34,17 @@
 void *xmalloc(size_t);
 void *xrealloc(void *,size_t);
 char *xstrdup(const char *s,int spare=0);
-#define alloca_strdup(s) ((s)?strcpy((char*)alloca(strlen((s))+1),(s)):0)
+char *xstrset(char *&mem,const char *s);
+char *xstrset(char *&mem,const char *s,size_t n);
+#define alloca_strdup(s) alloca_strdup2((s),0)
+#define alloca_strdup2(s,n) ((s)?strcpy((char*)alloca(strlen((s))+1+n),(s)) \
+                                :((n)==0?0:(char*)alloca((n))))
+#define alloca_append(s1,s2) strcat(alloca_strdup2((s1),strlen((s2))),(s2));
+
+void xfree(void *p);
+void xmalloc_register_block(void *);
+
+#include "xstring.h"
 
 static inline void *xmemdup(const void *m,int len)
 {
@@ -44,8 +53,5 @@ static inline void *xmemdup(const void *m,int len)
    memcpy(buf,m,len);
    return buf;
 }
-void xfree(void *p);
-
-void xmalloc_register_block(void *);
 
 #endif /* XMALLOC_H */
